#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# class stuff for circles
#

import tolerance
import point
import graphicobject
import style
import linetype
import color
import math
import tree
import maptree

class _Circle(object):
    """A base-class for Circles, Arcs, and Construction Circles.

A _Circle has two attributes:

center: A Point object
radius: The _Circle's radius

A _Circle has the following methods:

{get/set}Center(): Get/Set the center Point of a _Circle.
{get/set}Radius(): Get/Set the radius of a _Circle.
move(): Move the _Circle.
circumference(): Get the _Circle's circumference.
area(): Get the _Circle's area.
mapPoint(): Find the nearest Point on the _Circle to some other Point.
inRegion(): Returns whether or not a _Circle can be seen in a bounded area.
clone(): Return an indentical copy of a _Circle.
    """

    # member functions

    def __init__(self, center, radius):
        """Initialize a _Circle.

_Circle(center, radius)

The center should be a Point, or a two-entry tuple of floats,
and the radius should be a float greater than 0.
        """
        _c = center
        if not isinstance(_c, point._Point):
            _c = point._Point(center)
        _r = radius
        if not isinstance(_r, float):
            _r = float(radius)
        if _r < 1e-10:
            raise ValueError, "Invalid radius: %g" % _r
        self.__center = _c
        self.__radius = _r

    def __cmp__(self, obj):
        """Compare two _Circles or a _Circle and Point.

The _Circle comparison goes as follows:

1) Get the two center points
2) Get the two radii

Having obtained those values, the minimum and maximum
"x" and "y" coordinates of the circle are calculated.

xmin = ("x" of center) - radius
xmax = ("x" of center) + radius
ymin = ("y" of center) - radius
ymax = ("y" of center) + radius

xmin of C1 < xmin of C2 ==> return -1
xmax of C1 > xmax of C2 ==> return 1
ymin of C1 < ymin of C2 ==> return -1
ymax of C1 > ymax of C2 ==> return 1

The radii are compared if we make it here

radius of C1 < radius of C2 ==> return -1
radius of C1 > radius of C2 ==> return 1

If we get here, return 0

The _Circle/Point comparison goes as follows.

1) Get the center point and radius of C and coordinates of P
2) Circle minima and maxima are calculated as above

The function returns the following

1: P.x < C.xmin
0: S.xmin <= P.x <= S.xmax and S.ymin <= P.y <= S.ymax
-1: Any other condition
        """
        if not isinstance(obj, (_Circle, point._Point)):
            raise TypeError, "Invalid object for _Circle comparison: " + `obj`
        _xs, _ys = self.__center.getCoords()
        _rs = self.__radius
        _sxmin = _xs - _rs
        if isinstance(obj, _Circle):
            _x, _y = obj.getCenter().getCoords()
            _r = obj.getRadius()
            _xmin = _x - _r
            if abs(_sxmin - _xmin) < 1e-10:
                _symin = _ys - _rs
                _ymin = _y - _r
                if abs(_symin - _ymin) < 1e-10:
                    if abs(_rs - _r) < 1e-10:
                        _val = 0
                    elif _rs > _r:
                        _val = 1
                    else:
                        _val = -1
                elif _symin > _ymin:
                    _val = 1
                else:
                    _val = -1
            elif _sxmin > _xmin:
                _val = 1
            else:
                _val = -1
        else:
            _x, _y = obj.getCoords()
            if _x < _sxmin - 1e-10:
                _val = 1
            else:
                _val = -1
                _symin = _ys - _rs - 1e-10
                _sxmax = _xs + _rs + 1e-10
                _symax = _ys + _rs + 1e-10
                if _x < _sxmax and _symin < _y < _symax:
                    _val = 0
        return _val

    def getCenter(self):
        """Return the center Point of the _Circle.

getCenter()        
        """
        return self.__center

    def setCenter(self, c):
        """Set the center Point of the _Circle.

setCenter(c)

The argument must be a Point or a tuple containing
two float values.
        """
        _c = c
        if not isinstance(_c, point._Point):
            _c = point._Point(c)
        self.__center = _c
        
    center = property(getCenter, setCenter, None, "_Circle center")
    
    def getRadius(self):
        """Return the radius of the the _Circle.

getRadius()        
        """
        return self.__radius

    def setRadius(self, r):
        """Set the radius of the _Circle.

setRadius(r)

The argument must be float value greater than 0.
        """
        _r = r
        if not isinstance(_r, float):
            _r = float(r)
        if _r < 1e-10:
            raise ValueError, "Invalid radius: %g" % _r
        self.__radius = _r
        
    radius = property(getRadius, setRadius, None, "_Circle radius")

    def move(self, dx, dy):
        """Move a _Circle.

move(dx, dy)

The first argument gives the "x" coordinate displacement,
and the second gives the "y" coordinate displacement. Both
values should be floats.
        """
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        self.__center.move(_dx, _dy)

    def circumference(self):
        """Return the circumference of the _Circle.

circumference()
        """
        return 2.0 * math.pi * self.__radius

    def area(self):
        """Return the area enclosed by the _Circle.

area()
        """
        return math.pi * pow(self.__radius, 2)
    
    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest Point on the Circle to the Point p.

mapPoint(p [, tol])

The function has one required argument:

p: A Point object or a tuple of two floats

There is a single optional argument:

tol: A float value equal or greater than 0.

This function is used to map a possibly near-by Point object to an
actual Point on the _Circle. If the distance between the actual
Point and the Point used as an argument is less than the Tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _p = p
        if not isinstance(_p, point._Point):
            _p = point._Point(p)
        _t = tolerance.toltest(tol)
        _xp, _yp = _p.getCoords()
        _xc, _yc = self.__center.getCoords()
        _rad = self.__radius
        _dist = math.hypot((_xp - _xc), (_yp - _yc))
        _ap = None
        _sep = abs(_dist - _rad)
        if _sep < _t or abs(_sep - _t) < 1e-10:
            _angle = math.atan2((_yp - _yc),(_xp - _xc))
            _xoff = _rad * math.cos(_angle)
            _yoff = _rad * math.sin(_angle)
            _ap = point._Point((_xc + _xoff), (_yc + _yoff))
        return _ap

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not an _Circle exists within a region.

inRegion(xmin, ymin, xmax, ymax[, fully])

The first four arguments define the boundary. The optional
fifth argument "fully" indicates whether or not the _Circle
must be completely contained within the region or just pass
through it.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        _x, _y = self.__center.getCoords()
        _r = self.__radius
        # print "cx: %g; cy: %g; r: %g" % (_x, _y, _r)
        _cxmin = _x - _r
        _cymin = _y - _r
        _cxmax = _x + _r
        _cymax = _y + _r
        # print "cxmin: %g" % _cxmin
        # print "cymin: %g" % _cymin
        # print "cxmax: %g" % _cxmax
        # print "cymax: %g" % _cymax
        _val = False
        if (_xmin < _cxmin and _xmax > _cxmax and
            _ymin < _cymin and _ymax > _cymax):
            _val = True
        elif (not fully and
              not (_cxmax < _xmin or _cymax < _ymin or
                   _cxmin > _xmax or _cymin > _ymax)):
            _d1 = math.hypot((_xmin - _x), (_ymin - _y))
            _d2 = math.hypot((_xmin - _x), (_ymax - _y))
            _d3 = math.hypot((_xmax - _x), (_ymax - _y))
            _d4 = math.hypot((_xmax - _x), (_ymin - _y))
            _dmax = max(_d1, _d2, _d3, _d4)
            _dmin = min(_d1, _d2, _d3, _d4)
            if _xmin < _x < _xmax and _ymin < _y < _ymax:
                _dmin = -1e-10
            else:
                if _x > _xmax and _ymin < _y < _ymax:
                    _dmin = _x - _xmax
                elif _x < _xmin and _ymin < _y < _ymax:
                    _dmin = _xmin - _x
                elif _y > _ymax and _xmin < _x < _xmax:
                    _dmin = _y - _ymax
                elif _y < _ymin and _xmin < _x < _xmax:
                    _dmin = _ymin - _y
            if _dmin < _r < _dmax:
                _val = True
        return _val

    def clone(self):
        """Create an identical copy of a _Circle

clone()
        """
        _cp = self.__center.clone()
        return _Circle(_cp, self.__radius)
        
class Circle(_Circle, tree.TreeItem, graphicobject.GraphicObject):
    """The storable and displayed representation of a _Circle.

A Circle object is derived from a _Circle object, TreeItem object,
and a GraphicObject, so it has the attributes and member functions
as in those objects.
    """

    # member functions
    
    def __init__(self, center, radius, st=style.Style('Default'),
                 lt=None, col=None, t=None):
        """Initialize a Circle object.

Circle(center, radius, [st, lt, col, t])

There are two required arguments:

center: A Point object for the Circle's center
radius: The Circle's radius

Optional arguments:
st: A Style object
lt: A Linetype object that overrides the Style linetype
col: A Color object that overrides the Style color
t: A float that overrides the Style thickness
        """
        _Circle.__init__(self, center, radius)
        tree.TreeItem.__init__(self)
        graphicobject.GraphicObject.__init__(self, st, lt, col, t)

    def __str__(self):
        _c = self.getCenter().getCoords()
        _r = self.getRadius()
        return "Circle: Center: %s and radius: %g" % (_c, _r)

    def isModified(self):
        """Indicate if the Circle has been changed.

isModified()

This method extends the TreeItem::isModified() method.
        """
        if self.getCenter().isModified():
            self.modified()
        return tree.TreeItem.isModified(self)

    def setCenter(self, c):
        """Set the center Point of the Circle.

setCenter(c)

This method extends the _Circle::setCenter() method and sets
the modified flag indicating the Circle has been altered.
        """
        _Circle.setCenter(self, c)
        self.modified()

    center = property(_Circle.getCenter, setCenter, None,
                      "Circle center point.")
    
    def setRadius(self, r):
        """Set the radius of the Circle.

setRadius(r)

This method extends the _Circle::setRadius() method and sets
the modified flag indicating the Circle has been altered.
        """
        _Circle.setRadius(self, r)
        self.modified()

    radius = property(_Circle.getRadius, setRadius, None,
                      "Circle radius.")

    def move(self, dx, dy):
        """Move a Circle.

move(dx, dy)

This method extends the _Circle::move() method and sets
the modified flag indicating the Circle has been altered.
        """
        _Circle.move(self, dx, dy)
        self.modified()
        
    def clone(self):
        """Create an identical copy of a Circle.

clone()        
        """
        _ccp = self.getCenter().clone()
        _rad = self.getRadius()
        _st = self.getStyle()
        _lt = self.getLinetype()
        _col = self.getColor()
        _th = self.getThickness()
        return Circle(_ccp, _rad, _st, _lt, _col, _th)
    
    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest Point on the Circle to the Point p.

mapPoint(p [, tol])

This method extends _Circle::mapPoint().
        """
        _pt = _Circle.mapPoint(self, p, tol)
        if _pt is not None:
            _x, _y = _pt.getCoords()
            _pt = point.Point(_x, _y)
        return _pt
    
class CircleTree(maptree.PointMapTree):
    """A Tree class specific for Circle objects.

This class is based on the PointMapTree class, and
holds only Circle objects.
    """
    def __init__(self):
        """Initialize a CircleTree.

CircleTree()
        """
        maptree.PointMapTree.__init__(self, Circle)

    def binscan(self, obj, tol=tolerance.TOL):
        """Find a Circle object held in a CircleTree via binary search.

binscan(obj [, tol])

This method has one required parameter:

obj: A Circle object

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     points are close enough to be considered equal.

If a Circle object is found in the CircleTree within the
tolerance, the index of that object in the CircleTree is
returned. Otherwise, the function returns None.
        """
        if not isinstance(obj, Circle):
            raise TypeError, "Invalid type for searching in CircleTree: " + `obj`
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "CircleTree in modified state."
        _lo = 0
        _hi = len(self)
        _scanlist = []
        _ro = None
        while _lo < _hi:
            _mid = (_hi + _lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _cir = self[_mid]
            _res = cmp(obj, _cir)
            if _res == 0:
                _ro = _cir
                break
            elif _res == 1:
                _lo = _mid + 1
            else:
                _hi = _mid
        if _ro is None:
            _cx, _cy = obj.getCenter().getCoords()
            _cr = obj.getRadius()
            _xmin = _cx - _cr
            for _i in xrange(_mid,-1,-1):
                _tc = self[_i]
                _tcx, _tcy = _tc.getCenter().getCoords()
                _tcr = _tc.getRadius()
                _txmin = _tcx - _tcr
                if _txmin < _xmin:
                    break
                if abs(_tcx - _cx) < _t:
                    if abs(_tcy - _cy) < _t:
                        if abs(_tcr - _cr) < 1e-10:
                            _ro = _tc
                            break
        return _ro
    
    def scan(self, obj, tol=tolerance.TOL):
        """Find a Circle object held in a CircleTree via linear search.

scan(obj [, tol])

This method has one required parameter:

obj: A Circle object

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     points are close enough to be considered equal.

If a Circle object is found in the CircleTree within the
tolerance, the index of that object in the CircleTree is
returned. Otherwise, the function returns None.
        """
        if not isinstance(obj, Circle):
            raise TypeError, "Invalid type for searching in CircleTree: " + `obj`
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "Circle tree in modified state"
        _cx, _cy = obj.getCenter().getCoords()
        _cr = obj.getRadius()
        _xmin = _cx - _cr
        _ro = None
        for _sobj in self:
            _sx, _sy = _sobj.getCenter().getCoords()
            _sr = _sobj.getRadius()
            _sxmin = _sx - _sr
            if _sxmin > _xmin:
                break
            if abs(_sx - _cx) < _t:
                if abs(_sy - _cy) < _t:
                    if abs(_sr - _cr) < 1e-10:
                        _ro = _sobj
                        break
        return _ro

    def atCoords(self, rad, x, y):
        """Find a Circle at a specific location and some radius.

atCoords(rad, x, y)

The arguments for this method are:

rad: A float giving the radius of the Circle.
x: A float giving the center "x" coordinate.
y: A float giving the center "y" coordinate.


This method returns a Circle if there is one in the tree
at the specified location and with the desired radius.
Otherwise this method returns None.
        """
        _rad = rad
        if not isinstance(_rad, float):
            _rad = float(rad)
        if _rad < 1e-10:
            raise ValueError, "Invalid radius: " + `_rad`
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        if self.isModified():
            raise ValueError, "CircleTree in modified state."
        _lo = 0
        _hi = len(self)
        _scanlist = []
        _ro = None
        while _lo < _hi:
            _mid = (_hi + _lo)//2
            if _mid in _scanlist:
                break
            _scanlist.append(_mid)
            _cir = self[_mid]
            _cx, _cy = _cir.getCenter().getCoords()
            _cr = _cir.getRadius()
            if abs(_cx - _x) < 1e-10:
                if abs(_cy - _y) < 1e-10:
                    if abs(_cr - _rad) < 1e-10:
                        _ro = _cir
                        break
                    elif _rad < _cr:
                        _lo = _mid + 1
                    else:
                        _hi = _mid
                elif _y < (_cy - _cr):
                    _lo = _mid + 1
                else:
                    _hi = _mid
            elif _x < (_cx - _cr):
                _lo = _mid + 1
            else:
                _hi = _mid
        return _ro
