#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# generic dimension classes
#

from math import pi, hypot, sqrt, atan2, sin, cos
import array

import text
import point
import circle
import arc
import color
import layer
import units
import util
import tolerance
import options

class DimString(text.TextStyle):
    """A class for the visual presentation of the dimensional value.

The DimString class is used to present the numerical display for
the dimension. A DimString object is derived from the text.TextStyle
class, so it shares all of that classes methods and attributes.

A DimString class has the following methods

{get/set}Prefix(): Get/Set the text preceding the dimension value.
{get/set}Suffix(): Get/Set the text following the dimension value.
{get/set}Units(): Define what units the dimension will be in.
{get/set}Precision(): Get/Set how many fractional digits are displayed.
{get/set}PrintZero(): Get/Set whether or not a dimension less than one
                      unit long should have a leading 0 printed
{get/set}PrintDecimal(): Get/Set whether or not a dimension with 0
                         fractional digits prints out the decimal point.
                         """
    def __init__(self):
        """Instatiate a DimString object

ds = DimString()

By default, a DimString object has the following values ...

prefix: Empty string
suffix: Empty string
units: Millimeters
precision: 3 decimal places
print zero: True
print decimal: True
        """
        text.TextStyle.__init__(self)
        self.__prefix = u''
        self.__suffix = u''
        self.__unit = units.Unit(units.MILLIMETERS)
        self.__precision = 3
        self.__print_zero = True
        self.__print_decimal = True

    def getPrefix(self):
        """Return the prefix for the DimString object.

getPrefix()
        """
        return self.__prefix

    def setPrefix(self, prefix=u''):
        """Set the prefix for the DimString object.

setPrefix([prefix])

Invoking this method without arguments sets the prefix
to an empty string. When an argument is passed, the argument
should be a Unicode string.
        """
        _p = prefix
        if not isinstance(_p, unicode):
            _p = unicode(prefix)
        self.__prefix = _p

    prefix = property(getPrefix, setPrefix, None, "Dimension string prefix")

    def getSuffix(self):
        """Return the suffix for the DimString object.

getSuffix()        
        """
        return self.__suffix

    def setSuffix(self, suffix=u''):
        """Set the suffix for the DimString object.

setSuffix([suffix])

Invoking this method without arguments sets the suffix
to an empty string. When an argument is passed, the argument
should be a Unicode string.
        """
        _s = suffix
        if not isinstance(_s, unicode):
            _s = unicode(suffix)
        self.__suffix = _s

    suffix = property(getSuffix, setSuffix, None, "Dimension string suffix")
    
    def getPrecision(self):
        """Return the number of decimal points used for the DimString.

getPrecision()
        """
        return self.__precision

    def setPrecision(self, p):
        """Set the number of decimal points used for the DimString.

setPrecision(p)

The valid range of "p" is 0 <= p <= 15.
        """
        _p = p
        if not isinstance(_p, int):
            _p = int(p)
        if _p < 0 or _p > 15:
            raise ValueError, "Invalid precision: " + `_p`
        self.__precision = _p

    precision = property(getPrecision, setPrecision, None,
                         "Dimension precision")

    def getUnits(self):
        """Return the current units used in the DimString().

getUnits()        
        """
        return self.__unit.getUnit()

    def setUnits(self, unit):
        """The the units for the DimString.

setUnits(unit)

The value units are given in the units module.
        """
        self.__unit.setUnit(unit)

    units = property(getUnits, setUnits, None, "Dimensional units.")
        
    def getPrintZero(self):
        """Return whether or not a leading 0 is printed for the DimString.

getPrintZero()
        """
        return self.__print_zero

    def setPrintZero(self, pz=True):
        """Set whether or not a leading 0 is printed for the DimString.

setPrintZero([pz])

Invoking this method without arguments sets the value to True.
If called with an argument, the argument should be either True
or False.
        """
        if pz is True or pz is False:
            self.__print_zero = pz
        else:
            raise ValueError, "Invalid zero print flag: " + `pz`
        
    print_zero = property(getPrintZero, setPrintZero, None,
                          "Print a leading 0 for decimal dimensions")

    def getPrintDecimal(self):
        """Return whether or not the DimString will print a trailing decimal.

getPrintDecimal()        
        """
        return self.__print_decimal

    def setPrintDecimal(self, pd=True):
        """Set whether or not the DimString will print a trailing decimal.

setPrintDecimal([pd])

Invoking this method without arguments sets the value to True.
If called with an argument, the argument should be either True
or False.
        """
        if pd is True or pd is False:
            self.__print_decimal = pd
        else:
            raise ValueError, "Invalid print decimal flag: " + `pd`

    print_decimal = property(getPrintDecimal, setPrintDecimal, None,
                             "Print a decimal point after the dimension value")

    def formatDimension(self, dist):
        """Return a formatted numerical value for a dimension.

formatDimension(dist)

The argument "dist" should be a float value representing the
distance in millimeters. The returned value will have the
prefix prepended and suffix appended to the numerical value
that has been formatted with the precision.
        """
        _d = dist
        if not isinstance(_d, float):
            _d = float(dist)
        _d = abs(_d)
        _fmtstr = u"%%#.%df" % self.__precision
        _dstr = _fmtstr % self.__unit.fromMillimeters(_d)
        if _d < 1.0 and self.__print_zero is False:
            _dstr = _dstr[1:]
        if _dstr.endswith('.') and self.__print_decimal is False:
            _dstr = _dstr[:-1]
        return self.__prefix + _dstr + self.__suffix

dim_color = color.Color(255,165,0)

class Dimension(object):
    """The base class for Dimensions

A Dimension object is meant to be a base class for specialized
dimensions.

Every Dimension object holds two DimString objects, so any
dimension can be displayed with two separate formatting options
and units.

A Dimension has the following methods

{get/set}DimStyle(): Get/Set the DimStyle used for this Dimension.
applyDimStyle(): Set all the Dimension attributes to the values given
                 in the DimStyle
getPrimaryDimstring(): Return the DimString used for formatting the
                       Primary dimension.
getSecondaryDimstring(): Return the DimString used for formatting the
                         Secondary dimension.
{get/set}Endpoint(): Get/Set the type of endpoints used in the Dimension
{get/set}EndpointSize(): Get/Set the size of the dimension endpoints
{get/set}DualDimMode(): Get/Set whether or not to display both the Primary
                        and Secondary DimString objects
{get/set}Offset(): Get/Set how far from the dimension endpoints to draw
                   dimension lines at the edges of the dimension.
{get/set}Extension(): Get/Set how far past the dimension crossbar line
                      to draw.
{get/set}Position(): Get/Set where the dimensional values are placed on the
                     dimension cross bar.
{get/set}Color(): Get/Set the color used to draw the dimension lines.
{get/set}Location(): Get/Set where to draw the dimensional values.
getDimensions(): Return the formatted dimensional values in this Dimension.
isModified(): Return if the dimension has had its endpoints adjusted.
modified(): Flag the dimension endpoints or circle have been modified.
reset(): Set the dimension state to be unmodified.
inRegion(): Return if the dimension is visible within some are.
calcDimValues(): Calculate the dimension lines endpoints.
onDimension(): Test if an x/y coordinate pair hit the dimension lines.
    """
    #
    # Endpoint
    #
    NO_ENDPOINT = 0
    ARROW = 1
    FILLED_ARROW = 2
    SLASH = 3
    CIRCLE = 4

    #
    # Dimension position on dimline
    #
    SPLIT = 0
    ABOVE = 1
    BELOW = 2
    
    def __init__(self, ds):
        """Instatiate a Dimension object

dim = Dimension(ds)

A Dimension object requires a DimStyle object to construct
itself. The DimStyle object provides a collection of values
used to format the two DimString objects and the various other
parts of a dimension.
        """
        if not isinstance(ds, DimStyle):
            raise TypeError, "Invalid DimStyle: " + `ds`
        self.__dimstyle = ds
        self.__dimstring1 = DimString()
        self.__dimstring2 = DimString()
        self.__dual_mode = False # Display primary dimension only
        self.__offset = 0.0
        self.__extension = 0.0
        self.__dim_pos = Dimension.SPLIT
        self.__endpoint_type = Dimension.NO_ENDPOINT
        self.__endpoint_size = 1.0
        self.__color = dim_color        
        self.__dim_loc_x = None
        self.__dim_loc_y = None
        self.__modified = False

    def getDimStyle(self):
        """Return the DimStyle used in this Dimension.

getDimStyle()        
        """
        return self.__dimstyle

    def setDimStyle(self, ds):
        """Set the DimStyle used for this Dimension.

setDimStyle(ds)

After setting the DimStyle, the values stored in it
are applied to the DimensionObject.
        """
        if not isinstance(ds, DimStyle):
            raise TypeError, "Invalid DimStyle: " + `ds`
        self.__dimstyle = ds
        self.applyDimStyle()

    dimstyle = property(getDimStyle, setDimStyle, None,
                        "Dimension DimStyle object.")

    def applyDimStyle(self):
        """Apply the values in the DimStyle to the Dimension.

applyDimStyle()
        """
        _ds = self.__dimstyle
        self.__offset = _ds.getValue('DIM_OFFSET')
        self.__extension = _ds.getValue('DIM_EXTENSION')
        self.__dim_pos = _ds.getValue('DIM_POSITION')
        self.__endpoint_type = _ds.getValue('DIM_ENDPOINT')
        self.__endpoint_size = _ds.getValue('DIM_ENDPOINT_SIZE')
        self.__color = _ds.getValue('DIM_COLOR')
        self.__dual_mode = _ds.getValue('DIM_DUAL_MODE')
        #
        # primary dimension string
        #
        _pds = self.getPrimaryDimstring()
        _pds.setPrefix(_ds.getValue('DIM_PRIMARY_PREFIX'))
        _pds.setSuffix(_ds.getValue('DIM_PRIMARY_SUFFIX'))
        _pds.setPrecision(_ds.getValue('DIM_PRIMARY_PRECISION'))
        _pds.setUnits(_ds.getValue('DIM_PRIMARY_UNITS'))
        _pds.setPrintZero(_ds.getValue('DIM_PRIMARY_LEADING_ZERO'))
        _pds.setPrintDecimal(_ds.getValue('DIM_PRIMARY_TRAILING_DECIMAL'))
        _pds.setFamily(_ds.getValue('DIM_PRIMARY_FONT_FAMILY'))
        _pds.setSize(_ds.getValue('DIM_PRIMARY_FONT_SIZE'))
        _pds.setWeight(_ds.getValue('DIM_PRIMARY_FONT_WEIGHT'))
        _pds.setStyle(_ds.getValue('DIM_PRIMARY_FONT_STYLE'))
        _pds.setColor(_ds.getValue('DIM_PRIMARY_FONT_COLOR'))
        #
        # secondary dimension string
        #
        _sds = self.getSecondaryDimstring()
        _sds.setPrefix(_ds.getValue('DIM_SECONDARY_PREFIX'))
        _sds.setSuffix(_ds.getValue('DIM_SECONDARY_SUFFIX'))
        _sds.setPrecision(_ds.getValue('DIM_SECONDARY_PRECISION'))
        _sds.setUnits(_ds.getValue('DIM_SECONDARY_UNITS'))
        _sds.setPrintZero(_ds.getValue('DIM_SECONDARY_LEADING_ZERO'))
        _sds.setPrintDecimal(_ds.getValue('DIM_SECONDARY_TRAILING_DECIMAL'))
        _sds.setFamily(_ds.getValue('DIM_SECONDARY_FONT_FAMILY'))
        _sds.setSize(_ds.getValue('DIM_SECONDARY_FONT_SIZE'))
        _sds.setWeight(_ds.getValue('DIM_SECONDARY_FONT_WEIGHT'))
        _sds.setStyle(_ds.getValue('DIM_SECONDARY_FONT_STYLE'))
        _sds.setColor(_ds.getValue('DIM_SECONDARY_FONT_COLOR'))
        
    def getEndpointType(self):
        """Return what type of endpoints the Dimension uses.

getEndpointType()
        """
        return self.__endpoint_type

    def setEndpointType(self, e):
        """Set what type of endpoints the Dimension will use.

setEndpointType(e)

The argument "e" should be one of the following

dimension.NO_ENDPOINT => no special marking at the dimension crossbar ends
dimension.ARROW => an arrowhead at the dimension crossbar ends
dimension.FILLED_ARROW => a filled arrohead at the dimension crossbar ends
dimension.SLASH => a slash mark at the dimension crossbar ends
dimension.CIRCLE => a filled circle at the dimension crossbar ends
        """
        if (e != Dimension.NO_ENDPOINT and
            e != Dimension.ARROW and
            e != Dimension.FILLED_ARROW and
            e != Dimension.SLASH and
            e != Dimension.CIRCLE):
            raise ValueError, "Invalid endpoint value: '%s'" % str(e)
        self.__endpoint_type = e

    endpoint = property(getEndpointType, setEndpointType,
                        None, "Dimension endpoint type.")

    def getEndpointSize(self):
        """Return the size of the Dimension endpoints.

getEndpointSize()        
        """
        return self.__endpoint_size

    def setEndpointSize(self, size):
        """Set the size of the Dimension endpoints.

setEndpointSize(size)

The argument size should be a float greater than or
equal to 0.0
        """
        _size = size
        if not isinstance(_size, float):
            _size = float(size)
        if _size < 0.0:
            raise ValueError, "Invalid endpoint size: %g" % _size
        self.__endpoint_size = _size
        
    def getPrimaryDimstring(self):
        """ Return the DimString used for formatting the primary dimension.

getPrimaryDimstring()
        """
        return self.__dimstring1

    def getSecondaryDimstring(self):
        """Return the DimString used for formatting the secondary dimension.

getSecondaryDimstring()
        """
        return self.__dimstring2

    def getDualDimMode(self):
        """Return if the Dimension is displaying primary and secondary values.

getDualDimMode(self)        
        """
        return self.__dual_mode

    def setDualDimMode(self, mode):
        """Set the Dimension to display both primary and secondary values.

setDualDimMode(mode)

The argument "mode" should be True or False.
        """
        if mode is True or mode is False:
            self.__dual_mode = mode
        else:
            raise ValueError, "Invalid dual dimension mode: '%s'" % str(mode)

    dual_mode = property(getDualDimMode, setDualDimMode, None,
                         "Display both primary and secondary dimensions")
    
    def getOffset(self):
        """Return the current offset value for the Dimension.

getOffset()        
        """
        return self.__offset

    def setOffset(self, offset=0.0):
        """Set the offset value for the Dimension.

setOffset([offset])

Calling this method without arguments sets the value to 0.
If the argument "offset" is supplied, it should be a positive
float value.
        """
        _o = offset
        if not isinstance(_o, float):
            _o = float(offset)
        if _o < 0.0:
            raise ValueError, "Invalid dimension offset length: %g" % _o
        self.__offset = _o

    offset = property(getOffset, setOffset, None, "Dimension offset.")

    def getExtension(self):
        """Get the extension length of the Dimension.

getExtension()        
        """
        return self.__extension

    def setExtension(self, ext=0.0):
        """Set the extension length of the Dimension.

setExtension([ext])

If this method is called without arguments, the extension
length is set to 0. If the argument "ext" is supplied, it
should be a positive float value.
        """
        _e = ext
        if not isinstance(_e, float):
            _e = float(ext)
        if _e < 0.0:
            raise ValueError, "Invalid dimension extension length: %g" % _e
        self.__extension = _e

    extension = property(getExtension, setExtension, None,
                         "Dimension extension length.")
    
    def getPosition(self):
        """Return how the dimension text intersects the crossbar.

getPosition()        
        """
        return self.__dim_pos

    def setPosition(self, pos):
        """Set where the dimension text should be placed at the crossbar.

setPosition(pos)

The choices are:

dimension.SPLIT => In the middle of the crossbar.
dimension.ABOVE => Beyond the crossbar from the dimensioned objects.
dimension.BELOW => Between the crossbar and the dimensioned objects.
        """
        if (pos != Dimension.SPLIT and
            pos != Dimension.ABOVE and
            pos != Dimension.BELOW):
            raise ValueError, "Invalid dimension text position: '%s'" % str(pos)
        self.__dim_pos = pos

    position = property(getPosition, setPosition, None,
                        "Dimension text position")
    
    def getColor(self):
        """Return the color of the dimension lines.

getColor()        
        """
        return self.__color

    def setColor(self, c):
        """Set the color of the dimension lines.

setColor(c)
        """
        _c = c
        if not isinstance(_c, color.Color):
            _c = color.Color(c)
        self.__color = _c

    color = property(getColor, setColor, None, "Dimension Color")

    def getLocation(self):
        """Return the location of the dimensional text values.

getLocation()        
        """
        return (self.__dim_loc_x, self.__dim_loc_y)

    def setLocation(self, x, y):
        """Set the location of the dimensional text values.

setLocation(x, y)

The "x" and "y" arguments should be float values. The text is
centered around that point.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        self.__dim_loc_x = _x
        self.__dim_loc_y = _y

    location = property(getLocation, setLocation, None,
                        "Dimension location")

    def getDimensions(self, dimlen):
        """Return the formatted dimensional values.

getDimensions(dimlen)

The argument "dimlen" should be the length in millimeters.

This method returns a list of the primary and secondary
dimensional values.
        """
        _dl = dimlen
        if not isinstance(dimlen, float):
            _dl = float(dimlen)
        dims = []
        dims.append(self.__dimstring1.formatDimension(_dl))
        dims.append(self.__dimstring2.formatDimension(_dl))
        return dims

    def isModified(self):
        """Return if the dimension has adjusted itself.

isModified()        
        """
        return self.__modified
    
    def modified(self):
        """Set the dimension to be in the modified state.

modified()        
        """
        self.__modified = True

    def reset(self):
        """Set the dimension to be in the unmodified state.

reset()        
        """
        self.__modified = False

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a Dimension exists with a region.

isRegion(xmin, ymin, xmax, ymax[, fully])

The first four arguments define the boundary. The optional
fifth argument "fully" indicates whether or not the Dimension
must be completely contained within the region or just pass
through it.

This method should be overriden in classes derived from Dimension.
        """
        return False
    
class LinearDimension(Dimension):
    """A class for Linear dimensions.

The LinearDimension class is derived from the Dimension
class, so it shares all of those methods and attributes.
A LinearDimension should be used to display the absolute
distance between two Point objects.

A LinearDimension object has the following methods:

{get/set}P1(): Get/Set the first Point for the LinearDimension.
{get/set}P2(): Get/Set the second Point for the LinearDimension.
getDimPoints(): Return the two Points used in this dimension.
getDimLayers(): Return the two Layers holding the Points.
getDimXPoints(): Get the x-coordinates of the dimension bar positions.
getDimYPoints(): Get the y-coordinates of the dimension bar positions.
getDimMarkerPoints(): Get the locaiton of the dimension endpoint markers.
    """
    def __init__(self, l1, p1, l2, p2, dx, dy, ds):
        """Instantiate a LinearDimension object.

ldim = LinearDimension(l1, p1, l2, p2, dx, dy, ds)

l1: A Layer holding Point p1
p1: A Point contained in Layer l1
l2: A Layer holding Point p2
p2: A Point contained in Layer l2
dx: The "x" coordinate of the dimensional text
dy: The "y" coordinate of the dimensional text
ds: The DimStyle used for this Dimension.
        """
        if not isinstance(l1, layer.Layer):
            raise TypeError, "Invalid Layer: " + `l1`
        if not isinstance(p1, point.Point):
            raise TypeError, "Invalid Point: " + `p1`
        _lp = l1.findObject(p1)
        if _lp is not p1:
            raise ValueError, "Point p1 not found in layer"
        if not isinstance(l2, layer.Layer):
            raise TypeError, "Invalid Layer: " + `l2`
        if not isinstance(p2, point.Point):
            raise TypeError, "Invalid Point: " + `p2`
        _lp = l2.findObject(p2)
        if _lp is not p2:
            raise ValueError, "Point p2 not found in layer"
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        if not isinstance(ds, DimStyle):
            raise TypeError, "Invalid DimStyle: " + `ds`
        Dimension.__init__(self, ds)
        self.__l1 = l1
        self.__p1 = p1
        self.__l2 = l2
        self.__p2 = p2
        self.__dim_xpts = array.array('d', [0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        self.__dim_ypts = array.array('d', [0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        self.__dim_mpts = array.array('d', [0.0, 0.0, 0.0, 0.0])
        Dimension.setLocation(self, _dx, _dy)
        self.applyDimStyle()

    def __eq__(self, ldim):
        """Test two LinearDimension objects for equality.
        """
        if not isinstance(ldim, LinearDimension):
            raise TypeError, "Invalid object for comparison: " + `ldim`
        _val = False
        _l1, _l2 = ldim.getDimLayers()
        _p1, _p2 = ldim.getDimPoints()
        if (self.__l1 is _l1 and
            self.__l2 is _l2 and
            self.__p1 == _p1 and
            self.__p2 == _p2):
            _val = True
        elif (self.__l1 is _l2 and
              self.__l2 is _l1 and
              self.__p1 == _p2 and
              self.__p2 == _p1):
            _val = True
        return _val

    def __ne__(self, ldim):
        """Test two LinearDimension objects for equality.
        """
        if not isinstance(ldim, LinearDimension):
            raise TypeError, "Invalid object for comparison: " + `ldim`
        _val = True
        _l1, _l2 = ldim.getDimLayers()
        _p1, _p2 = ldim.getDimPoints()
        if (self.__l1 is _l1 and
            self.__l2 is _l2 and
            self.__p1 == _p1 and
            self.__p2 == _p2):
            _val = True
        elif (self.__l1 is _l2 and
              self.__l2 is _l1 and
              self.__p1 == _p2 and
              self.__p2 == _p1):
            _val = False
        return _val

    def modified(self):
        """Indicate that the LinearDimension has been modified.

modified()

This method extends the Dimension::modified() method.
        """
        Dimension.modified(self)
        self.calcDimValues()
        
    def getP1(self):
        """Return the first Point of a LinearDimension.

getP1()        
        """
        return self.__p1

    def setP1(self, l, p):
        """Set the first Point of a LinearDimension.

setP1(l, p)

There are two required arguments for this method

l: The layer holding the Point p
p: A point in Layer l
        """
        if not isinstance(l, layer.Layer):
            raise TypeError, "Invalid layer: " + `l`
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _lp = l.findObject(p)
        if _lp is not p:
            raise ValueError, "Point not found in layer!"
        self.modified()
        self.__l1 = l
        self.__p1 = p

    p1 = property(getP1, None, None, "Dimension first point.")

    def getP2(self):
        """Return the second point of a LinearDimension.

getP2()        
        """
        return self.__p2

    def setP2(self, l, p):
        """Set the second Point of a LinearDimension.

setP2(l, p)

There are two required arguments for this method

l: The layer holding the Point p
p: A point in Layer l
        """
        if not isinstance(l, layer.Layer):
            raise TypeError, "Invalid layer: " + `l`
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _lp = l.findObject(p)
        if _lp is not p:
            raise ValueError, "Point not found in layer!"
        self.modified()
        self.__l2 = l
        self.__p2 = p
        
    p2 = property(getP2, None, None, "Dimension second point.")
    
    def getDimPoints(self):
        """Return both points used in the LinearDimension.

getDimPoints()

The two points are returned in a tuple.
        """
        return (self.__p1, self.__p2)

    def getDimXPoints(self):
        """Get the array of dimension bar x-coordinates.

getDimXPoints()
        """
        return self.__dim_xpts

    def getDimYPoints(self):
        """Get the array of dimension bar y-coordinates.

getDimYPoints()
        """
        return self.__dim_ypts

    def getDimMarkerPoints(self):
        """Get the array of dimension endpoint points.

getDimMarkerPoints()
        """
        return self.__dim_mpts
    
    def getDimLayers(self):
        """Return both layers used in the LinearDimension.

getDimLayers()

The two layers are returned in a tuple.
        """
        return (self.__l1, self.__l2)

    def setLocation(self, x, y):
        """Set the location of the dimensional text values.

setLocation(x, y)

This method extends the Dimension::setLocation() method.
        """
        Dimension.setLocation(self, x, y)
        self.calcDimValues()
        
    def calculate(self):
        """Determine the length of this LinearDimension.

calculate()        
        """
        return self.__p1 - self.__p2

    def applyDimStyle(self):
        """Apply the values in the DimStyle to the LinearDimension.

applyDimStyle()

This method extends the Dimension::applyDimStyle() method.
        """
        Dimension.applyDimStyle(self)
        self.calcDimValues()

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a LinearDimension exists within a region.

isRegion(xmin, ymin, xmax, ymax[, fully])

The four arguments define the boundary of an area, and the
function returns True if the LinearDimension lies within that
area. If the optional argument "fully" is used and is True,
then the dimension points and the location of the dimension
text must lie within the boundary. Otherwise, the function
returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _dx, _dy = self.getLocation()
        _dxpts = self.__dim_xpts
        _dypts = self.__dim_ypts
        _pxmin = min(_x1, _x2, _dx, min(_dxpts))
        _pymin = min(_y1, _y2, _dy, min(_dypts))
        _pxmax = max(_x1, _x2, _dx, max(_dxpts))
        _pymax = max(_y1, _y2, _dy, max(_dypts))
        _val = False
        if (_xmin < _pxmin and
            _xmax > _pxmax and
            _ymin < _pymin and
            _ymax > _pymax):
            _val = True
        elif (not fully and
              not (_pxmax < _xmin or _pymax < _ymin or
                   _pxmin > _xmax or _pymin > _ymax)):
            if _xmin < _x1 < _xmax and _ymin < _y1 < _ymax: # p1 point
                _val = True
            elif _xmin < _x2 < _xmax and _ymin < _y2 < _ymax: # p2 point
                _val = True
            elif _xmin < _dx < _xmax and _ymin < _dy < _ymax: # dim text
                _val = True
            else:
                #
                # bar at p1
                #
                _t1 = (_dxpts[0], _dypts[0])
                _t2 = (_dxpts[2], _dypts[2])
                _val =_test_dim_line(_t1, _t2, _xmin, _ymin, _xmax, _ymax)
                if _val:
                    return _val 
                #
                # bar at p2
                #
                _t1 = (_dxpts[3], _dypts[3])
                _t2 = (_dxpts[3], _dypts[3])
                _val =_test_dim_line(_t1, _t2, _xmin, _ymin, _xmax, _ymax)
                if _val:
                    return _val
                #
                # crossbar
                #
                _t1 = (_dxpts[1], _dypts[1])
                _t2 = (_dxpts[4], _dypts[4])
                _val =_test_dim_line(_t1, _t2, _xmin, _ymin, _xmax, _ymax)
        return _val

    def calcDimValues(self):
        """Recalculate the values for dimensional display.

calcDimValues()

This method calculates where the points for the dimension
bars and crossbar are located.
        """
        _p1, _p2 = self.getDimPoints()
        _p1x, _p1y = _p1.getCoords()
        _p2x, _p2y = _p2.getCoords()
        _dx, _dy = self.getLocation()
        _offset = self.getOffset()
        _ext = self.getExtension()
        #
        # see comp.graphics.algorithms.faq section on calcuating
        # the distance between a point and line for info about
        # the following equations ...
        #
        _dpx = _p2x - _p1x
        _dpy = _p2y - _p1y
        _rnum = ((_dx - _p1x)*_dpx) + ((_dy - _p1y)*_dpy)
        _snum = ((_p1y - _dy)*_dpx) - ((_p1x - _dx)*_dpy)
        _den = pow(_dpx, 2) + pow(_dpy, 2)
        _r = _rnum/_den
        _s = _snum/_den
        _sep = abs(_s) * sqrt(_den)
        _slope = atan2(_dpy, _dpx)
        if _s < 0.0: # dim point left of p1-p2 line
            _angle = _slope + (pi/2.0)
        else: # dim point right of p1-p2 line (or on it)
            _angle = _slope - (pi/2.0)
        _sin_angle = sin(_angle)
        _cos_angle = cos(_angle)
        _dim_xpts = self.__dim_xpts
        _dim_ypts = self.__dim_ypts
        _dim_mpts = self.__dim_mpts
        _dim_xpts[0] = _p1x + (_offset * _cos_angle)
        _dim_ypts[0] = _p1y + (_offset * _sin_angle)
        if _r < 0.0:
            _px = _p1x + (_r * _dpx)
            _py = _p1y + (_r * _dpy)
            _dim_xpts[1] = _px + (_sep * _cos_angle)
            _dim_ypts[1] = _py + (_sep * _sin_angle)
        else:
            _dim_xpts[1] = _p1x + (_sep * _cos_angle)
            _dim_ypts[1] = _p1y + (_sep * _sin_angle)
        _dim_mpts[0] = _p1x + (_sep * _cos_angle)
        _dim_mpts[1] = _p1y + (_sep * _sin_angle)
        _dim_xpts[2] = _p1x + ((_sep + _ext) * _cos_angle)
        _dim_ypts[2] = _p1y + ((_sep + _ext) * _sin_angle)
        _dim_xpts[3] = _p2x + (_offset * _cos_angle)
        _dim_ypts[3] = _p2y + (_offset * _sin_angle)
        if _r > 1.0:
            _px = _p1x + (_r * _dpx)
            _py = _p1y + (_r * _dpy)
            _dim_xpts[4] = _px + (_sep * _cos_angle)
            _dim_ypts[4] = _py + (_sep * _sin_angle)
        else:
            _dim_xpts[4] = _p2x + (_sep * _cos_angle)
            _dim_ypts[4] = _p2y + (_sep * _sin_angle)
        _dim_mpts[2] = _p2x + (_sep * _cos_angle)
        _dim_mpts[3] = _p2y + (_sep * _sin_angle)
        _dim_xpts[5] = _p2x + ((_sep + _ext) * _cos_angle)
        _dim_ypts[5] = _p2y + ((_sep + _ext) * _sin_angle)
        
    def onDimension(self, x, y, tol=tolerance.TOL):
        """Test an x/y coordinate pair if it could lay on the dimension.

onDimension(x, y[, tol])

This method has two required parameters:

x: The x-coordinate
y: The y-coordinate

These should both be float values.

There is an optional third parameter, "tol", giving
the maximum distance from the dimension bars that the
x/y coordinates may sit.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _dxpts = self.__dim_xpts
        _dypts = self.__dim_ypts
        #
        # test p1 bar
        #
        _val = _test_on_dimension(_dxpts[0], _dypts[0],
                                  _dxpts[2], _dypts[2],
                                  _x, _y, _t)
        if _val:
            return _val
        #
        # test p2 bar
        #
        _val = _test_on_dimension(_dxpts[3], _dypts[3],
                                  _dxpts[5], _dypts[5],
                                  _x, _y, _t)
        if _val:
            return _val
        #
        # test crossbar
        #
        return  _test_on_dimension(_dxpts[1], _dypts[1],
                                   _dxpts[4], _dypts[4],
                                   _x, _y, _t)
        
        
class HorizontalDimension(LinearDimension):
    """A class representing Horizontal dimensions.

This class is derived from the LinearDimension class, so
it shares all those attributes and methods of its parent.
    """
    def __init__(self, l1, p1, l2, p2, dx, dy, ds):
        """Instantiate a Horizontal Dimension.

hdim = HorizontalDimension(l1, p1, l2, p2, dx, dy, ds)

l1: A Layer holding Point p1
p1: A Point contained in Layer l1
l2: A Layer holding Point p2
p2: A Point contained in Layer l2
dx: The "x" coordinate of the dimensional text
dy: The "y" coordinate of the dimensional text
ds: The DimStyle used for this Dimension.
        """        
        LinearDimension.__init__(self, l1, p1, l2, p2, dx, dy, ds)
        self.calcDimValues()

    def calculate(self):
        """Determine the length of this HorizontalDimension.

calculate()        
        """
        _p1, _p2 = self.getDimPoints()
        return abs(_p1.x - _p2.x)

    def calcDimValues(self):
        """Recalculate the values for dimensional display.

calcDimValues()

This method overrides the LinearDimension::calcDimValues() method.
        """
        _p1, _p2 = self.getDimPoints()
        _p1x, _p1y = _p1.getCoords()
        _p2x, _p2y = _p2.getCoords()
        _pxmin = min(_p1x, _p2x)
        _pxmax = max(_p1x, _p2x)
        _pymin = min(_p1y, _p2y)
        _pymax = max(_p1y, _p2y)            
        _dx, _dy = self.getLocation()
        _offset = self.getOffset()
        _ext = self.getExtension()
        _dim_xpts = self.getDimXPoints()
        _dim_xpts[0] = _p1x
        _dim_xpts[1] = _p1x
        _dim_xpts[2] = _p1x
        _dim_xpts[3] = _p2x
        _dim_xpts[4] = _p2x
        _dim_xpts[5] = _p2x
        if _dx < _pxmin or _dx > _pxmax:
            if _p1x < _p2x:
                if _dx < _p1x:
                    _dim_xpts[1] = _dx
                if _dx > _p2x:
                    _dim_xpts[4] = _dx
            if _p2x < _p1x:
                if _dx < _p2x:
                    _dim_xpts[4] = _dx
                if _dx > _p1x:
                    _dim_xpts[1] = _dx
        _dim_ypts = self.getDimYPoints()
        _dim_ypts[1] = _dy
        _dim_ypts[4] = _dy
        if _dy < _pymin:
            _dim_ypts[0] = _p1y - _offset
            _dim_ypts[2] = _dy - _ext
            _dim_ypts[3] = _p2y - _offset
            _dim_ypts[5] = _dy - _offset
        elif _dy > _pymax:
            _dim_ypts[0] = _p1y + _offset
            _dim_ypts[2] = _dy + _ext
            _dim_ypts[3] = _p2y + _offset
            _dim_ypts[5] = _dy + _ext
        else:
            if _dy > _p1y:
                _dim_ypts[0] = _p1y + _offset
                _dim_ypts[2] = _dy + _offset
            else:
                _dim_ypts[0] = _p1y - _offset
                _dim_ypts[2] = _dy - _offset
            if _dy > _p2y:
                _dim_ypts[3] = _p2y + _offset
                _dim_ypts[5] = _dy + _offset
            else:
                _dim_ypts[3] = _p2y - _offset
                _dim_ypts[5] = _dy - _offset
        _dim_mpts = self.getDimMarkerPoints()
        _dim_mpts[0] = _p1x
        _dim_mpts[1] = _dy
        _dim_mpts[2] = _p2x
        _dim_mpts[3] = _dy

class VerticalDimension(LinearDimension):
    """A class representing Vertical dimensions.

This class is derived from the LinearDimension class, so
it shares all those attributes and methods of its parent.
    """

    def __init__(self, l1, p1, l2, p2, dx, dy, ds):
        """Instantiate a Horizontal Dimension.

vdim = VerticalDimension(l1, p1, l2, p2, dx, dy, ds)

l1: A Layer holding Point p1
p1: A Point contained in Layer l1
l2: A Layer holding Point p2
p2: A Point contained in Layer l2
dx: The "x" coordinate of the dimensional text
dy: The "y" coordinate of the dimensional text
ds: The DimStyle used for this Dimension.
        """        
        LinearDimension.__init__(self, l1, p1, l2, p2, dx, dy, ds)
        self.calcDimValues()
        
    def calculate(self):
        """Determine the length of this VerticalDimension.

calculate()        
        """
        _p1, _p2 = self.getDimPoints()
        return abs(_p1.y - _p2.y)

    def calcDimValues(self):
        """Recalculate the values for dimensional display.

calcDimValues()

This method overrides the LinearDimension::calcDimValues() method.
        """
        _p1, _p2 = self.getDimPoints()
        _p1x, _p1y = _p1.getCoords()
        _p2x, _p2y = _p2.getCoords()
        _pxmin = min(_p1x, _p2x)
        _pxmax = max(_p1x, _p2x)
        _pymin = min(_p1y, _p2y)
        _pymax = max(_p1y, _p2y)
        _dx, _dy = self.getLocation()
        _offset = self.getOffset()
        _ext = self.getExtension()
        _dim_ypts = self.getDimYPoints()
        _dim_ypts[0] = _p1y
        _dim_ypts[1] = _p1y
        _dim_ypts[2] = _p1y
        _dim_ypts[3] = _p2y
        _dim_ypts[4] = _p2y
        _dim_ypts[5] = _p2y
        if _dy < _pymin or _dy > _pymax:
            if _p1y < _p2y:
                if _dy < _p1y:
                    _dim_ypts[1] = _dy
                if _dy > _p2y:
                    _dim_ypts[4] = _dy
            if _p2y < _p1y:
                if _dy < _p2y:
                    _dim_ypts[4] = _dy
                if _dy > _p1y:
                    _dim_ypts[1] = _dy
        _dim_xpts = self.getDimXPoints()
        _dim_xpts[1] = _dx
        _dim_xpts[4] = _dx
        if _dx < _pxmin:
            _dim_xpts[0] = _p1x - _offset
            _dim_xpts[2] = _dx - _ext
            _dim_xpts[3] = _p2x - _offset
            _dim_xpts[5] = _dx - _offset
        elif _dx > _pxmax:
            _dim_xpts[0] = _p1x + _offset
            _dim_xpts[2] = _dx + _ext
            _dim_xpts[3] = _p2x + _offset
            _dim_xpts[5] = _dx + _ext
        else:
            if _dx > _p1x:
                _dim_xpts[0] = _p1x + _offset
                _dim_xpts[2] = _dx + _offset
            else:
                _dim_xpts[0] = _p1x - _offset
                _dim_xpts[2] = _dx - _offset
            if _dx > _p2x:
                _dim_xpts[3] = _p2x + _offset
                _dim_xpts[5] = _dx + _offset
            else:
                _dim_xpts[3] = _p2x - _offset
                _dim_xpts[5] = _dx - _offset
        _dim_mpts = self.getDimMarkerPoints()
        _dim_mpts[0] = _dx
        _dim_mpts[1] = _p1y
        _dim_mpts[2] = _dx
        _dim_mpts[3] = _p2y
        
class RadialDimension(Dimension):
    """A class for Radial dimensions.

The RadialDimension class is derived from the Dimension
class, so it shares all of those methods and attributes.
A RadialDimension should be used to display either the
radius or diamter of a Circle object.

A RadialDimension object has the following methods:

{get/set}DimCircle(): Get/Set the measured circle object.
getDimLayer(): Return the layer containing the measured circle.
{get/set}DiaMode(): Get/Set if the RadialDimension should return diameters.
getDimXPoints(): Get the x-coordinates of the dimension bar positions.
getDimYPoints(): Get the y-coordinates of the dimension bar positions.
getDimMarkerPoints(): Get the locaiton of the dimension endpoint markers.
calcDimValues(): Calculate the endpoint of the dimension line.
onDimension(): Test if an x/y coordinate pair fall on the dimension line.
    """

    def __init__(self, cl, cir, dx, dy, ds):
        """Initialize a RadialDimension object.
        
rdim = RadialDimension(cl, cir, dx, dy, ds)

cl: A Layer object
cir: A Circle or Arc object
dx: The "x" coordinate of the dimensional text
dy: The "y" coordinate of the dimensional text
ds: The DimStyle used for this Dimension.
        """
        if not isinstance(cl, layer.Layer):
            raise TypeError, "Invalid Layer: " + `cl`
        if not isinstance(cir, circle.Circle):
            raise TypeError, "Invalid Circular object: " + `cir`
        _lc = cl.findObject(cir)
        if _lc is not cir:
            raise ValueError, "Circle not found in layer"
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        if not isinstance(ds, DimStyle):
            raise TypeError, "Invalid DimStyle: " + `ds`
        Dimension.__init__(self, ds)
        self.__layer = cl
        self.__circle = cir
        self.__dia_mode = False
        self.__dim_xpts = array.array('d', [0.0, 0.0])
        self.__dim_ypts = array.array('d', [0.0, 0.0])
        self.__dim_mpts = array.array('d', [0.0, 0.0, 0.0, 0.0])
        Dimension.setLocation(self, _dx, _dy)
        self.applyDimStyle()
        
    def __eq__(self, rdim):
        """Compare two RadialDimensions for equality.
        """
        if not isinstance(rdim, RadialDimension):
            raise TypeError, "Invalid object for comparison: " + `rdim`
        _val = False
        _rc = rdim.getDimCircle()
        _rl = rdim.getDimLayer()
        if self.__layer is _rl and self.__circle == _rc:
            _val = True
        return _val

    def __ne__(self, rdim):
        """Compare two RadialDimensions for inequality.
        """
        if not isinstance(rdim, RadialDimension):
            raise TypeError, "Invalid object for comparison: " + `rdim`
        _val = True
        _rc = rdim.getDimCircle()
        _rl = rdim.getDimLayer()
        if self.__layer is _rl and self.__circle == _rc:
            _val = False
        return _val

    def modified(self):
        """Indicate that the RadialDimension has been modified.

modified()

This method extends the Dimension::modified() method.
        """
        Dimension.modified(self)
        self.calcDimValues()
        
    def applyDimStyle(self):
        """Apply the values in the DimStyle to the RadialDimension.

applyDimStyle()

This method extends the Dimension::applyDimStyle() method.
        """
        Dimension.applyDimStyle(self)
        _ds = self.getDimStyle()
        _pds = self.getPrimaryDimstring()
        _pds.setPrefix(_ds.getValue('RADIAL_DIM_PRIMARY_PREFIX'))
        _pds.setSuffix(_ds.getValue('RADIAL_DIM_PRIMARY_SUFFIX'))
        _sds = self.getSecondaryDimstring()
        _sds.setPrefix(_ds.getValue('RADIAL_DIM_SECONDARY_PREFIX'))
        _sds.setSuffix(_ds.getValue('RADIAL_DIM_SECONDARY_SUFFIX'))
        self.setDiaMode(_ds.getValue('RADIAL_DIM_DIA_MODE'))
        self.calcDimValues()

    def getDiaMode(self):
        """Return if the RadialDimension will return diametrical values.

getDiaMode()

This method returns True if the diameter value is returned,
and False otherwise.
        """
        return self.__dia_mode
    
    def setDiaMode(self, mode=False):
        """Set the RadialDimension to return diametrical values.

setDiaMode([mode])

Calling this method without an argument sets the RadialDimension
to return radial measurements. If the argument "mode" is supplied,
it should be either True or False.

If the RadialDimension is measuring an arc, the returned value
will always be set to return a radius.
        """
        if mode is True or mode is False:
            if not isinstance(self.__circle, arc.Arc):
                _old_mode = self.__dia_mode
                self.__dia_mode = mode
                if mode is not _old_mode:
                    self.calcDimValues()
        else:
            raise ValueError, "Invalid diameter mode: " + `mode`

    dia_mode = property(getDiaMode, setDiaMode, None,
                        "Draw the Dimension as a diameter")

    def getDimLayer(self):
        """Return the Layer object holding the Circle for this RadialDimension.

getDimLayer()        
        """
        return self.__layer

    def getDimCircle(self):
        """Return the Circle object this RadialDimension is measuring.

getDimCircle()        
        """
        return self.__circle

    def setDimCircle(self, l, c):
        """Set the Circle object measured by this RadialDimension.

setDimCircle(l, c)

The arguments for this method are:

l: A Layer containing circle c
c: A circle contained in layer l
        """
        if not isinstance(l, layer.Layer):
            raise TypeError, "Invalid layer: " + `l`
        if not isinstance(c, circle.Circle):
            raise TypeError, "Invalid circle: " + `c`
        _lc = l.findObject(c)
        if _lc is not c:
            raise ValueError, "Circle not found in layer!"
        self.modified()
        self.__layer = l
        self.__circle = c
        
    circle = property(getDimCircle, None, None,
                      "Radial dimension circle object.")

    def getDimXPoints(self):
        """Get the array of dimension bar x-coordinates.

getDimXPoints()
        """
        return self.__dim_xpts

    def getDimYPoints(self):
        """Get the array of dimension bar y-coordinates.

getDimYPoints()
        """
        return self.__dim_ypts

    def getDimMarkerPoints(self):
        """Get the array of dimension endpoint points.

getDimMPoints()
        """
        return self.__dim_mpts
    
    def setLocation(self, x, y):
        """Set the location of the dimensional text values.

setLocation(x, y)

This method extends the Dimension::setLocation() method.
        """
        Dimension.setLocation(self, x, y)
        self.calcDimValues()
        
    def calcDimValues(self):
        """Recalculate the values for dimensional display.

calcDimValues()
        """
        _c = self.__circle
        _cx, _cy = _c.getCenter().getCoords()
        _dx, _dy = self.getLocation()
        _rad = _c.getRadius()
        _sep = hypot((_dx - _cx), (_dy - _cy))
        _angle = atan2((_dy - _cy), (_dx - _cx))
        _sx = _rad * cos(_angle)
        _sy = _rad * sin(_angle)
        self.__dim_mpts[0] = _cx + _sx
        self.__dim_mpts[1] = _cy + _sy
        self.__dim_xpts[0] = _cx
        self.__dim_ypts[0] = _cy
        if isinstance(_c, arc.Arc):
            _sa = _c.getStartAngle()
            _ea = _c.getEndAngle()
            _angle = (180.0/pi) * _angle
            if _angle < 0.0:
                _angle = _angle + 360.0
            if not _c.throughAngle(_angle):
                _ep1, _ep2 = _c.getEndpoints()
                if _angle < _sa:
                    _sa = (pi/180.0) * _sa
                    _sx = _rad * cos(_sa)
                    _sy = _rad * sin(_sa)
                    if _sep > _rad:
                        _dx = _cx + _sep * cos(_sa)
                        _dy = _cy + _sep * sin(_sa)
                if _angle > _ea:
                    _ea = (pi/180.0) * _ea
                    _sx = _rad * cos(_ea)
                    _sy = _rad * sin(_ea)
                    if _sep > _rad:
                        _dx = _cx + _sep * cos(_ea)
                        _dy = _cy + _sep * sin(_ea)
        if _sep > _rad:
            self.__dim_xpts[1] = _dx
            self.__dim_ypts[1] = _dy
        else:
            self.__dim_xpts[1] = _cx + _sx
            self.__dim_ypts[1] = _cy + _sy
        if not isinstance(_c, arc.Arc) and self.__dia_mode:
            _x = _cx - _sx
            _y = _cy - _sy
            self.__dim_xpts[0] = _x
            self.__dim_ypts[0] = _y
            self.__dim_mpts[2] = _x
            self.__dim_mpts[3] = _y

    def calculate(self):
        """Return the radius or diamter of this RadialDimension.

calculate()

By default, a RadialDimension will return the radius of the
circle. The setDiaMode() method can be called to set the
returned value to corresponed to a diameter.
        """
        _val = self.__circle.getRadius()
        if self.__dia_mode is True:
            _val = _val * 2.0
        return _val

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a RadialDimension exists within a region.

isRegion(xmin, ymin, xmax, ymax[, fully])

The four arguments define the boundary of an area, and the
function returns True if the RadialDimension lies within that
area. If the optional argument "fully" is used and is True,
then the dimensioned circle and the location of the dimension
text must lie within the boundary. Otherwise, the function
returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        _circ = self.__circle
        _cx, _cy = _circ.getCenter().getCoords()
        _rad = _circ.getRadius()
        _dx, _dy = self.getLocation()        
        _cxmin = min(_cx - _rad, _dx)
        _cymin = min(_cy - _rad, _dy)
        _cxmax = max(_cx + _rad, _dx)
        _cymax = max(_cy + _rad, _dy)
        _val = False
        if (_xmin < _cxmin and _xmax > _cxmax and
            _ymin < _cymin and _ymax > _cymax):
            _val = True
        elif (not fully and
              not (_cxmax < _xmin or
                   _cymax < _ymin or
                   _cxmin > _xmax or
                   _cymin > _ymax)):
            if _xmin < _dx < _xmax and _ymin < _dy < _ymax: # dim text
                _val = True
            elif _xmin < _cx < _xmax and _ymin < _cy < _ymax: # circle center
                _val = True
            else: # dimension line
                _dxpts = self.__dim_xpts
                _dypts = self.__dim_ypts
                _t1 = (_dxpts[0], _dypts[0])
                _t2 = (_dxpts[1], _dypts[1])
                _val =_test_dim_line(_t1, _t2, _xmin, _ymin, _xmax, _ymax)
        return _val

    def onDimension(self, x, y, tol=tolerance.TOL):
        """Test an x/y coordinate pair if it could lay on the dimension.

onDimension(x, y[, tol])

This method has two required parameters:

x: The x-coordinate
y: The y-coordinate

These should both be float values.

There is an optional third parameter, "tol", giving
the maximum distance from the dimension bars that the
x/y coordinates may sit.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _dxpts = self.__dim_xpts
        _dypts = self.__dim_ypts
        return _test_on_dimension(_dxpts[0], _dypts[0],
                                  _dxpts[1], _dypts[1],
                                  _x, _y, _t)
        
class AngularDimension(Dimension):
    """A class for Angular dimensions.

The AngularDimension class is derived from the Dimension
class, so it shares all of those methods and attributes.

AngularDimension objects have the following methods:

{get/set}VertexPoint(): Get/Set the vertex point for the AngularDimension.
{get/set}P1(): Get/Set the first Point for the AngularDimension.
{get/set}P2(): Get/Set the second Point for the AngularDimension.
getDimPoints(): Return the two Points used in this dimension.
getDimLayers(): Return the two Layers holding the Points.
getDimXPoints(): Get the x-coordinates of the dimension bar positions.
getDimYPoints(): Get the y-coordinates of the dimension bar positions.
getDimAngles(): Get the angles at which the dimension bars should be drawn.
getDimMarkerPoints(): Get the locaiton of the dimension endpoint markers.
calcDimValues(): Calculate the endpoint of the dimension line.
onDimension(): Test if an x/y coordinate pair fall on the dimension line.
invert(): Switch the endpoints used to measure the dimension
    """
    
    def __init__(self, vl, vp, l1, p1, l2, p2, dx, dy, ds):
        """Instantiate an AngularDimension

adim = AngularDimension(vl, vp, l1, p1, l2, p2, dx, dy, ds)

vl: A Layer holding Point vp
vp: A Point contained in layer vl
l1: A Layer holding Point p1
p1: A Point contained in Layer l1
l2: A Layer holding Point p2
p2: A Point contained in Layer l2
dx: The "x" coordinate of the dimensional text
dy: The "y" coordinate of the dimensional text
ds: The DimStyle used for this Dimension.
        """
        if not isinstance(vl, layer.Layer):
            raise TypeError, "Invalid Layer: " + `vl`
        if not isinstance(vp, point.Point):
            raise TypeError, "Invalid Point: " + `vp`
        _lp = vl.findObject(vp)
        if _lp is not vp:
            raise ValueError, "Vertex point vp not found in layer"
        if not isinstance(l1, layer.Layer):
            raise TypeError, "Invalid Layer: " + `l1`
        if not isinstance(p1, point.Point):
            raise TypeError, "Invalid Point: " + `p1`
        _lp = l1.findObject(p1)
        if _lp is not p1:
            raise ValueError, "Point p1 not found in layer"
        if not isinstance(l2, layer.Layer):
            raise TypeError, "Invalid Layer: " + `l2`
        if not isinstance(p2, point.Point):
            raise TypeError, "Invalid Point: " + `p2`
        _lp = l2.findObject(p2)
        if _lp is not p2:
            raise ValueError, "Point p2 not found in layer"
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        if not isinstance(ds, DimStyle):
            raise TypeError, "Invalid DimStyle: " + `ds`
        Dimension.__init__(self, ds)        
        self.__vl = vl
        self.__vp = vp
        self.__l1 = l1
        self.__p1 = p1
        self.__l2 = l2
        self.__p2 = p2
        self.__dim_xpts = array.array('d', [0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        self.__dim_ypts = array.array('d', [0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        self.__dim_angles = array.array('d', [0.0, 0.0])
        self.__dim_mpts = array.array('d', [0.0, 0.0, 0.0, 0.0])
        self.__dim_radius = 0.0
        Dimension.setLocation(self, _dx, _dy)
        self.applyDimStyle()

    def __eq__(self, adim):
        """Compare two AngularDimensions for equality.
        """
        if not isinstance(adim, AngularDimension):
            raise TypeError, "Invalid object for comparison: " + `adim`
        _val = False
        _vl, _l1, _l2 = adim.getDimLayers()
        _vp, _p1, _p2 = adim.getDimPoints()
        if (self.__vl is _vl and
            self.__vp == _vp and
            self.__l1 is _l1 and
            self.__p1 == _p1 and
            self.__l2 is _l2 and
            self.__p2 == _p2):
            _val = True
        return _val

    def __ne__(self, adim):
        """Compare two AngularDimensions for inequality.
        """
        if not isinstance(adim, AngularDimension):
            raise TypeError, "Invalid object for comparison: " + `adim`
        _val = True
        _vl, _l1, _l2 = adim.getDimLayers()
        _vp, _p1, _p2 = adim.getDimPoints()
        if (self.__vl is _vl and
            self.__vp == _vp and
            self.__l1 is _l1 and
            self.__p1 == _p1 and
            self.__l2 is _l2 and
            self.__p2 == _p2):
            _val = False
        return _val

    def modified(self):
        """Indicate that the AngularDimension has been modified.

modified()

This method extends the Dimension::modified() method.
        """
        Dimension.modified(self)
        self.calcDimValues()
        
    def applyDimStyle(self):
        """Set the values in the AngularDimension's DimStyle.

applyDimStyle()

This method extends the Dimension::applyDimStyle() method.
        """
        Dimension.applyDimStyle(self)
        ds = self.getDimStyle()
        pds = self.getPrimaryDimstring()
        pds.setPrefix(ds.getValue('ANGULAR_DIM_PRIMARY_PREFIX'))
        pds.setSuffix(ds.getValue('ANGULAR_DIM_PRIMARY_SUFFIX'))
        sds = self.getSecondaryDimstring()
        sds.setPrefix(ds.getValue('ANGULAR_DIM_SECONDARY_PREFIX'))
        sds.setSuffix(ds.getValue('ANGULAR_DIM_SECONDARY_SUFFIX'))
        self.calcDimValues()
        
    def getDimLayers(self):
        """Return the layers used in an AngularDimension.

getDimLayers()        
        """
        return (self.__vl, self.__l1, self.__l2)

    def getDimPoints(self):
        """Return the points used in an AngularDimension.

getDimPoints()        
        """
        return (self.__vp, self.__p1, self.__p2)

    def getVertexPoint(self):
        """Return the vertex point used in an AngularDimension.

getVertexPoint()
        """
        return self.__vp

    def setVertexPoint(self, l, p):
        """Set the vertex point for an AngularDimension.

setVertexPoint(l, p)

There are two required arguments for this method

l: The layer holding the Point p
p: A point in Layer l
        """
        if not isinstance(l, layer.Layer):
            raise TypeError, "Invalid layer: " + `l`
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _lp = l.findObject(p)
        if _lp is not p:
            raise ValueError, "Point not found in layer!"
        self.modified()
        self.__vl = l
        self.__vp = p

    vp = property(getVertexPoint, None, None,
                  "Angular Dimension vertex point.")

    def getP1(self):
        """Return the first angle point used in an AngularDimension.

getP1()        
        """
        return self.__p1
    
    def setP1(self, l, p):
        """Set the first Point for an AngularDimension.

setP1(l, p)

There are two required arguments for this method

l: The layer holding the Point p
p: A point in Layer l
        """
        if not isinstance(l, layer.Layer):
            raise TypeError, "Invalid layer: " + `l`
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _lp = l.findObject(p)
        if _lp is not p:
            raise ValueError, "Point not found in layer!"
        self.modified()
        self.__l1 = l
        self.__p1 = p
        
    p1 = property(getP1, None, None, "Dimension first point.")

    def getP2(self):
        """Return the second angle point used in an AngularDimension.

getP2()        
        """
        return self.__p2

    def setP2(self, l, p):
        """Set the second Point for an AngularDimension.

setP2(l, p)

There are two required arguments for this method

l: The layer holding the Point p
p: A point in Layer l
        """
        if not isinstance(l, layer.Layer):
            raise TypeError, "Invalid layer: " + `l`
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid point: " + `p`
        _lp = l.findObject(p)
        if _lp is not p:
            raise ValueError, "Point not found in layer!"
        self.modified()
        self.__l2 = l
        self.__p2 = p

    p2 = property(getP2, None, None, "Dimension second point.")

    def getDimXPoints(self):
        """Get the array of dimension bar x-coordinates.

getDimXPoints()
        """
        return self.__dim_xpts

    def getDimYPoints(self):
        """Get the array of dimension bar y-coordinates.

getDimYPoints()
        """
        return self.__dim_ypts

    def getDimAngles(self):
        """Get the array of dimension bar angles.

geDimAngles()        
        """
        return self.__dim_angles

    def getDimRadius(self):
        """Get the radius of the dimension crossarc.

getDimRadius()        
        """
        return self.__dim_radius

    def getDimMarkerPoints(self):
        """Get the array of dimension endpoint points.

getDimMPoints()
        """
        return self.__dim_mpts
    
    def setLocation(self, x, y):
        """Set the location of the dimensional text values.

setLocation(x, y)

This method extends the Dimension::setLocation() method.
        """
        Dimension.setLocation(self, x, y)
        self.calcDimValues()

    def invert(self):
        """Switch the endpoints used in this object.

invert()

Invoking this method on an AngularDimension will result in
it measuring the opposite angle than what it currently measures.
        """
        _ltemp = self.__l1
        _ptemp = self.__p1
        self.__l1 = self.__l2
        self.__p1 = self.__p2
        self.__l2 = _ltemp
        self.__p2 = _ptemp
        self.calcDimValues()

    def calculate(self):
        """Find the value of the angle measured by this AngularDimension.

calculate()  
        """
        _vx, _vy = self.__vp.getCoords()
        _p1x, _p1y = self.__p1.getCoords()
        _p2x, _p2y = self.__p2.getCoords()
        _a1 = (180.0/pi) * atan2((_p1y - _vy), (_p1x - _vx))
        if _a1 < 0.0:
            _a1 = _a1 + 360.0
        _a2 = (180.0/pi) * atan2((_p2y - _vy), (_p2x - _vx))
        if _a2 < 0.0:
            _a2 = _a2 + 360.0
        _val = _a2 - _a1
        if _a1 > _a2:
            _val = _val + 360.0
        return _val

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not an AngularDimension exists within a region.

isRegion(xmin, ymin, xmax, ymax[, fully])

The four arguments define the boundary of an area, and the
function returns True if the RadialDimension lies within that
area. If the optional argument "fully" is used and is True,
then the dimensioned circle and the location of the dimension
text must lie within the boundary. Otherwise, the function
returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        _vx, _vy = self.__vp.getCoords()
        _p1x, _p1y = self.__p1.getCoords()
        _p2x, _p2y = self.__p2.getCoords()
        _dx, _dy = self.getLocation()
        _dxpts = self.__dim_xpts
        _dypts = self.__dim_ypts
        _pxmin = min(_vx, _p1x, _p2x, _dx, min(_dxpts))
        _pymin = min(_vy, _p1y, _p2y, _dy, min(_dypts))
        _pxmax = max(_vx, _p1x, _p2x, _dx, max(_dxpts))
        _pymax = max(_vy, _p1y, _p2y, _dy, max(_dypts))
        _val = False
        if (_xmin < _pxmin and _xmax > _pxmax and
            _ymin < _pymin and _ymax > _pymax and
            _xmin < _dx < _xmax and _ymin < _dy < _ymax):
            _val = True
        elif not fully:
            if _xmin < _dx < _xmax and _ymin < _dy < _ymax:
                _val = True
            elif _xmin < _vx < _xmax and _ymin < _vy < _ymax:
                _val = True
            elif _xmin < _p1x < _xmax and _ymin < _p1y < _ymax:
                _val = True
            elif _xmin < _p2x < _xmax and _ymin < _p2y < _ymax:
                _val = True
            else:
                #
                # bar on vp-p1 line
                #
                _t1 = (_dxpts[0], _dypts[0])
                _t2 = (_dxpts[2], _dypts[2])
                _val =_test_dim_line(_t1, _t2, _xmin, _ymin, _xmax, _ymax)
                if _val:
                    return _val 
                #
                # bar at vp-p2 line
                #
                _t1 = (_dxpts[3], _dypts[3])
                _t2 = (_dxpts[5], _dypts[5])
                _val =_test_dim_line(_t1, _t2, _xmin, _ymin, _xmax, _ymax)
                if _val:
                    return _val
                #
                # dimension crossarc
                #
                _r = self.__dim_radius
                _d1 = hypot((_xmin - _vx), (_ymin - _vy))
                _d2 = hypot((_xmin - _vx), (_ymax - _vy))
                _d3 = hypot((_xmax - _vx), (_ymax - _vy))
                _d4 = hypot((_xmax - _vx), (_ymin - _vy))
                _dmin = min(_d1, _d2, _d3, _d4)
                _dmax = max(_d1, _d2, _d3, _d4)
                if _xmin < _vx < _xmax and _ymin < _vy < _ymax:
                    _dmin = -1e-10
                else:
                    if _vx > _xmax and _ymin < _vy < _ymax:
                        _dmin = _vx - _xmax
                    elif _vx < _xmin and _ymin < _vy < _ymax:
                        _dmin = _xmin - _vx
                    elif _vy > _ymax and _xmin < _vx < _xmax:
                        _dmin = _vy - _ymax
                    elif _vy < _ymin and _xmin < _vx < _xmax:
                        _dmin = _ymin - _vy
                if _dmin < _r < _dmax:
                    _rtd = 180.0/pi
                    _da = _rtd * atan2((_ymin - _vy), (_xmin - _vx))
                    if _da < 0.0:
                        _da = _da + 360.0
                    _val = self._throughAngle(_da)
                    if _val:
                        return _val
                    _da = _rtd * atan2((_ymin - _vy), (_xmax - _vx))
                    if _da < 0.0:
                        _da = _da + 360.0
                    _val = self._throughAngle(_da)
                    if _val:
                        return _val
                    _da = _rtd * atan2((_ymax - _vy), (_xmax - _vx))
                    if _da < 0.0:
                        _da = _da + 360.0
                    _val = self._throughAngle(_da)
                    if _val:
                        return _val
                    _da = _rtd * atan2((_ymax - _vy), (_xmin - _vx))
                    if _da < 0.0:
                        _da = _da + 360.0
                    _val = self._throughAngle(_da)
        return _val

    def _throughAngle(self, angle):
        """Test if the angular crossarc exists at a certain angle.

_throughAngle()

This method is private to the AngularDimension class.
        """
        _angle1 = self.__dim_angles[0]
        _angle2 = self.__dim_angles[1]
        _val = False
        if _angle2 > _angle1:
            if _angle2 > angle > _angle1:
                _val = True
        else:
            if ((_angle1 < angle < (360.0 + 1e-10)) or
                (-1e-10 < angle < _angle2)):
                _val = True
        return _val

    def calcDimValues(self):
        """Recalculate the values for dimensional display.

calcDimValues()
        """
        _vx, _vy = self.__vp.getCoords()
        _p1x, _p1y = self.__p1.getCoords()
        _p2x, _p2y = self.__p2.getCoords()
        _dx, _dy = self.getLocation()
        _angle1 = atan2((_p1y - _vy), (_p1x - _vx))
        _angle2 = atan2((_p2y - _vy), (_p2x - _vx))
        _dv1 = hypot((_p1x - _vx), (_p1y - _vy))
        _dv2 = hypot((_p2x - _vx), (_p2y - _vy))
        _ddp = hypot((_dx - _vx), (_dy - _vy))
        self.__dim_radius = _ddp
        _offset = self.getOffset()
        _ext = self.getExtension()
        _dim_xpts = self.__dim_xpts
        _dim_ypts = self.__dim_ypts
        _sin_angle = sin(_angle1)
        _cos_angle = cos(_angle1)
        _ex = _vx + (_ddp * _cos_angle)
        _ey = _vy + (_ddp * _sin_angle)
        _dim_xpts[1] = _ex
        _dim_ypts[1] = _ey
        self.__dim_mpts[0] = _ex
        self.__dim_mpts[1] = _ey
        if _ddp > _dv1: # dim point is radially further to vp than p1
            _dim_xpts[0] = _p1x + (_offset * _cos_angle)
            _dim_ypts[0] = _p1y + (_offset * _sin_angle)
            _dim_xpts[2] = _vx + ((_ddp + _ext) * _cos_angle)
            _dim_ypts[2] = _vy + ((_ddp + _ext) * _sin_angle)
        else: # dim point is radially closer to vp than p1
            _dim_xpts[0] = _p1x - (_offset * _cos_angle)
            _dim_ypts[0] = _p1y - (_offset * _sin_angle)
            _dim_xpts[2] = _vx + ((_ddp - _ext) * _cos_angle)
            _dim_ypts[2] = _vy + ((_ddp - _ext) * _sin_angle)
        _sin_angle = sin(_angle2)
        _cos_angle = cos(_angle2)
        _ex = _vx + (_ddp * _cos_angle)
        _ey = _vy + (_ddp * _sin_angle)
        _dim_xpts[4] = _ex
        _dim_ypts[4] = _ey
        self.__dim_mpts[2] = _ex
        self.__dim_mpts[3] = _ey
        if _ddp > _dv2: # dim point is radially further to vp than p2
            _dim_xpts[3] = _p2x + (_offset * _cos_angle)
            _dim_ypts[3] = _p2y + (_offset * _sin_angle)
            _dim_xpts[5] = _vx + ((_ddp + _ext) * _cos_angle)
            _dim_ypts[5] = _vy + ((_ddp + _ext) * _sin_angle)
        else: # dim point is radially closers to vp than p2
            _dim_xpts[3] = _p2x - (_offset * _cos_angle)
            _dim_ypts[3] = _p2y - (_offset * _sin_angle)
            _dim_xpts[5] = _vx + ((_ddp - _ext) * _cos_angle)
            _dim_ypts[5] = _vy + ((_ddp - _ext) * _sin_angle)
        if _angle1 < 0.0:
            _angle1 = _angle1 + (2.0 * pi)
        if _angle2 < 0.0:
            _angle2 = _angle2 + (2.0 * pi)
        self.__dim_angles[0] = (180.0/pi) * _angle1
        self.__dim_angles[1] = (180.0/pi) * _angle2

    def onDimension(self, x, y, tol=tolerance.TOL):
        """Test an x/y coordinate pair hit the dimension lines and arc.

onDimension(x, y[, tol])

This method has two required parameters:

x: The x-coordinate
y: The y-coordinate

These should both be float values.

There is an optional third parameter, "tol", giving
the maximum distance from the dimension bars that the
x/y coordinates may sit.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _dxpts = self.__dim_xpts
        _dypts = self.__dim_ypts
        #
        # test vp-p1 bar
        #
        _val = _test_on_dimension(_dxpts[0], _dypts[0],
                                  _dxpts[2], _dypts[2],
                                  _x, _y, _t)
        if _val:
            return _val
        #
        # test vp-p2 bar
        #
        _val = _test_on_dimension(_dxpts[3], _dypts[3],
                                  _dxpts[5], _dypts[5],
                                  _x, _y, _t)
        if _val:
            return _val
        #
        # test the arc
        #
        _vx, _vy = self.__vp.getCoords()
        _psep = hypot((_vx - _x), (_vy - y))
        _dx, _dy = self.getLocation()
        _dsep = hypot((_vx - _dx), (_vy - _dy))
        _val = False
        if abs(_psep - _dsep) < _t:
            _angle1 = self.__dim_angles[0]
            _angle2 = self.__dim_angles[1]
            _amax = max(_angle1, _angle2)
            _amin = min(_angle1, _angle2)
            _angle = (180.0/pi) * atan2((_vy - _y), (_vx - _x))
            if _angle1 < _angle2:
                if _angle1 < _angle < _angle2:
                    _val = True
            else:
                if ((_angle1 < _angle < (360.0 + 1e-10)) or
                    (-1e-10 < _angle < _angle2)):
                    _val = True
        return _val

#
# class stuff for dimension styles
#

class DimStyle(object):
    """A class storing preferences for Dimensions

The DimStyle class stores a set of dimension parameters
that will be used when creating dimensions when the
particular style is active.

A DimStyle object has the following methods:

getName(): Return the name of the DimStyle.
getOption(): Return a single value in the DimStyle.
getOptions(): Return all the options in the DimStyle.
getValue(): Return the value of one of the DimStyle options.

getOption() and getValue() are synonymous.
    """
    def __init__(self, name, keywords={}):
        """Instantiate a DimStyle object.

ds = DimStyle(name, keywords)

The argument "name" should be a unicode name, and the
"keyword" argument should be a dict. The keys should
be the same keywords used to set option values, such
as DIM_OFFSET, DIM_EXTENSION, etc, and the value corresponding
to each key should be set appropriately.
        """
        _n = name
        if not isinstance(_n, (str, unicode)):
            raise TypeError, "Invalid DimStyle name: "+ `_n`
        if isinstance(_n, str):
            _n = unicode(_n)
        if not isinstance(keywords, dict):
            raise TypeError, "Invalid keywords argument: " + `keywords`
        self.__opts = {}
        self.__name = _n
        for _kw in keywords:
            _val = keywords[_kw]
            options.test_option(self.__opts, _kw, _val)

    def __eq__(self, obj):
        """Test a DimStyle object for equality with another DimStyle.
        """
        if not isinstance(obj, DimStyle):
            raise TypeError, "Invalid object for DimStyle comparison: " + `obj`
        return self.__opts == obj.__opts

    def __ne__(self, obj):
        """Test a DimStyle object for inequality with another DimStyle.
        """
        if not isinstance(obj, DimStyle):
            raise TypeError, "Invalid object for DimStyle comparison: " + `obj`
        return self.__opts != obj.__opts
        
    def getName(self):
        """Return the name of the DimStyle.

getName()        
        """
        return self.__name

    name = property(getName, None, None, "DimStyle name.")

    def getOptions(self):
        """Return all the options stored within the DimStyle.

getOptions()        
        """
        return self.__opts.keys()

    def getOption(self, key):
        """Return the value of a particular option in the DimStyle.

getOption(key)

The key should be one of the strings returned from getOptions. If
there is no value found in the DimStyle for the key, the value None
is returned.
        """
        _val = None
        if self.__opts.has_key(key):
            _val = self.__opts[key]
        return _val

    def getValue(self, key):
        """Return the value of a particular option in the DimStyle.

getValue(key)

The key should be one of the strings returned from getOptions. If
there is no value found in the DimStyle for the key, the value None
is returned.
        """
        _val = None
        if self.__opts.has_key(key):
            _val = self.__opts[key]
        return _val

dimstyle_defaults = {
    'DIM_PRIMARY_FONT_FAMILY' : 'Sans',
    'DIM_PRIMARY_FONT_SIZE' : 12,
    'DIM_PRIMARY_FONT_WEIGHT' : text.NORMAL,
    'DIM_PRIMARY_FONT_STYLE' : text.NORMAL,
    'DIM_PRIMARY_FONT_COLOR' : color.Color(255,255,255),
    'DIM_PRIMARY_PREFIX' : u'',
    'DIM_PRIMARY_SUFFIX' : u'',
    'DIM_PRIMARY_PRECISION' : 3,
    'DIM_PRIMARY_UNITS' : units.MILLIMETERS,
    'DIM_PRIMARY_LEADING_ZERO' : True,
    'DIM_PRIMARY_TRAILING_DECIMAL': True,
    'DIM_SECONDARY_FONT_FAMILY' : 'Sans',
    'DIM_SECONDARY_FONT_SIZE' : 12,
    'DIM_SECONDARY_FONT_WEIGHT' : text.NORMAL,
    'DIM_SECONDARY_FONT_STYLE' : text.NORMAL,
    'DIM_SECONDARY_FONT_COLOR' : color.Color(255,255,255),
    'DIM_SECONDARY_PREFIX' : u'',
    'DIM_SECONDARY_SUFFIX' : u'',
    'DIM_SECONDARY_PRECISION' : 3,
    'DIM_SECONDARY_UNITS' : units.MILLIMETERS,
    'DIM_SECONDARY_LEADING_ZERO' : True,
    'DIM_SECONDARY_TRAILING_DECIMAL': True,
    'DIM_OFFSET' : 1.0,
    'DIM_EXTENSION': 1.0,
    'DIM_COLOR' : color.Color(255,165,0),
    'DIM_POSITION': Dimension.SPLIT,
    'DIM_ENDPOINT': Dimension.NO_ENDPOINT,
    'DIM_ENDPOINT_SIZE' : 1.0,
    'DIM_DUAL_MODE' : False,
    'RADIAL_DIM_PRIMARY_PREFIX' : u'',
    'RADIAL_DIM_PRIMARY_SUFFIX' : u'',
    'RADIAL_DIM_SECONDARY_PREFIX' : u'',
    'RADIAL_DIM_SECONDARY_SUFFIX' : u'',
    'RADIAL_DIM_DIA_MODE' : False,
    'ANGULAR_DIM_PRIMARY_PREFIX' : u'',
    'ANGULAR_DIM_PRIMARY_SUFFIX' : u'',
    'ANGULAR_DIM_SECONDARY_PREFIX' : u'',
    'ANGULAR_DIM_SECONDARY_SUFFIX' : u'',
    }

#
# dimension line testing
#
# test if a dimension line from (x1, y1), (x2, y2) is
# visiible in a region from (xmin, ymin, xmax, ymax)
#

def _test_dim_line(t1, t2, xmin, ymin, xmax, ymax):
    _x1, _y1 = t1
    _x2, _y2 = t2
    _val = False
    #
    # see if dim line intersects an imaginary segment
    # from (xmin, ymax) to (xmax, ymin)
    #
    _d = ((xmax - xmin)*(_y2 - _y1)) - ((ymin - ymax)*(_x2 - _x1))
    if abs(_d) > 1e-10: # not parallel
        _rn = ((ymax - _y1)*(_x2 - _x1)) - ((xmin - _x1)*(_y2 - _y1))
        _r = _rn/_d
        if 0.0 < _r < 1.0:
            _sn = ((ymax - _y1)*(xmax - xmin)) - ((xmin - _x1)*(ymin - ymax))
            _s = _sn/_d
            if 0.0 < _s < 1.0:
                _val = True
    if _val:
        return _val
    #
    # see if dim line intersects an imaginary segment
    # from (xmin, ymin) to (xmax, ymax)
    #
    _d = ((xmax - xmin)*(_y2 - _y1)) - ((ymax - ymin)*(_x2 - _x1))
    if abs(_d) > 1e-10: # not parallel
        _rn = ((ymin - _y1)*(_x2 - _x1)) - ((xmin - _x1)*(_y2 - _y1))
        _r = _rn/_d
        if 0.0 < _r < 1.0:
            _sn = ((ymax - _y1)*(xmax - xmin)) - ((xmin - _x1)*(ymax - ymin))
            _s = _sn/_d
            if 0.0 < _s < 1.0:
                _val = True
    return _val

#
# test if a projected point from an x/y coordinate pair
# is within a tolerance limited distance from the coordinate
#

def _test_on_dimension(dx1, dy1, dx2, dy2, x, y, t):
        _dx = dx2 - dx1
        _dy = dy2 - dy1
        _d = pow(_dx, 2) + pow(_dy, 2)
        _r = ((x - dx1)*(_dx) + (y - dy1)*(_dy))/_d
        if _r < 0.0 : _r = 0.0
        if _r > 1.0 : _r = 1.0
        _px = dx1 + (_r * _dx)
        _py = dy1 + (_r * _dy)
        _val = False
        if ((abs(_px - x) < t) and (abs(_py - y) < t)):
            _val = True
        return _val
