#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# code for doing file saves/loads
#

import layer
import color
import linetype
import style
import point
import segment
import circle
import arc
import hcline
import vcline
import acline
import cline
import ccircle
import units
import segjoint
import leader
import dimension

import xml.dom.minidom
import fileio
import os

def _save_graph_bits(obj, node, doc, map):
    _st = obj.getStyle()
    assert _st in map, "Object %s style %s not in map!" % (str(obj), `_st`)
    _attr = doc.createAttributeNS("image", "style")
    node.setAttributeNodeNS(_attr)
    node.setAttributeNS("image", "style", `map[_st]`)
    _color = obj.getColor()
    if _color != _st.getColor():
        assert _color in map, "Object %s color not in map!" % str(obj)
        _attr = doc.createAttributeNS("image", "color")
        node.setAttributeNodeNS(_attr)
        node.setAttributeNS("image", "color", `map[_color]`)
    _linetype = obj.getLinetype()
    if _linetype != _st.getLinetype():
        assert _linetype in map, "Object %s linetype not in map!" % str(obj)
        _attr = doc.createAttributeNS("image", "linetype")
        node.setAttributeNodeNS(_attr)
        node.setAttributeNS("image", "linetype", `map[_linetype]`)
    _thickness = obj.getThickness()
    _tdiff = _thickness - _st.getThickness()
    if abs(_tdiff) > 1e-10:
        _attr = doc.createAttributeNS("image", "thickness")
        node.setAttributeNodeNS(_attr)
        node.setAttributeNS("image", "thickness", `_thickness`)

def _save_leaders(lyr, node, doc, map):
    _leaders = lyr.getLayerEntities("leader")
    if len(_leaders):
        _leads = doc.createElementNS("image", "image:Leaders")
        node.appendChild(_leads)
        _i = 0
        for _lead in _leaders:
            _p1, _p2, _p3 = _lead.getLeaderPoints()
            assert _p1 in map, "Leader %s p1 not in map!" % str(_lead)
            assert _p2 in map, "Leader %s p2 not in map!" % str(_lead)
            assert _p3 in map, "Leader %s p3 not in map!" % str(_lead)
            _child = doc.createElementNS("image", "image:Leader")
            _leads.appendChild(_child)
            _attr = doc.createAttributeNS("image", "id")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "id", `_i`)
            _attr = doc.createAttributeNS("image", "p1")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "p1", `map[_p1]`)
            _attr = doc.createAttributeNS("image", "p2")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "p2", `map[_p2]`)
            _attr = doc.createAttributeNS("image", "p3")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "p3", `map[_p3]`)
            _attr = doc.createAttributeNS("image", "size")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "size", `_lead.getArrowSize()`)
            _save_graph_bits(_lead, _child, doc, map)
            _i = _i + 1
            
def _save_chamfers(lyr, node, doc, map):
    _chamferlist = lyr.getLayerEntities("chamfer")
    if len(_chamferlist):
        _chs = doc.createElementNS("image", "image:Chamfers")
        node.appendChild(_chs)
        _i = 0
        for _ch in _chamferlist:
            _s1, _s2 = _ch.getSegments()
            assert _s1 in map, "Chamfer %s seg1 not in map!" % `_ch`
            assert _s2 in map, "Chamfer %s seg2 not in map!" % `_ch`
            _child = doc.createElementNS("image", "image:Chamfer")
            _chs.appendChild(_child)
            _attr = doc.createAttributeNS("image", "id")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "id", `_i`)
            _attr = doc.createAttributeNS("image", "s1")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "s1", `map[_s1]`)
            _attr = doc.createAttributeNS("image", "s2")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "s2", `map[_s2]`)
            _attr = doc.createAttributeNS("image", "length")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "length", `_ch.getLength()`)
            _save_graph_bits(_ch, _child, doc, map)
            _i = _i + 1

def _save_fillets(lyr, node, doc, map):
    _filletlist = lyr.getLayerEntities("fillet")
    if len(_filletlist):
        _flts = doc.createElementNS("image", "image:Fillets")
        node.appendChild(_flts)
        _i = 0
        for _flt in _filletlist:
            _s1, _s2 = _flt.getSegments()
            assert _s1 in map, "Fillet %s seg1 not in map!" % `_flt`
            assert _s2 in map, "Fillet %s seg2 not in map!" % `_flt`
            _child = doc.createElementNS("image", "image:Fillet")
            _flts.appendChild(_child)
            _attr = doc.createAttributeNS("image", "id")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "id", `_i`)
            _attr = doc.createAttributeNS("image", "_s1")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "s1", `map[_s1]`)
            _attr = doc.createAttributeNS("image", "s2")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "s2", `map[_s2]`)
            _attr = doc.createAttributeNS("image", "radius")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "radius", `_flt.getRadius()`)
            _save_graph_bits(_flt, _child, doc, map)
            _i = _i + 1

def _save_dim_override(node, doc, map, stkey, value):
    _child = doc.createElementNS("image", "image:DimOption")
    node.appendChild(_child)
    _attr = doc.createAttributeNS("image", "opt")
    _child.setAttributeNodeNS(_attr)
    _child.setAttributeNS("image", "opt", stkey)
    if (stkey == 'DIM_COLOR' or
        stkey == 'DIM_PRIMARY_FONT_COLOR' or
        stkey == 'DIM_SECONDARY_FONT_COLOR'):
        _attr = doc.createAttributeNS("image", "cid")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "cid", `map[value]`)
    elif (stkey == 'DIM_PRIMARY_FONT_FAMILY' or
          stkey == 'DIM_SECONDARY_FONT_FAMILY'):
        _attr = doc.createAttributeNS("image", "fid")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "fid", `map[value]`)
    else:
        _attr = doc.createAttributeNS("image", "val")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "val", str(value))
    
def _save_dim_common(dim, node, doc, map):
    print "in _save_dim_common()"
    ds = dim.getDimStyle()
    _offset = dim.getOffset()
    if abs(ds.getValue('DIM_OFFSET') - _offset) > 1e-10:
        _save_dim_override(node, doc, map, 'DIM_OFFSET', _offset)
    _ext = dim.getExtension()
    if abs(ds.getValue('DIM_EXTENSION') - _ext) > 1e-10:
        _save_dim_override(node, doc, map, 'DIM_EXTENSION', _ext)
    _endpt = dim.getEndpointType()
    if ds.getValue('DIM_ENDPOINT') != _endpt:
        _save_dim_override(node, doc, map, 'DIM_ENDPOINT', _endpt)
    _size = dim.getEndpointSize()
    if abs(ds.getValue('DIM_ENDPOINT_SIZE') - _size) > 1e-10:
        _save_dim_override(node, doc, map, 'DIM_ENDPOINT_SIZE', _size)
    _dual_mode = dim.getDualDimMode()
    if ds.getValue('DIM_DUAL_MODE') != _dual_mode:
        print "different dual dimension modes!"
        _save_dim_override(node, doc, map, 'DIM_DUAL_MODE', _dual_mode)
    else:
        print "same dual dimension modes!"
    #
    # primary dimension string
    #
    pds = dim.getPrimaryDimstring()
    prefix = pds.getPrefix()
    if isinstance(dim, dimension.LinearDimension):
        if ds.getValue('DIM_PRIMARY_PREFIX') != prefix:
            _save_dim_override(node, doc, map,
                               'DIM_PRIMARY_PREFIX', prefix)
    elif isinstance(dim, dimension.RadialDimension):
        if ds.getValue('RADIAL_DIM_PRIMARY_PREFIX') != prefix:
            _save_dim_override(node, doc, map,
                               'RADIAL_DIM_PRIMARY_PREFIX', prefix)
    elif isinstance(dim, dimension.AngularDimension):
        if ds.getValue('ANGULAR_DIM_PRIMARY_PREFIX') != prefix:
            _save_dim_override(node, doc, map,
                               'ANGULAR_DIM_PRIMARY_PREFIX', prefix)
    else:
        raise TypeError, "Unexpected dimension type: " + `dim`
    suffix = pds.getSuffix()
    if isinstance(dim, dimension.LinearDimension):
        if ds.getValue('DIM_PRIMARY_SUFFIX') != suffix:
            _save_dim_override(node, doc, map,
                               'DIM_PRIMARY_SUFFIX', suffix)
    elif isinstance(dim, dimension.RadialDimension):
        if ds.getValue('RADIAL_DIM_PRIMARY_SUFFIX') != suffix:
            _save_dim_override(node, doc, map,
                               'RADIAL_DIM_PRIMARY_SUFFIX', suffix)
    elif isinstance(dim, dimension.AngularDimension):
        if ds.getValue('ANGULAR_DIM_PRIMARY_SUFFIX') != suffix:
            _save_dim_override(node, doc, map,
                               'ANGULAR_DIM_PRIMARY_SUFFIX', prefix)
    else:
        raise TypeError, "Unexpected dimension type: " + `dim`
    unit = pds.getUnits()
    if ds.getValue('DIM_PRIMARY_UNITS') != unit:
        _save_dim_override(node, doc, map, 'DIM_PRIMARY_UNITS', unit)
    precision = pds.getPrecision()
    if ds.getValue('DIM_PRIMARY_PRECISION') != precision:
        _save_dim_override(node, doc, map, 'DIM_PRIMARY_PRECISION',
                           precision)
    print_zero = pds.getPrintZero()
    if ds.getValue('DIM_PRIMARY_LEADING_ZERO') != print_zero:
        _save_dim_override(node, doc, map, 'DIM_PRIMARY_LEADING_ZERO',
                           print_zero)
    print_dec = pds.getPrintDecimal()
    if ds.getValue('DIM_PRIMARY_TRAILING_DECIMAL') != print_dec:
        _save_dim_override(node, doc, map, 'DIM_PRIMARY_TRAILING_DECIMAL',
                           print_dec)
    family = pds.getFamily()
    if ds.getValue('DIM_PRIMARY_FONT_FAMILY') != family:
        assert family in map, "Font family missing in map: " + `family`
        _save_dim_override(node, doc, map, 'DIM_PRIMARY_FONT_FAMILY', family)
    size = pds.getSize()
    if ds.getValue('DIM_PRIMARY_FONT_SIZE') != size:
        _save_dim_override(node, doc, map, 'DIM_PRIMARY_FONT_SIZE', size)
    font_style = pds.getStyle()
    if ds.getValue('DIM_PRIMARY_FONT_STYLE') != font_style:
        _save_dim_override(node, doc, map, 'DIM_PRIMARY_FONT_STYLE',
                           font_style)
    weight = pds.getWeight()
    if ds.getValue('DIM_PRIMARY_FONT_WEIGHT') != weight:
        _save_dim_override(node, doc, map, 'DIM_PRIMARY_FONT_WEIGHT', weight)
    font_color = pds.getColor()
    if ds.getValue('DIM_PRIMARY_FONT_COLOR') != font_color:
        assert font_color in map, "Font color missing in map: " + `font_color`
        _save_dim_override(node, doc, map, 'DIM_PRIMARY_FONT_COLOR',
                           font_color)
    #
    # secondary dimension string
    #
    sds = dim.getSecondaryDimstring()
    prefix = sds.getPrefix()
    if isinstance(dim, dimension.LinearDimension):
        if ds.getValue('DIM_SECONDARY_PREFIX') != prefix:
            _save_dim_override(node, doc, map,
                               'DIM_SECONDARY_PREFIX', prefix)
    elif isinstance(dim, dimension.RadialDimension):
        if ds.getValue('RADIAL_DIM_SECONDARY_PREFIX') != prefix:
            _save_dim_override(node, doc, map,
                               'RADIAL_DIM_SECONDARY_PREFIX', prefix)
    elif isinstance(dim, dimension.AngularDimension):
        if ds.getValue('ANGULAR_DIM_SECONDARY_PREFIX') != prefix:
            _save_dim_override(node, doc, map,
                               'ANGULAR_DIM_SECONDARY_PREFIX', prefix)
    else:
        raise TypeError, "Unexpected dimension type: " + `dim`
    suffix = sds.getSuffix()
    if isinstance(dim, dimension.LinearDimension):
        if ds.getValue('DIM_SECONDARY_SUFFIX') != suffix:
            _save_dim_override(node, doc, map,
                               'DIM_SECONDARY_SUFFIX', suffix)
    elif isinstance(dim, dimension.RadialDimension):
        if ds.getValue('RADIAL_DIM_SECONDARY_SUFFIX') != suffix:
            _save_dim_override(node, doc, map,
                               'RADIAL_DIM_SECONDARY_SUFFIX', suffix)
    elif isinstance(dim, dimension.AngularDimension):
        if ds.getValue('ANGULAR_DIM_SECONDARY_SUFFIX') != suffix:
            _save_dim_override(node, doc, map,
                               'ANGULAR_DIM_SECONDARY_SUFFIX', prefix)
    else:
        raise TypeError, "Unexpected dimension type: " + `dim`
    unit = sds.getUnits()
    if ds.getValue('DIM_SECONDARY_UNITS') != unit:
        _save_dim_override(node, doc, map, 'DIM_SECONDARY_UNITS', unit)
    precision = sds.getPrecision()
    if ds.getValue('DIM_SECONDARY_PRECISION') != precision:
        _save_dim_override(node, doc, map, 'DIM_SECONDARY_PRECISION',
                           precision)
    print_zero = sds.getPrintZero()
    if ds.getValue('DIM_SECONDARY_LEADING_ZERO') != print_zero:
        _save_dim_override(node, doc, map, 'DIM_SECONDARY_LEADING_ZERO',
                           print_zero)
    print_dec = sds.getPrintDecimal()
    if ds.getValue('DIM_SECONDARY_TRAILING_DECIMAL') != print_dec:
        _save_dim_override(node, doc, map, 'DIM_SECONDARY_TRAILING_DECIMAL',
                           print_dec)
    family = sds.getFamily()
    if ds.getValue('DIM_SECONDARY_FONT_FAMILY') != family:
        assert family in map, "Font family missing in map: " + `family`
        _save_dim_override(node, doc, map, 'DIM_SECONDARY_FONT_FAMILY', family)
    size = sds.getSize()
    if ds.getValue('DIM_SECONDARY_FONT_SIZE') != size:
        _save_dim_override(node, doc, map, 'DIM_SECONDARY_FONT_SIZE', size)
    font_style = sds.getStyle()
    if ds.getValue('DIM_SECONDARY_FONT_STYLE') != font_style:
        _save_dim_override(node, doc, map, 'DIM_SECONDARY_FONT_STYLE',
                           font_style)
    weight = sds.getWeight()
    if ds.getValue('DIM_SECONDARY_FONT_WEIGHT') != weight:
        _save_dim_override(node, doc, map, 'DIM_SECONDARY_FONT_WEIGHT', weight)
    font_color = sds.getColor()
    if ds.getValue('DIM_SECONDARY_FONT_COLOR') != font_color:
        assert font_color in map, "Font color missing in map: " + `font_color`
        _save_dim_override(node, doc, map, 'DIM_SECONDARY_FONT_COLOR',
                           font_color)

def _save_linear_dims(l, node, doc, map, dimmap, dimtype):
    if dimtype == 'linear_dimension':
        _tag = 'image:LDims'
    elif dimtype == 'horizontal_dimension':
        _tag = 'image:HDims'
    elif dimtype == 'vertical_dimension':
        _tag = 'image:VDims'
    else:
        raise ValueError, "Unexpected linear dimension type: " + `dimtype`
    _ldimlist = l.getLayerEntities(dimtype)
    if len(_ldimlist):
        _item_tag = _tag[:-1]
        _block = doc.createElementNS("image", _tag)
        node.appendChild(_block)
        _i = 0
        for _dim in _ldimlist:
            _p1, _p2 = _dim.getDimPoints()
            _l1, _l2 = _dim.getDimLayers()
            _x, _y = _dim.getLocation()
            _ds = _dim.getDimStyle()
            _child = doc.createElementNS("image", _item_tag)
            _block.appendChild(_child)
            _attr = doc.createAttributeNS("image", "id")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "id", `_i`)
            _attr = doc.createAttributeNS("image", "l1")
            _child.setAttributeNodeNS(_attr)
            if _l1 in map:
                _child.setAttributeNS("image", "l1", `map[_l1]`)
            else:
                dimmap[_child] = []
                dimmap[_child].append((_l1, "l1"))
            _attr = doc.createAttributeNS("image", "p1")
            _child.setAttributeNodeNS(_attr)
            if _p1 in map:
                _child.setAttributeNS("image", "p1", `map[_p1]`)
            else:
                if _child not in dimmap:
                    dimmap[_child] = []
                dimmap[_child].append((_p1, "p1"))
            _attr = doc.createAttributeNS("image", "l2")
            _child.setAttributeNodeNS(_attr)
            if _l2 in map:
                _child.setAttributeNS("image", "l2", `map[_l2]`)
            else:
                if _child not in dimmap:
                    dimmap[_child] = []
                dimmap[_child].append((_l2, "l2"))
            _attr = doc.createAttributeNS("image", "p2")
            _child.setAttributeNodeNS(_attr)
            if _p2 in map:
                _child.setAttributeNS("image", "p2", `map[_p2]`)
            else:
                if _child not in dimmap:
                    dimmap[_child] = []
                dimmap[_child].append((_p2, "p2"))
            _attr = doc.createAttributeNS("image", "x")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "x", `_x`)
            _attr = doc.createAttributeNS("image", "y")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "y", `_y`)
            _attr = doc.createAttributeNS("image", "ds")
            assert _ds in map, "Missing DimStyle in map: " + `_ds`
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "ds", `map[_ds]`)
            _save_dim_common(_dim, _child, doc, map)
            
def _save_radial_dims(lyr, node, doc, map, dimmap):
    _rdimlist = lyr.getLayerEntities("radial_dimension")
    if len(_rdimlist):
        _rdims = doc.createElementNS("image", "image:RDims")
        node.appendChild(_rdims)
        _i = 0
        for _rdim in _rdimlist:
            _dl = _rdim.getDimLayer()
            _dc = _rdim.getDimCircle()
            _x, _y = _rdim.getLocation()
            _ds = _rdim.getDimStyle()
            _child = doc.createElementNS("image", "image:RDim")
            _rdims.appendChild(_child)
            _attr = doc.createAttributeNS("image", "id")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "id", `_i`)
            _attr = doc.createAttributeNS("image", "l")
            _child.setAttributeNodeNS(_attr)
            if _dl in map:
                _child.setAttributeNS("image", "l", `map[_dl]`)
            else:
                dimmap[_child] = []
                dimmap[_child].append((_dl, "l"))
            _attr = doc.createAttributeNS("image", "c")
            _child.setAttributeNodeNS(_attr)
            if _dc in map:
                _child.setAttributeNS("image", "c", `map[_dc]`)
            else:
                if _child not in dimmap:
                    dimmap[_child] = []
                dimmap[_child].append((_dc, "c"))
            _attr = doc.createAttributeNS("image", "x")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "x", `_x`)
            _attr = doc.createAttributeNS("image", "y")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "y", `_y`)
            _attr = doc.createAttributeNS("image", "ds")
            assert _ds in map, "Missing DimStyle in map: " + `_ds`
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "ds", `map[_ds]`)
            _save_dim_common(_rdim, _child, doc, map)
            _dia_mode = _rdim.getDiaMode()
            if _ds.getValue('RADIAL_DIM_DIA_MODE') != _dia_mode:
                _opt = doc.createElementNS("image", "image:DimOption")
                _child.appendChild(_opt)
                _attr = doc.createAttributeNS("image", "opt")
                _opt.setAttributeNodeNS(_attr)
                _opt.setAttributeNS("image", "opt", 'RADIAL_DIM_DIA_MODE')
                _attr = doc.createAttributeNS("image", "val")
                _opt.setAttributeNodeNS(_attr)
                _opt.setAttributeNS("image", "val", `_dia_mode`)

def _save_angular_dims(lyr, node, doc, map, dimmap):
    _adimlist = lyr.getLayerEntities("angular_dimension")
    if len(_adimlist):
        _adims = doc.createElementNS("image", "image:ADims")
        node.appendChild(_adims)
        _i = 0
        for _adim in _adimlist:
            _p1, _p2, _p3 = _adim.getDimPoints()
            _l1, _l2, _l3 = _adim.getDimLayers()
            _x, _y = _adim.getLocation()
            _ds = _adim.getDimStyle()
            _child = doc.createElementNS("image", "image:ADim")
            _adims.appendChild(_child)
            _attr = doc.createAttributeNS("image", "id")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "id", `_i`)
            _attr = doc.createAttributeNS("image", "l1")
            _child.setAttributeNodeNS(_attr)
            if _l1 in map:
                _child.setAttributeNS("image", "l1", `map[_l1]`)
            else:
                dimmap[_child] = []
                dimmap[_child].append((_l1, "_l1"))
            _attr = doc.createAttributeNS("image", "p1")
            _child.setAttributeNodeNS(_attr)
            if _p1 in map:
                _child.setAttributeNS("image", "p1", `map[_p1]`)
            else:
                if _child not in dimmap:
                    dimmap[_child] = []
                dimmap[_child].append((_p1, "p1"))
            _attr = doc.createAttributeNS("image", "l2")
            _child.setAttributeNodeNS(_attr)
            if _l2 in map:
                _child.setAttributeNS("image", "l2", `map[_l2]`)
            else:
                if _child not in dimmap:
                    dimmap[_child] = []
                dimmap[_child].append((_l2, "l2"))
            _attr = doc.createAttributeNS("image", "p2")
            _child.setAttributeNodeNS(_attr)
            if _p2 in map:
                _child.setAttributeNS("image", "p2", `map[_p2]`)
            else:
                if _child not in dimmap:
                    dimmap[_child] = []
                dimmap[_child].append((_p2, "p2"))
            _attr = doc.createAttributeNS("image", "l3")
            _child.setAttributeNodeNS(_attr)
            if _l3 in map:
                _child.setAttributeNS("image", "l3", `map[_l3]`)
            else:
                if _child not in dimmap:
                    dimmap[_child] = []
                dimmap[_child].append((_l3, "l3"))
            _attr = doc.createAttributeNS("image", "p3")
            _child.setAttributeNodeNS(_attr)
            if _p3 in map:
                _child.setAttributeNS("image", "p3", `map[_p3]`)
            else:
                if _child not in dimmap:
                    dimmap[_child] = []
                dimmap[_child].append((_p3, "p3"))
            _attr = doc.createAttributeNS("image", "x")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "x", `_x`)
            _attr = doc.createAttributeNS("image", "y")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "y", `_y`)
            _attr = doc.createAttributeNS("image", "ds")
            assert _ds in map, "Missing DimStyle in map: " + `_ds`
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "ds", `map[_ds]`)
            _save_dim_common(_adim, _child, doc, map)

def _save_ccircles(lyr, node, doc, map):
    _cclist = lyr.getLayerEntities("ccircle")
    if len(_cclist):
        _ccircs = doc.createElementNS("image", "image:CCircles")
        node.appendChild(_ccircs)
        _i = 0
        for _cc in _cclist:
            _cp = _cc.getCenter()
            assert _cp in map, "CCircle %s center not in map!" % str(_cc)
            _child = doc.createElementNS("image", "image:CCircle")
            _ccircs.appendChild(_child)
            _attr = doc.createAttributeNS("image", "id")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "id", `_i`)
            _attr = doc.createAttributeNS("image", "cp")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "cp", `map[_cp]`)
            _attr = doc.createAttributeNS("image", "r")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "r", `_cc.getRadius()`)
            _i = _i + 1
    
def _save_clines(lyr, node, doc, map):
    _cllist = lyr.getLayerEntities("cline")
    if len(_cllist):
        _clines = doc.createElementNS("image", "image:CLines")
        node.appendChild(_clines)
        _i = 0
        for _cl in _cllist:
            _p1, _p2 = _cl.getKeypoints()
            assert _p1 in map, "CLine %s p1 not in map!" % str(_cl)
            assert _p2 in map, "CLine %s p2 not in map!" % str(_cl)
            _child = doc.createElementNS("image", "image:CLine")
            _clines.appendChild(_child)
            _attr = doc.createAttributeNS("image", "id")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "id", `_i`)
            _attr = doc.createAttributeNS("image", "p1")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "p1", `map[_p1]`)
            _attr = doc.createAttributeNS("image", "p2")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "p2", `map[_p2]`)
            _i = _i + 1

def _save_aclines(lyr, node, doc, map):
    _acllist = lyr.getLayerEntities("acline")
    if len(_acllist):
        _acls = doc.createElementNS("image", "image:ACLines")
        node.appendChild(_acls)
        _i = 0
        for _acl in _acllist:
            _loc = _acl.getLocation()
            assert _loc in map, "ACline %s point not in map!" % str(_acl)
            _child = doc.createElementNS("image", "image:ACLine")
            _acls.appendChild(_child)
            _attr = doc.createAttributeNS("image", "id")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "id", `_i`)
            _attr = doc.createAttributeNS("image", "angle")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "angle", `_acl.getAngle()`)
            _attr = doc.createAttributeNS("image", "location")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "location", `map[_loc]`)
            _i = _i + 1
    
def _save_vclines(lyr, node, doc, map):
    _vcllist = lyr.getLayerEntities("vcline")
    if len(_vcllist):
        _vcls = doc.createElementNS("image", "image:VCLines")
        node.appendChild(_vcls)
        _i = 0
        for _vcl in _vcllist:
            _loc = _vcl.getLocation()
            assert _loc in map, "VCline %s point not in map!" % str(_vcl)
            _child = doc.createElementNS("image", "image:VCLine")
            _vcls.appendChild(_child)
            _attr = doc.createAttributeNS("image", "id")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "id", `_i`)
            _attr = doc.createAttributeNS("image", "location")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "location", `map[_loc]`)
            _i = _i + 1
    
def _save_hclines(lyr, node, doc, map):
    _hcllist = lyr.getLayerEntities("hcline")
    if len(_hcllist):
        _hcls = doc.createElementNS("image", "image:HCLines")
        node.appendChild(_hcls)
        _i = 0
        for _hcl in _hcllist:
            _loc = _hcl.getLocation()
            assert _loc in map, "HCline %s point not in map!" % str(_hcl)
            _child = doc.createElementNS("image", "image:HCLine")
            _hcls.appendChild(_child)
            _attr = doc.createAttributeNS("image", "id")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "id", `_i`)
            _attr = doc.createAttributeNS("image", "location")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "location", `map[_loc]`)
            _i = _i + 1
    
def _save_arcs(lyr, node, doc, map):
    _arclist = lyr.getLayerEntities("arc")
    if len(_arclist):
        _arcs = doc.createElementNS("image", "image:Arcs")
        node.appendChild(_arcs)
        _i = 0
        for _arc in _arclist:
            _cp = _arc.getCenter()
            assert _cp in map, "Arc %s center not in map!" % str(_arc)
            _child = doc.createElementNS("image", "image:Arc")
            _arcs.appendChild(_child)
            _attr = doc.createAttributeNS("image", "id")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "id", `_i`)
            _attr = doc.createAttributeNS("image", "cp")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "cp", `map[_cp]`)
            _attr = doc.createAttributeNS("image", "r")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "r", `_arc.getRadius()`)
            _attr = doc.createAttributeNS("image", "sa")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "sa", `_arc.getStartAngle()`)
            _attr = doc.createAttributeNS("image", "ea")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "ea", `_arc.getEndAngle()`)
            _save_graph_bits(_arc, _child, doc, map)
            map[_arc] = _i
            _i = _i + 1
    
def _save_circles(lyr, node, doc, map):
    _circles = lyr.getLayerEntities("circle")
    if len(_circles):
        _circs = doc.createElementNS("image", "image:Circles")
        node.appendChild(_circs)
        _i = 0
        for _circle in _circles:
            _cp = _circle.getCenter()
            assert _cp in map, "Circle %s center not in map!" % str(_circle)
            _child = doc.createElementNS("image", "image:Circle")
            _circs.appendChild(_child)
            _attr = doc.createAttributeNS("image", "id")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "id", `_i`)
            _attr = doc.createAttributeNS("image", "cp")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "cp", `map[_cp]`)
            _attr = doc.createAttributeNS("image", "r")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "r", `_circle.getRadius()`)
            _save_graph_bits(_circle, _child, doc, map)
            map[_circle] = _i
            _i = _i + 1

def _save_segments(lyr, node, doc, map):
    _segments = lyr.getLayerEntities("segment")
    if len(_segments):
        _segs = doc.createElementNS("image", "image:Segments")
        node.appendChild(_segs)
        _i = 0
        for _seg in _segments:
            _p1, _p2 = _seg.getEndpoints()
            assert _p1 in map, "Segment %s p1 not in map!" % str(_seg)
            assert _p2 in map, "Segment %s p2 not in map!" % str(_seg)
            _child = doc.createElementNS("image", "image:Segment")
            _segs.appendChild(_child)
            _attr = doc.createAttributeNS("image", "id")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "id", `_i`)
            _attr = doc.createAttributeNS("image", "p1")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "p1", `map[_p1]`)
            _attr = doc.createAttributeNS("image", "p2")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "p2", `map[_p2]`)
            _save_graph_bits(_seg, _child, doc, map)
            map[_seg] = _i
            _i = _i + 1
        
def _save_points(lyr, node, doc, map):
    _points = lyr.getLayerEntities("point")
    if len(_points):
        _pts = doc.createElementNS("image", "image:Points")
        node.appendChild(_pts)
        _i = 0
        for _pt in _points:
            _x, _y = _pt.getCoords()
            _child = doc.createElementNS("image", "image:Point")
            _pts.appendChild(_child)
            _attr = doc.createAttributeNS("image", "id")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "id", `_i`)
            _attr = doc.createAttributeNS("image", "x")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "x", `_x`)
            _attr = doc.createAttributeNS("image", "y")
            _child.setAttributeNodeNS(_attr)
            _child.setAttributeNS("image", "y", `_y`)
            map[_pt] = _i
            _i = _i + 1
    
def _make_xml_doc():
    newdoc = xml.dom.minidom.Document()
    root = newdoc.createElementNS("image", "image:Image")
    newdoc.appendChild(root)
    attr = newdoc.createAttributeNS("xmlns", "xmlns:image")
    root.setAttributeNodeNS(attr)
    root.setAttributeNS("xmlns", "xmlns:image", "http://www.pythoncad.org")
    attr = newdoc.createAttributeNS("xmlns", "xmlns:xsi")
    root.setAttributeNodeNS(attr)
    root.setAttributeNS("xmlns", "xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
    attr = newdoc.createAttributeNS("xsi", "xsi:schemaLocation")
    root.setAttributeNodeNS(attr)
    root.setAttributeNS("xmlns", "xsi:schemaLocation", "http://www.pythoncad.org")
    return newdoc

def _save_dimstyles(attmap, node, doc, map):
    _dimstyles = doc.createElementNS("image", "image:DimStyles")
    node.appendChild(_dimstyles)
    _dslist = attmap['dimstyles']
    _i = 0
    for _ds in _dslist:
        _child = doc.createElementNS("image", "image:DimStyle")
        _dimstyles.appendChild(_child)
        _attr = doc.createAttributeNS("image", "id")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "id", `_i`)
        _attr = doc.createAttributeNS("image", "name")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "name", _ds.getName())
        _stkeys = _ds.getOptions()
        _stkeys.sort()
        for _stkey in _stkeys:
            _value = _ds.getValue(_stkey)
            _dsfield = doc.createElementNS("image", "image:DimOption")
            _child.appendChild(_dsfield)
            _attr = doc.createAttributeNS("image", "opt")
            _dsfield.setAttributeNodeNS(_attr)
            _dsfield.setAttributeNS("image", "opt", _stkey)
            if (_stkey == 'DIM_COLOR' or
                _stkey == 'DIM_PRIMARY_FONT_COLOR' or
                _stkey == 'DIM_SECONDARY_FONT_COLOR'):
                assert _value in map, "color %s not found in map!" % str(_value)
                _attr = doc.createAttributeNS("image", "cid")
                _dsfield.setAttributeNodeNS(_attr)
                _dsfield.setAttributeNS("image", "cid", `map[_value]`)
            elif (_stkey == 'DIM_PRIMARY_FONT_FAMILY' or
                  _stkey == 'DIM_SECONDARY_FONT_FAMILY'):
                assert _value in map, "font %s not found in map!" % str(_value)
                _attr = doc.createAttributeNS("image", "fid")
                _dsfield.setAttributeNodeNS(_attr)
                _dsfield.setAttributeNS("image", "fid", `map[_value]`)
            elif (_stkey == 'DIM_PRIMARY_PREFIX' or
                  _stkey == 'DIM_PRIMARY_SUFFIX' or
                  _stkey == 'DIM_SECONDARY_PREFIX' or
                  _stkey == 'DIM_SECONDARY_SUFFIX' or
                  _stkey == 'RADIAL_DIM_PRIMARY_PREFIX' or
                  _stkey == 'RADIAL_DIM_PRIMARY_SUFFIX' or
                  _stkey == 'RADIAL_DIM_SECONDARY_PREFIX' or
                  _stkey == 'RADIAL_DIM_SECONDARY_SUFFIX' or
                  _stkey == 'ANGULAR_DIM_PRIMARY_PREFIX' or
                  _stkey == 'ANGULAR_DIM_PRIMARY_SUFFIX' or
                  _stkey == 'ANGULAR_DIM_SECONDARY_PREFIX' or
                  _stkey == 'ANGULAR_DIM_SECONDARY_SUFFIX'):
                _attr = doc.createAttributeNS("image", "val")
                _dsfield.setAttributeNodeNS(_attr)
                _dsfield.setAttributeNS("image", "val", str(_value))
            else:
                _attr = doc.createAttributeNS("image", "val")
                _dsfield.setAttributeNodeNS(_attr)
                _dsfield.setAttributeNS("image", "val", `_value`)
        map[_ds] = _i
        _i = _i + 1

def _save_font_families(attmap, node, doc, map):
    _families = doc.createElementNS("image", "image:FontFamilies")
    node.appendChild(_families)
    _i = 0
    _fontlist = attmap['fonts']
    for _font in _fontlist:
        _child = doc.createElementNS("image", "image:FontFamily")
        _families.appendChild(_child)
        _attr = doc.createAttributeNS("image", "id")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "id", `_i`)
        _attr = doc.createAttributeNS("image", "name")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "name", _font)
        map[_font] = _i
        _i = _i + 1
    
def _save_styles(attmap, node, doc, map):
    _styles = doc.createElementNS("image", "image:Styles")
    node.appendChild(_styles)
    _i = 0
    _stylelist = attmap['styles']
    for _style in _stylelist:
        _child = doc.createElementNS("image", "image:Style")
        _styles.appendChild(_child)
        _attr = doc.createAttributeNS("image", "id")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "id", `_i`)
        _attr = doc.createAttributeNS("image", "name")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "name", _style.getName())
        _color = _style.getColor()
        assert _color in map, "Color not found in map!"
        _attr = doc.createAttributeNS("image", "cid")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "cid", `map[_color]`)
        _linetype = _style.getLinetype()
        assert _linetype in map, "Linetype not found in map!"
        _attr = doc.createAttributeNS("image", "ltid")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "ltid", `map[_linetype]`)
        _attr = doc.createAttributeNS("image", "thickness")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "thickness", `_style.getThickness()`)
        map[_style] = _i
        _i = _i + 1
    
def _save_colors(attmap, node, doc, map):
    _colors = doc.createElementNS("image", "image:Colors")
    node.appendChild(_colors)
    _i = 0
    _colorlist = attmap['colors']
    for _color in _colorlist:
        _r, _g, _b = _color.getColors()
        _child = doc.createElementNS("image", "image:Color")
        _colors.appendChild(_child)
        _attr = doc.createAttributeNS("image", "id")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "id", `_i`)
        _attr = doc.createAttributeNS("image", "r")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "r", `_r`)
        _attr = doc.createAttributeNS("image", "g")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "g", `_g`)
        _attr = doc.createAttributeNS("image", "b")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "b", `_b`)
        map[_color] = _i
        _i = _i + 1
        
def _save_linetypes(attmap, node, doc, map):
    _linetypes = doc.createElementNS("image", "image:Linetypes")
    node.appendChild(_linetypes)
    _i = 0
    _linetypelist = attmap['linetypes']
    for _linetype in _linetypelist:
        _child = doc.createElementNS("image", "image:Linetype")
        _linetypes.appendChild(_child)
        _attr = doc.createAttributeNS("image", "id")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "id", `_i`)
        _attr = doc.createAttributeNS("image", "name")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "name", _linetype.getName())
        _attr = doc.createAttributeNS("image", "pattern")
        _child.setAttributeNodeNS(_attr)
        _child.setAttributeNS("image", "pattern", `_linetype.getList()`)
        map[_linetype] = _i
        _i = _i + 1

def _save_units(image, node, doc):
    _uts = doc.createElementNS("image", "image:Units")
    node.appendChild(_uts)
    _attr = doc.createAttributeNS("image", "unit")
    _uts.setAttributeNodeNS(_attr)
    _uts.setAttributeNS("image", "unit", `image.getUnits()`)

_entlist = [
    'segment',
    'circle',
    'arc',
    'chamfer',
    'fillet',
    'leader',
    ]

_conobjs = [
    'hcline',
    'vcline',
    'acline',
    'cline',
    'ccircle',
    ]

_dimlist = [
    'linear_dimension',
    'horizontal_dimension',
    'vertical_dimension',
    'radial_dimension',
    'angular_dimension',
    ]

def _get_image_attributes(image):
    _attmap = {}
    _styles = []
    _attmap['styles'] = _styles
    _colors = []
    _attmap['colors'] = _colors
    _linetypes = []
    _attmap['linetypes'] = _linetypes
    _fonts = []
    _attmap['fonts'] = _fonts
    _dimstyles = []
    _attmap['dimstyles'] = _dimstyles
    _check_conobjs = True
    _layers = []
    _layers.append(image.getTopLayer())
    while len(_layers):
        _layer = _layers.pop()
        for _ent in _entlist:
            for _obj in _layer.getLayerEntities(_ent):
                #
                # get style properties
                #
                _style = _obj.getStyle()
                if _style not in _styles:
                    _styles.append(_style)
                _color = _style.getColor() # style color
                if _color not in _colors:
                    _colors.append(_color)
                _linetype = _style.getLinetype() # style linetype
                if _linetype not in _linetypes:
                    _linetypes.append(_linetype)
                #
                # object may have overridden style values
                #
                _color = _obj.getColor()
                if _color not in _colors:
                    _colors.append(_color)
                _linetype = _obj.getLinetype()
                if _linetype not in _linetypes:
                    _linetypes.append(_linetype)
        if _check_conobjs:
            for _conobj in _conobjs:
                for _obj in _layer.getLayerEntities(_conobj):
                    _check_conobjs = False
                    _style = _obj.getStyle()
                    if _style not in _styles:
                        _styles.append(_style)
                    _color = _style.getColor()
                    if _color not in _colors:
                        _colors.append(_color)
                    _linetype = _style.getLinetype()
                    if _linetype not in _linetypes:
                        _linetypes.append(_linetype)
                    break
                if not _check_conobjs:
                    break
        for _dim in _dimlist:
            for _obj in _layer.getLayerEntities(_dim):
                #
                # get style properties
                #
                _ds = _obj.getDimStyle()
                if _ds not in _dimstyles:
                    _dimstyles.append(_ds)
                _color = _ds.getOption('DIM_COLOR')
                if _color not in _colors:
                    _colors.append(_color)
                _color = _ds.getOption('DIM_PRIMARY_FONT_COLOR')
                if _color not in _colors:
                    _colors.append(_color)
                _color = _ds.getOption('DIM_SECONDARY_FONT_COLOR')
                if _color not in _colors:
                    _colors.append(_color)
                _font = _ds.getOption('DIM_PRIMARY_FONT_FAMILY')
                if _font not in _fonts:
                    _fonts.append(_font)
                _font = _ds.getOption('DIM_SECONDARY_FONT_FAMILY')
                if _font not in _fonts:
                    _fonts.append(_font)
                #
                # object may have overridden style values
                #
                _color = _obj.getColor()
                if _color not in _colors: # dimension bar colors
                    _colors.append(_color)
                _pds = _obj.getPrimaryDimstring()
                _color = _pds.getColor() # primary dim text color
                if _color not in _colors:
                    _colors.append(_color)
                _font = _pds.getFamily() # primary dim font
                if _font not in _fonts:
                    _fonts.append(_font)
                _sds = _obj.getSecondaryDimstring()
                _color = _sds.getColor() # secondary dim color
                if _color not in _colors:
                    _colors.append(_color)
                _font = _sds.getFamily() # secondary dim font
                if _font not in _fonts:
                    _fonts.append(_font)
        for _child in _layer.getChildren():
            _layers.append(_child)
    # print "colors:"
    _colors.sort()
    # for _color in _colors:
        # print "Color: " + str(_color)
    # print "styles:"
    _styles.sort()
    # for _style in _styles:
        # print "Style: " + str(_style)
    # print "linetypes:"
    _linetypes.sort()
    # for _linetype in _linetypes:
        # print "Linetype: " + str(_linetype)
    # print "fonts:"
    _fonts.sort()
    # for _font in _fonts:
        # print "Font: " + str(_font)
    # print "dimstyles:"
    _dimstyles.sort()
    # for _dimstyle in _dimstyles:
        # print "Dimstyle: " + `_dimstyle`
    return _attmap
            
def save_image(image, filename=None):
    """Save a file.

save_image(image[, filename])

By default the file is saved to the filename stored in the image.
Passing a filename argument saves the image data to that location
instead.
    """
    if filename is None:
        filename = image.getFilename()
        if filename is None:
            raise ValueError, "No filename given for the image!"
        if not filename.endswith('.gz'):
            filename = filename + '.gz'
    if os.path.islink(filename):
        raise IOError, "No writing to symbolic links."
    fdir = os.path.dirname(filename)
    if not os.access(fdir, os.W_OK):
        raise IOError, "Unable to write into directory %s" % fdir
    backup = filename + '~'
    if os.path.exists(filename):
        try:
            os.unlink(backup)
        except OSError:
            pass
        os.rename(filename, backup)
    try:
        imfile = fileio.CompFile(filename, "w")
        try:
            doc = _make_xml_doc()
            root = doc.documentElement
            map = {}
            dimmap = {}
            #
            # to only save colors, linetypes, etc. actually used
            # in the drawing, the get_image_attributes() scans
            # the image and stores them and returns them in a
            # dictionary, then then image entities are written out.
            # this approach requires two passes over the image, and
            # it would be nice if these two passes could be combined
            # into a single pass.
            #
            _attmap = _get_image_attributes(image)
            _save_colors(_attmap, root, doc, map)
            _save_linetypes(_attmap, root, doc, map)
            _save_styles(_attmap, root, doc, map)
            _save_font_families(_attmap, root, doc, map)
            _save_dimstyles(_attmap, root, doc, map)
            _save_units(image, root, doc)
            lyrs = doc.createElementNS("image", "image:Layers")
            root.appendChild(lyrs)
            layers = []
            layers.append(image.getTopLayer())
            i = 0
            while(len(layers)):
                l = layers.pop()
                map[l] = i
                parent = l.getParent()
                child = doc.createElementNS("image", "image:Layer")
                lyrs.appendChild(child)
                attr = doc.createAttributeNS("image", "id")
                child.setAttributeNodeNS(attr)
                child.setAttributeNS("image", "id", `i`)
                attr = doc.createAttributeNS("image", "name")
                child.setAttributeNodeNS(attr)
                child.setAttributeNS("image", "name", l.getName())
                attr = doc.createAttributeNS("image", "scale")
                child.setAttributeNodeNS(attr)
                child.setAttributeNS("image", "scale", `l.getScale()`)
                parent = l.getParent()
                if parent is not None:
                    assert parent in map, "Layer %s parent not in map!" % l.getName()
                    attr = doc.createAttributeNS("image", "parent")
                    child.setAttributeNodeNS(attr)
                    child.setAttributeNS("image", "parent", `map[parent]`)
                _save_points(l, child, doc, map)
                _save_segments(l, child, doc, map)
                _save_circles(l, child, doc, map)
                _save_arcs(l, child, doc, map)
                _save_hclines(l, child, doc, map)
                _save_vclines(l, child, doc, map)
                _save_aclines(l, child, doc, map)
                _save_clines(l, child, doc, map)
                _save_ccircles(l, child, doc, map)
                _save_chamfers(l, child, doc, map)
                _save_fillets(l, child, doc, map)
                _save_leaders(l, child, doc, map)
                _save_linear_dims(l, child, doc, map, dimmap, 'linear_dimension')
                _save_linear_dims(l, child, doc, map, dimmap, 'horizontal_dimension')
                _save_linear_dims(l, child, doc, map, dimmap, 'vertical_dimension')
                _save_radial_dims(l, child, doc, map, dimmap)
                _save_angular_dims(l, child, doc, map, dimmap)
                i = i + 1
                if l.hasChildren():
                    layers = layers + l.getChildren()
            if len(dimmap):
                for elem in dimmap.keys():
                    print "fixing element: " + `elem`
                    for obj, attr in dimmap[elem]:
                        print "obj: " + `obj`
                        print "attr : " + `attr`
                        assert map.has_key(obj), "Missing object in map: " + `obj`
                        elem.setAttributeNS("image", attr, `map[obj]`)
            doc.writexml(imfile, "  ", "  ", "\n")
        finally:
            imfile.close()
    except IOError, e:
        os.unlink(filename)
        os.rename(backup, filename)
        raise IOError, "Unable to write new file: " + `e`
    except StandardError, e:
        print "Standard error: " + str(e)
            
#
# reload an image
#

def _apply_dim_overrides(dim, overrides):
    print "in _apply_dim_overrides()"
    for key in overrides:
        val = overrides[key]
        if key == 'DIM_OFFSET':
            dim.setOffset(val)
        elif key == 'DIM_EXTENSION':
            dim.setExtension(val)
        elif key == 'DIM_COLOR':
            dim.setColor(val)
        elif key == 'DIM_POSITION':
            dim.setPosition(val)
        elif key == 'DIM_ENDPOINT':
            dim.setEndpointType(val)
        elif key == 'DIM_ENDPOINT_SIZE':
            dim.setEndpointSize(val)
        elif key == 'DIM_DUAL_MODE':
            dim.setDualDimMode(val)
        elif key == 'DIM_PRIMARY_FONT_FAMILY':
            dim.getPrimaryDimstring().setFamily(val)
        elif key == 'DIM_PRIMARY_FONT_SIZE':
            dim.getPrimaryDimstring().setSize(val)
        elif key == 'DIM_PRIMARY_FONT_WEIGHT':
            dim.getPrimaryDimstring().setWeight(val)
        elif key == 'DIM_PRIMARY_FONT_STYLE':
            dim.getPrimaryDimstring().setStyle(val)
        elif key == 'DIM_PRIMARY_FONT_COLOR':
            dim.getPrimaryDimstring().setColor(val)
        elif (key == 'DIM_PRIMARY_PREFIX' or
              key == 'RADIAL_DIM_PRIMARY_PREFIX' or
              key == 'ANGULAR_DIM_PRIMARY_PREFIX'):
            dim.getPrimaryDimstring().setPrefix(val)
        elif (key == 'DIM_PRIMARY_SUFFIX' or
              key == 'RADIAL_DIM_PRIMARY_SUFFIX' or
              key == 'ANGULAR_DIM_PRIMARY_SUFFIX'):
            dim.getPrimaryDimstring().setSuffix(val)
        elif key == 'DIM_PRIMARY_PRECISION':
            dim.getPrimaryDimstring().setPrecision(val)
        elif key == 'DIM_PRIMARY_UNITS':
            dim.getPrimaryDimstring().setUnits(val)
        elif key == 'DIM_PRIMARY_LEADING_ZERO':
            dim.getPrimaryDimstring().setPrintZero(val)
        elif key == 'DIM_PRIMARY_TRAILING_DECIMAL':
            dim.getPrimaryDimstring().setPrintDecimal(val)
        elif key == 'DIM_SECONDARY_FONT_FAMILY':
            dim.getSecondaryDimstring().setFamily(val)
        elif key == 'DIM_SECONDARY_FONT_SIZE':
            dim.getSecondaryDimstring().setSize(val)
        elif key == 'DIM_SECONDARY_FONT_WEIGHT':
            dim.getSecondaryDimstring().setWeight(val)
        elif key == 'DIM_SECONDARY_FONT_STYLE':
            dim.getSecondaryDimstring().setStyle(val)
        elif key == 'DIM_SECONDARY_FONT_COLOR':
            dim.getSecondaryDimstring().setColor(val)
        elif (key == 'DIM_SECONDARY_PREFIX' or
              key == 'RADIAL_DIM_SECONDARY_PREFIX' or
              key == 'ANGULAR_DIM_SECONDARY_PREFIX'):
            dim.getSecondaryDimstring().setPrefix(val)
        elif (key == 'DIM_SECONDARY_SUFFIX' or
              key == 'RADIAL_DIM_SECONDARY_SUFFIX' or
              key == 'ANGULAR_DIM_SECONDARY_SUFFIX'):
            dim.getSecondaryDimstring().setSuffix(val)
        elif key == 'DIM_SECONDARY_PRECISION':
            dim.getSecondaryDimstring().setPrecision(val)
        elif key == 'DIM_SECONDARY_UNITS':
            dim.getSecondaryDimstring().setUnits(val)
        elif key == 'DIM_SECONDARY_LEADING_ZERO':
            dim.getSecondaryDimstring().setPrintZero(val)
        elif key == 'DIM_SECONDARY_TRAILING_DECIMAL':
            dim.getSecondaryDimstring().setPrintDecimal(val)
        elif key == 'RADIAL_DIM_DIA_MODE':
            dim.setDiaMode(val)
        else:
            pass # unused key ...
            
def _load_dim_options(node, map):
    print "in _load_dim_options()"
    options = {}
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:DimOption":
                opt = child.getAttribute("opt")
                if (opt == 'DIM_COLOR' or
                    opt == 'DIM_PRIMARY_FONT_COLOR' or
                    opt == 'DIM_SECONDARY_FONT_COLOR'):
                    cid = int(child.getAttribute("cid"))
                    assert cid in map['color'], "Color index missing: %d" % cid
                    val = map['color'][cid]
                elif (opt == 'DIM_PRIMARY_FONT_FAMILY' or
                      opt == 'DIM_SECONDARY_FONT_FAMILY'):
                    fid = int(child.getAttribute("fid"))
                    assert fid in map['family'], "Font index missing: %d" % fid
                    val = map['family'][fid]
                elif (opt == 'DIM_OFFSET' or
                      opt == 'DIM_ENDPOINT_SIZE' or
                      opt == 'DIM_EXTENSION'):
                    val = float(child.getAttribute("val"))
                elif (opt == 'DIM_PRIMARY_PRECISION' or
                      opt == 'DIM_SECONDARY_PRECISION' or
                      opt == 'DIM_PRIMARY_UNITS' or
                      opt == 'DIM_SECONDARY_UNITS' or
                      opt == 'DIM_PRIMARY_FONT_SIZE' or
                      opt == 'DIM_SECONDARY_FONT_SIZE' or
                      opt == 'DIM_PRIMARY_FONT_WEIGHT' or
                      opt == 'DIM_SECONDARY_FONT_WEIGHT' or
                      opt == 'DIM_PRIMARY_FONT_STYLE' or
                      opt == 'DIM_SECONDARY_FONT_STYLE' or
                      opt == 'DIM_POSITION' or
                      opt == 'DIM_ENDPOINT'):
                    val = int(child.getAttribute("val"))
                elif (opt == 'DIM_PRIMARY_LEADING_ZERO' or
                      opt == 'DIM_SECONDARY_LEADING_ZERO' or
                      opt == 'DIM_PRIMARY_TRAILING_DECIMAL' or
                      opt == 'DIM_SECONDARY_TRAILING_DECIMAL' or
                      opt == 'DIM_DUAL_MODE' or
                      opt == 'RADIAL_DIM_DIA_MODE'):
                    val = child.getAttribute("val")
                    if val == 0 or val == u'0' or val == u'False':
                        val = False
                    elif val == 1 or val == u'1' or val == u'True':
                        val = True
                else:
                    val = child.getAttribute("val")
                options[opt] = val
        child = child.nextSibling
    return options

def _load_linear_dimensions(node, map):
    print "in _load_linear_dimensions()"
    dims = []
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if (child.nodeName == "image:LDim" or
                child.nodeName == "image:HDim" or
                child.nodeName == "image:VDim"):
                dimopts = {}                
                id = int(child.getAttribute("id"))
                dimopts['id'] = id
                ds = int(child.getAttribute("ds"))
                dimopts['ds'] = ds
                assert ds in map['dimstyle'], "Missing style in map: " + `ds`
                l1 = int(child.getAttribute("l1"))
                dimopts['l1'] = l1
                p1 = int(child.getAttribute("p1"))
                dimopts['p1'] = p1
                l2 = int(child.getAttribute("l2"))
                dimopts['l2'] = l2
                p2 = int(child.getAttribute("p2"))
                dimopts['p2'] = p2
                x = float(child.getAttribute("x"))
                dimopts['x'] = x
                y = float(child.getAttribute("y"))
                dimopts['y'] = y
                overrides = _load_dim_options(child, map)
                for key in overrides:
                    print "override key: " + str(key)
                    val = overrides[key]
                    print "override val: " + str(val)
                    dimopts[key] = val
                for key in dimopts:
                    print "key: " + `key`
                    print "value: " + `dimopts[key]`
                dims.append(dimopts)
        child = child.nextSibling
    return dims

def _load_radial_dimensions(node, map):
    print "in _load_radial_dimensions()"
    rdims = []
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:RDim":
                dimopts = {}                
                id = int(child.getAttribute("id"))
                dimopts['id'] = id
                ds = int(child.getAttribute("ds"))
                dimopts['ds'] = ds
                assert ds in map['dimstyle'], "Missing style in map: " + `ds`
                l = int(child.getAttribute("l"))
                dimopts['l'] = l
                c = int(child.getAttribute("c"))
                dimopts['c'] = c
                x = float(child.getAttribute("x"))
                dimopts['x'] = x
                y = float(child.getAttribute("y"))
                dimopts['y'] = y
                overrides = _load_dim_options(child, map)
                for key in overrides:
                    val = overrides[key]
                    dimopts[key] = val
                for key in dimopts:
                    print "key: " + `key`
                    print "value: " + `dimopts[key]`
                rdims.append(dimopts)
        child = child.nextSibling
    return rdims

def _load_angular_dimensions(node, map):
    print "in _load_angular_dimensions()"
    adims = []
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:ADim":
                dimopts = {}                
                id = int(child.getAttribute("id"))
                dimopts['id'] = id
                ds = int(child.getAttribute("ds"))
                dimopts['ds'] = ds
                assert ds in map['dimstyle'], "Missing style in map: " + `ds`
                l1 = int(child.getAttribute("l1"))
                dimopts['l1'] = l1
                p1 = int(child.getAttribute("p1"))
                dimopts['p1'] = p1
                l2 = int(child.getAttribute("l2"))
                dimopts['l2'] = l2
                p2 = int(child.getAttribute("p2"))
                dimopts['p2'] = p2
                l3 = int(child.getAttribute("l3"))
                dimopts['l3'] = l3
                p3 = int(child.getAttribute("p3"))
                dimopts['p3'] = p3
                x = float(child.getAttribute("x"))
                dimopts['x'] = x
                y = float(child.getAttribute("y"))
                dimopts['y'] = y
                overrides = _load_dim_options(child, map)
                for key in overrides:
                    val = overrides[key]
                    dimopts[key] = val
                for key in dimopts:
                    print "key: " + `key`
                    print "value: " + `dimopts[key]`
                adims.append(dimopts)
        child = child.nextSibling
    return adims

def _set_graphic_attrs(obj, node, map):
    _ltid = node.getAttribute("linetype")
    if _ltid != "":
        _ltid = int(_ltid)
        assert _ltid in map['linetype'], "Object linetype index %d not in map: %s" % (_ltid, str(obj))
        _lt = map['linetype'][_ltid]
        obj.setLinetype(_lt)
    _cid = node.getAttribute("color")
    if _cid != "":
        _cid = int(_cid)
        assert _cid in map['color'], "Object color index %d not in map: %s" % (_cid, str(obj))
        _c = map['color'][_cid]
        obj.setColor(_c)
    _th = node.getAttribute("thickness")
    if _th != "":
        _th = float(_th)
        obj.setThickness(_th)

def _load_leaders(image, node, map):
    _active_layer = image.getActiveLayer()
    _child = node.firstChild
    while _child is not None:
        if _child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if _child.nodeName == "image:Leader":
                _id = int(_child.getAttribute("id"))
                _p1id = int(_child.getAttribute("p1"))
                assert _p1id in map[_active_layer]['point'], "Leader p1 index %d not in map!" % _p1id
                _p1 = map[_active_layer]['point'][_p1id]
                _p2id = int(_child.getAttribute("p2"))
                assert _p2id in map[_active_layer]['point'], "Leader p2 index %d not in map!" % _p2id
                _p2 = map[_active_layer]['point'][_p2id]
                _p3id = int(_child.getAttribute("p3"))
                assert _p3id in map[_active_layer]['point'], "Leader p3 index %d not in map!" % _p3id
                _p3 = map[_active_layer]['point'][_p3id]
                _sid = int(_child.getAttribute("style"))
                _size = float(_child.getAttribute("size"))
                assert _sid in map['style'], "Leader style index %d not in map!" % _sid
                _st = map['style'][_sid]
                _leader = leader.Leader(_p1, _p2, _p3, _size, _st)
                _set_graphic_attrs(_leader, _child, map)
                image.addObject(_leader)
        _child = _child.nextSibling

def _load_fillets(image, node, map):
    print "in load_fillets()"
    active_layer = image.getActiveLayer()
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:Fillet":
                id = int(child.getAttribute("id"))
                s1 = int(child.getAttribute("s1"))
                assert s1 in map[active_layer]['segment'], "Fillet segment s1 %d not in map!" % s1
                seg1 = map[active_layer]['segment'][s1]
                s2 = int(child.getAttribute("s2"))                
                assert s2 in map[active_layer]['segment'], "Fillet segment s2 %s not in map!" % s2
                seg2 = map[active_layer]['segment'][s2]
                sid = int(child.getAttribute("style"))
                assert sid in map['style'], "Fillet style index %d not in map!" % sid
                st = map['style'][sid]
                radius = float(child.getAttribute("radius"))
                fillet = segjoint.Fillet(seg1, seg2, radius, st)
                _set_graphic_attrs(fillet, child, map)
                image.addObject(fillet)
        child = child.nextSibling

def _load_chamfers(image, node, map):
    print "in load_chamfers()"
    active_layer = image.getActiveLayer()
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:Chamfer":
                id = int(child.getAttribute("id"))
                s1 = int(child.getAttribute("s1"))
                assert s1 in map[active_layer]['segment'], "Fillet segment s1 %d not in map!" % s1
                seg1 = map[active_layer]['segment'][s1]
                s2 = int(child.getAttribute("s2"))
                assert s2 in map[active_layer]['segment'], "Fillet segment s2 %s not in map!" % s2
                seg2 = map[active_layer]['segment'][s2]
                sid = int(child.getAttribute("style"))
                assert sid in map['style'], "Chamfer style index %d not in map!" % sid
                st = map['style'][sid]
                length = float(child.getAttribute("length"))
                chamfer = segjoint.Chamfer(seg1, seg2, length, st)
                _set_graphic_attrs(chamfer, child, map)
                image.addObject(chamfer)
        child = child.nextSibling
    
def _load_ccircles(image, node, map):
    active_layer = image.getActiveLayer()
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:CCircle":
                id = int(child.getAttribute("id"))
                cid = int(child.getAttribute("cp"))
                assert cid in map[active_layer]['point'], "CCircle center index %d not in map!" % cid
                cen = map[active_layer]['point'][cid]
                rad = float(child.getAttribute("r"))
                cc = ccircle.CCircle(cen, rad)
                image.addObject(cc)
        child = child.nextSibling

def _load_clines(image, node, map):
    active_layer = image.getActiveLayer()
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:CLine":
                id = int(child.getAttribute("id"))
                p1id = int(child.getAttribute("p1"))
                assert p1id in map[active_layer]['point'], "CLine p1 index %d not in map!" % p1id
                p1 = map[active_layer]['point'][p1id]
                p2id = int(child.getAttribute("p2"))
                assert p2id in map[active_layer]['point'], "CLine p2 index %d not in map!" % p2id
                p2 = map[active_layer]['point'][p2id]
                cl = cline.CLine(p1, p2)
                image.addObject(cl)
        child = child.nextSibling

def _load_aclines(image, node, map):
    active_layer = image.getActiveLayer()
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:ACLine":
                id = int(child.getAttribute("id"))
                lid = int(child.getAttribute("location"))
                angle = float(child.getAttribute("angle"))
                assert id in map[active_layer]['point'], "ACLine location index %d not in map!" % lid
                loc = map[active_layer]['point'][lid]
                acl = acline.ACLine(loc, angle)
                image.addObject(acl)
        child = child.nextSibling

def _load_vclines(image, node, map):
    active_layer = image.getActiveLayer()
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:VCLine":
                id = int(child.getAttribute("id"))
                lid = int(child.getAttribute("location"))
                assert id in map[active_layer]['point'], "VCLine location index %d not in map!" % lid
                loc = map[active_layer]['point'][lid]
                vcl = vcline.VCLine(loc)
                image.addObject(vcl)
        child = child.nextSibling

def _load_hclines(image, node, map):
    active_layer = image.getActiveLayer()
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:HCLine":
                id = int(child.getAttribute("id"))
                lid = int(child.getAttribute("location"))
                assert id in map[active_layer]['point'], "HCLine location index %d not in map!" % lid
                loc = map[active_layer]['point'][lid]
                hcl = hcline.HCLine(loc)
                image.addObject(hcl)
        child = child.nextSibling

def _load_arcs(image, node, map):
    active_layer = image.getActiveLayer()
    map[active_layer]['arc'] = {}    
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:Arc":
                id = int(child.getAttribute("id"))
                cid = int(child.getAttribute("cp"))
                assert cid in map[active_layer]['point'], "Arc center index %d not in map!" % cid
                cen = map[active_layer]['point'][cid]
                rad = float(child.getAttribute("r"))
                sa = float(child.getAttribute("sa"))
                ea = float(child.getAttribute("ea"))
                sid = int(child.getAttribute("style"))
                assert sid in map['style'], "Arc style index %d not in map!" % sid
                st = map['style'][sid]
                a = arc.Arc(cen, rad, sa, ea, st)
                _set_graphic_attrs(a, child, map)
                map[active_layer]['arc'][id] = a
                image.addObject(a)
        child = child.nextSibling

def _load_circles(image, node, map):
    active_layer = image.getActiveLayer()
    map[active_layer]['circle'] = {}    
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:Circle":
                id = int(child.getAttribute("id"))
                cid = int(child.getAttribute("cp"))
                assert cid in map[active_layer]['point'], "Circle center index %d not in map!" % cid
                cen = map[active_layer]['point'][cid]
                rad = float(child.getAttribute("r"))
                sid = int(child.getAttribute("style"))
                assert sid in map['style'], "Circle style index %d not in map!" % sid
                st = map['style'][sid]
                c = circle.Circle(cen, rad, st)
                _set_graphic_attrs(c, child, map)
                image.addObject(c)
                map[active_layer]['circle'][id] = c
        child = child.nextSibling

def _load_segments(image, node, map):
    active_layer = image.getActiveLayer()
    map[active_layer]['segment'] = {}
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:Segment":
                id = int(child.getAttribute("id"))
                p1id = int(child.getAttribute("p1"))
                assert p1id in map[active_layer]['point'], "Segment p1 index %d not in map!" % p1id
                p1 = map[active_layer]['point'][p1id]
                p2id = int(child.getAttribute("p2"))
                assert p2id in map[active_layer]['point'], "Segment p2 index %d not in map!" % p2id
                p2 = map[active_layer]['point'][p2id]

                sid = int(child.getAttribute("style"))
                assert sid in map['style'], "Segment style index %d not in map!" % sid
                st = map['style'][sid]
                s = segment.Segment(p1, p2, st)
                _set_graphic_attrs(s, child, map)
                image.addObject(s)
                map[active_layer]['segment'][id] = s
        child = child.nextSibling

def _load_points(image, node, map):
    map['point'] = {}
    active_layer = image.getActiveLayer()
    map[active_layer]['point'] = {}
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:Point":
                id = int(child.getAttribute("id"))
                x = float(child.getAttribute("x"))
                y = float(child.getAttribute("y"))
                p = point.Point(x,y)
                image.addObject(p)
                map[active_layer]['point'][id] = p
        child = child.nextSibling

def _load_layer(image, node, map, dimmap):
    active_layer = image.getActiveLayer()
    map[active_layer] = {}
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            name = child.nodeName
            if name == "image:Points":
                _load_points(image, child, map)
            elif name == "image:Segments":
                _load_segments(image, child, map)
            elif name == "image:Circles":
                _load_circles(image, child, map)
            elif name == "image:Arcs":
                _load_arcs(image, child, map)
            elif name == "image:HCLines":
                _load_hclines(image, child, map)
            elif name == "image:VCLines":
                _load_vclines(image, child, map)
            elif name == "image:ACLines":
                _load_aclines(image, child, map)
            elif name == "image:CLines":
                _load_clines(image, child, map)
            elif name == "image:CCircles":
                _load_ccircles(image, child, map)
            elif name == "image:Fillets":
                _load_fillets(image, child, map)
            elif name == "image:Chamfers":
                _load_chamfers(image, child, map)
            elif name == "image:Leaders":
                _load_leaders(image, child, map)
            elif name == "image:LDims":
                dims = _load_linear_dimensions(child, map)
                if len(dims):
                    if not dimmap.has_key(active_layer):
                        dimmap[active_layer] = {}
                    dimmap[active_layer]['linear'] = []
                    for dim in dims:
                        dimmap[active_layer]['linear'].append(dim)
            elif name == "image:HDims":
                dims = _load_linear_dimensions(child, map)
                if len(dims):
                    if not dimmap.has_key(active_layer):
                        dimmap[active_layer] = {}
                    dimmap[active_layer]['horiz'] = []
                    for dim in dims:
                        dimmap[active_layer]['horiz'].append(dim)
            elif name == "image:VDims":
                dims = _load_linear_dimensions(child, map)
                if len(dims):
                    if not dimmap.has_key(active_layer):
                        dimmap[active_layer] = {}
                    dimmap[active_layer]['vertical'] = []
                    for dim in dims:
                        dimmap[active_layer]['vertical'].append(dim)
            elif name == "image:RDims":
                dims = _load_radial_dimensions(child, map)
                if len(dims):
                    if not dimmap.has_key(active_layer):
                        dimmap[active_layer] = {}
                    dimmap[active_layer]['radial'] = []
                    for dim in dims:
                        dimmap[active_layer]['radial'].append(dim)
            elif name == "image:ADims":
                dims = _load_angular_dimensions(child, map)
                if len(dims):
                    if not dimmap.has_key(active_layer):
                        dimmap[active_layer] = {}
                    dimmap[active_layer]['angular'] = []
                    for dim in dims:
                        dimmap[active_layer]['angular'].append(dim)
        child = child.nextSibling

def _add_linear_dimension(dimdict, map, type, image, lyr):
    assert dimdict.has_key('id'), "Missing id key: " + `dimdict`
    assert dimdict.has_key('ds'), "Missing ds key: " + `dimdict`
    ds = dimdict['ds']
    assert map['dimstyle'].has_key(ds), "Missing ds key in map"
    ds = map['dimstyle'][dimdict['ds']]
    del dimdict['ds']                    
    assert dimdict.has_key('l1'), "Missing l1 key: " + `dimdict`
    l1 = dimdict['l1']
    del dimdict['l1']
    assert map['layer'].has_key(l1), "Missing l1 key in map"
    layer1 = map['layer'][l1]
    assert dimdict.has_key('p1'), "Missing p1 key: " + `dimdict`
    p1 = dimdict['p1']
    del dimdict['p1']
    assert map[layer1]['point'].has_key(p1), "Missing p1 key in map"
    point1 = map[layer1]['point'][p1]
    assert dimdict.has_key('l2'), "Missing l2 key: " + `dimdict`
    l2 = dimdict['l2']
    del dimdict['l2']
    assert map['layer'].has_key(l2), "Missing l2 key in map"
    layer2 = map['layer'][l2]
    assert dimdict.has_key('p2'), "Missing p2 key: " + `dimdict`
    p2 = dimdict['p2']
    del dimdict['p2']
    assert map[layer2]['point'].has_key(p2), "Missing p2 key in map"
    point2 = map[layer2]['point'][p2]
    assert dimdict.has_key('x'), "Missing x key: " + `dimdict`
    x = dimdict['x']
    del dimdict['x']
    assert dimdict.has_key('y'), "Missing y key: " + `dimdict`
    y = dimdict['y']
    del dimdict['y']
    dim = type(layer1, point1, layer2, point2, x, y, ds)
    if dimdict.has_key('id'):
        del dimdict['id']
    print "remaining keys ..."
    for key in dimdict:
        val = dimdict[key]
        print "key: " + `key`
        print "value: " + `val`
    if len(dimdict):
        _apply_dim_overrides(dim, dimdict)
    image.addObject(dim, lyr)

def _add_radial_dimension(dimdict, map, image, lyr):
    assert dimdict.has_key('id'), "Missing id key: " + `dimdict`
    assert dimdict.has_key('ds'), "Missing ds key: " + `dimdict`
    ds = dimdict['ds']
    assert map['dimstyle'].has_key(ds), "Missing ds key in map"
    ds = map['dimstyle'][dimdict['ds']]
    del dimdict['ds']                    
    assert dimdict.has_key('l'), "Missing l key: " + `dimdict`
    l = dimdict['l']
    del dimdict['l']
    assert map['layer'].has_key(l), "Missing l key in map"
    l1 = map['layer'][l]
    assert dimdict.has_key('c'), "Missing c key: " + `dimdict`
    c = dimdict['c']
    del dimdict['c']
    assert map[l1]['circle'].has_key(c), "Missing c key in map"
    c1 = map[l1]['circle'][c]
    assert dimdict.has_key('x'), "Missing x key: " + `dimdict`
    x = dimdict['x']
    del dimdict['x']
    assert dimdict.has_key('y'), "Missing y key: " + `dimdict`
    y = dimdict['y']
    del dimdict['y']
    dim = dimension.RadialDimension(l1, c1, x, y, ds)
    if dimdict.has_key('id'):
        del dimdict['id']
    print "remaining keys ..."
    for key in dimdict:
        val = dimdict[key]
        print "key: " + `key`
        print "value: " + `val`
    if len(dimdict):
        _apply_dim_overrides(dim, dimdict)
    image.addObject(dim, lyr)

def _add_angular_dimension(dimdict, map, image, lyr):
    assert dimdict.has_key('id'), "Missing id key: " + `dimdict`
    assert dimdict.has_key('ds'), "Missing ds key: " + `dimdict`
    ds = dimdict['ds']
    assert map['dimstyle'].has_key(ds), "Missing ds key in map"
    ds = map['dimstyle'][dimdict['ds']]
    del dimdict['ds']                    
    assert dimdict.has_key('l1'), "Missing l1 key: " + `dimdict`
    l1 = dimdict['l1']
    del dimdict['l1']
    assert map['layer'].has_key(l1), "Missing l1 key in map"
    layer1 = map['layer'][l1]
    assert dimdict.has_key('p1'), "Missing p1 key: " + `dimdict`
    p1 = dimdict['p1']
    del dimdict['p1']
    assert map[layer1]['point'].has_key(p1), "Missing p1 key in map"
    point1 = map[layer1]['point'][p1]
    assert dimdict.has_key('l2'), "Missing l2 key: " + `dimdict`
    l2 = dimdict['l2']
    del dimdict['l2']
    assert map['layer'].has_key(l2), "Missing l2 key in map"
    layer2 = map['layer'][l2]
    assert dimdict.has_key('p2'), "Missing p2 key: " + `dimdict`
    p2 = dimdict['p2']
    del dimdict['p2']
    assert map[layer2]['point'].has_key(p2), "Missing p2 key in map"
    point2 = map[layer2]['point'][p2]
    assert dimdict.has_key('l3'), "Missing l3 key: " + `dimdict`
    l3 = dimdict['l3']
    del dimdict['l3']
    assert map['layer'].has_key(l3), "Missing l3 key in map"
    layer3 = map['layer'][l3]
    assert dimdict.has_key('p3'), "Missing p3 key: " + `dimdict`
    p3 = dimdict['p3']
    del dimdict['p3']
    assert map[layer2]['point'].has_key(p3), "Missing p3 key in map"
    point3 = map[layer2]['point'][p3]
    assert dimdict.has_key('x'), "Missing x key: " + `dimdict`
    x = dimdict['x']
    del dimdict['x']
    assert dimdict.has_key('y'), "Missing y key: " + `dimdict`
    y = dimdict['y']
    del dimdict['y']
    dim = dimension.AngularDimension(layer1, point1, layer2, point2,
                                     layer3, point3, x, y, ds)
    if dimdict.has_key('id'):
        del dimdict['id']
    print "remaining keys ..."
    for key in dimdict:
        val = dimdict[key]
        print "key: " + `key`
        print "value: " + `val`
    if len(dimdict):
        _apply_dim_overrides(dim, dimdict)
    image.addObject(dim, lyr)

def _load_layers(image, node, map):
    map['layer'] = {}
    dimmap = {}
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:Layer":
                id = int(child.getAttribute("id"))
                name = str(child.getAttribute("name"))
                scale = float(child.getAttribute("scale"))
                pid = child.getAttribute("parent")
                if pid == "": # top layer created with image
                    l = image.getTopLayer()
                else:
                    pid = int(pid)
                    assert pid in map['layer'], "Layer parent id missing in map!" + `pid`
                    parent = map['layer'][pid]
                    l = layer.Layer(name)
                    image.addChildLayer(l,parent)
                l.setScale(scale)
                l.setName(name)
                image.setActiveLayer(l)
                map['layer'][id] = l
                _load_layer(image, child, map, dimmap)
        child = child.nextSibling
    if len(dimmap):
        for lyr in dimmap.keys():
            if dimmap[lyr].has_key('linear') and len(dimmap[lyr]['linear']):
                for dimdict in dimmap[lyr]['linear']:
                    _add_linear_dimension(dimdict, map, dimension.LinearDimension, image, lyr)
            if dimmap[lyr].has_key('horiz') and len(dimmap[lyr]['horiz']):
                for dimdict in dimmap[lyr]['horiz']:
                    _add_linear_dimension(dimdict, map, dimension.HorizontalDimension, image, lyr)
            if dimmap[lyr].has_key('vertical') and len(dimmap[lyr]['vertical']):
                for dimdict in dimmap[lyr]['vertical']:
                    _add_linear_dimension(dimdict, map, dimension.VerticalDimension, image, lyr)

            if dimmap[lyr].has_key('radial') and len(dimmap[lyr]['radial']):
                for dimdict in dimmap[lyr]['radial']:
                    _add_radial_dimension(dimdict, map, image, lyr)
            if dimmap[lyr].has_key('angular') and len(dimmap[lyr]['angular']):
                for dimdict in dimmap[lyr]['angular']:
                    _add_angular_dimension(dimdict, map, image, lyr)

def _load_dimstyles(image, node, map):
    map['dimstyle'] = {}
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:DimStyle":
                id = int(child.getAttribute("id"))
                name= child.getAttribute("name")
                dsopts = _load_dim_options(child, map)
                print "print keys ..."
                dskeys = dsopts.keys()
                dskeys.sort()
                for key in dskeys:
                    print "key: " + key + "; value: " + `dsopts[key]`
                dimstyle = dimension.DimStyle(name, dsopts)
                if image.hasDimStyle(dimstyle):
                    for ds in image.getAllDimStyles():
                        if dimstyle == ds:
                            map['dimstyle'][id] = ds
                else:
                    image.addDimStyle(dimstyle)
                    map['dimstyle'][id] = dimstyle
        child = child.nextSibling

def _load_font_families(image, node, map):
    map['family'] = {}
    child = node.firstChild
    families = image.getAllFonts()
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:FontFamily":
                id = int(child.getAttribute("id"))
                family = child.getAttribute("name")
                if family in families:
                    for f in families:
                        if f == family:
                            map['family'][id] = f
                else:
                    map['family'][id] = family
        child = child.nextSibling

def _load_styles(image, node, map):
    map['style'] = {}
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:Style":
                id = int(child.getAttribute("id"))
                cid = int(child.getAttribute("cid"))
                assert cid in map['color'], "Color index missing: %d" % cid
                c = map['color'][cid]
                ltid = int(child.getAttribute("ltid"))
                assert ltid in map['linetype'], "Linetype index missing :%d" % ltid
                lt = map['linetype'][ltid]
                name = child.getAttribute("name")
                t = float(child.getAttribute("thickness"))
                s = style.Style(name, lt, c, t)
                if image.hasStyle(s):
                    for st in image.getAllStyles():
                        if s == st:
                            map['style'][id] = st
                            break
                else:
                    image.addStyle(s)
                    map['style'][id] = s
        child = child.nextSibling

def _load_linetypes(image, node, map):
    map['linetype'] = {}
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:Linetype":
                id = int(child.getAttribute("id"))
                name = child.getAttribute("name")
                pat = child.getAttribute("pattern")
                if pat == "None":
                    pat = None
                else:
                    pat = eval(pat)
                lt = linetype.Linetype(name, pat)
                if image.hasLinetype(lt):
                    for l in image.getAllLinetypes():
                        if lt == l:
                            map['linetype'][id] = l
                            break
                else:
                    image.addLinetype(lt)
                    map['linetype'][id] = lt
        child = child.nextSibling

def _load_colors(image, node, map):
    map['color'] = {}
    child = node.firstChild
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            if child.nodeName == "image:Color":
                id = int(child.getAttribute("id"))
                r = int(child.getAttribute("r"))
                g = int(child.getAttribute("g"))
                b = int(child.getAttribute("b"))
                c = color.Color(r,g,b)
                if image.hasColor(c):
                    for col in image.getAllColors():
                        if c == col:
                            map['color'][id] = col
                else:
                    image.addColor(c)
                    map['color'][id] = c
        child = child.nextSibling

def load_image(image, filehandle):
    """Load an image from a filename or standard input

load_image(image, filehandle)

The argument image should be an empty image. The
filehandle argument should be an opened file object
that can be passed to the XML parser.
    """
    objmap = {}
    doc = xml.dom.minidom.parse(filehandle)
    root = doc.documentElement
    if root.nodeName != "image:Image":
        raise ValueError, "Expected 'image:Image' but got '%s' as document root!" % root.name
    child = root.firstChild
    while child.nodeType == xml.dom.Node.TEXT_NODE:
        child = child.nextSibling
    if child.nodeName != "image:Colors":
        raise ValueError, "Expected 'image:Colors' but got '%s'" % child.nodeName
    _load_colors(image, child, objmap)
    child = child.nextSibling
    while child.nodeType == xml.dom.Node.TEXT_NODE:
        child = child.nextSibling
    if child.nodeName != "image:Linetypes":
        raise ValueError, "Expected 'image:Linetypes' but got '%s'" % child.nodeName
    _load_linetypes(image, child, objmap)
    child = child.nextSibling
    while child.nodeType == xml.dom.Node.TEXT_NODE:
        child = child.nextSibling
    if child.nodeName != "image:Styles":
        raise ValueError, "Expected 'image:Styles' but got '%s'" % child.nodeName
    _load_styles(image, child, objmap)
    child = child.nextSibling
    while child is not None:
        if child.nodeType == xml.dom.Node.ELEMENT_NODE:
            name = child.nodeName
            if name == "image:Units":
                unit = int(child.getAttribute("unit"))
                image.setUnits(unit)
            elif name == "image:FontFamilies":
                _load_font_families(image, child, objmap)
            elif name == "image:DimStyles":
                _load_dimstyles(image, child, objmap)
            elif name == "image:Layers":
                _load_layers(image, child, objmap)
        child = child.nextSibling
