#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# The Layer class
#

import sys

import point
import segment
import circle
import arc
import hcline
import vcline
import acline
import cline
import ccircle
import segjoint
import leader
import dimension
import tolerance

class Layer(object):
    """The Layer class.

A Layer object holds all the various entities that can be
in a drawing. Each layer can have sublayers, and there is
no limit to the depth of the sublayering.

A Layer object has several attributes:

name: The Layer's name
parent: The parent Layer of the Layer
scale: The scale factor for object contained in the Layer
visibility: A flag to show the Layer's contents or not

A Layer object has the following methods:

{get/set}Name(): Get/Set the Layer's name.
{get/set}Parent(): Get/Set the Layer's parent.
hasChildren(): Test if the Layer has children.
getChildren(): Return the children of the Layer.
getEntities(): Return all of a type of entity in the Layer.
{add/del}Child(): Add/Remove a new child Layer to the Layer.
{add/del}Object(): Store/Remove a Point, Segment, etc. in the Layer.
delObjects(): Remove objects in the Layer.
usesObject(): List objects composed from other objects.
countUsers(): Return how many objects are using another object.
moveObject(): Move an object in the Layer.
moveObjects(): Move objects in the Layer.
findObject(): Return an object in the layer equivalent to a test object.
hasPoint(): See if the Layer has a Point object at a location.
hasPointAt(): Find a Point in the Layer.
hasCircleAt(): Find a Circle in the Layer.
extRef(): Add an external object reference to a Point in this Layer.
extDeref(): Remove an external object reference to a Point in this Layer.
getExtRefs(): Return a list of objects refering to an object in this Layer.
mapPoint(): See if a non-Point object in the layer crosses some location.
addLayer(): Add the contents of one layer to another.
{get/set}Visibility(): Get/Set the visibility of the Layer.
getBoundary(): Find the maximum and minimum coordinates of the Layer.
objsInRegion(): Return all the objects in the Layer that can be seen
                within some view.
update(): Check that all objects in the layer are stored properly.
    """

    def __init__(self, name=None):
        """Initializee a Layer.

Layer(name)

The 'name' argument is optional. It should be a string if used,
otherwise a default name "Layer" is assigned to the layer.
        """
        _n = name
        if _n is None:
            _n = u'Layer'
        if not isinstance(_n, unicode):
            if isinstance(_n, str):
                _n = unicode(_n)
            else:
                raise TypeError, "Invalid layer name: "+ `_n`
        self.__name = _n
        self.__parent = None
        self.__point_tree = point.PointTree()
        self.__segment_tree = segment.SegmentTree()
        self.__circle_tree = circle.CircleTree()
        self.__arc_tree = arc.ArcTree()
        self.__hcline_tree = hcline.HCLineTree()
        self.__vcline_tree = vcline.VCLineTree()
        self.__acline_tree = acline.ACLineTree()
        self.__cline_tree = cline.CLineTree()
        self.__ccircle_tree = ccircle.CCircleTree()
        self.__chamfers = []
        self.__fillets = []
        self.__leaders = []
        self.__linear_dimensions = []
        self.__horiz_dimensions = []
        self.__vert_dimensions = []
        self.__radial_dimensions = []
        self.__angular_dimensions = []
        self.__children = []
        self.__visibility = True
        self.__scale = 1.0
        #
        # self.__objdict keeps a reference to all objects stored in the layer
        #
        self.__objdict = {}
        #
        # self.__externals keeps reference to objects in this layer that are
        # used by objects stored in another layer
        #
        self.__externals = {}

    def __str__(self):
        if self.__parent is None:
            string = "Layer: %s [No Parent]" % self.__name
        else:
            string = "Layer: %s; Parent: %s" % (self.__name, self.__parent.__name)
        return string

    def __contains__(self, obj):
        """Find an object in the Layer.

This method permits the use of 'in' for test conditions.

if obj in layer:
    ....

This function tests for Point, Segment, Circle, etc. It returns
True if there is an equivalent object held in the Layer. Otherwise
the function returns False.
        """
        seen = False
        if obj in self.__objdict:
            seen = True
        if not seen:
            self.update()
            seen = False
            if isinstance(obj, point.Point):
                seen = obj in self.__point_tree
            elif isinstance(obj, segment.Segment):
                seen = obj in self.__segment_tree
            elif isinstance(obj, arc.Arc):
                seen = obj in self.__arc_tree
            elif isinstance(obj, circle.Circle):
                seen = obj in self.__circle_tree
            elif isinstance(obj, hcline.HCLine):
                seen = obj in self.__hcline_tree
            elif isinstance(obj, vcline.VCLine):
                seen = obj in self.__vcline_tree
            elif isinstance(obj, acline.ACLine):
                seen = obj in self.__acline_tree
            elif isinstance(obj, cline.CLine):
                seen = obj in self.__cline_tree
            elif isinstance(obj, ccircle.CCircle):
                seen = obj in self.__ccircle_tree
            elif isinstance(obj, segjoint.Fillet):
                seen = obj in self.__fillets
            elif isinstance(obj, segjoint.Chamfer):
                seen = obj in self.__chamfers
            elif isinstance(obj, leader.Leader):
                seen = obj in self.__leaders
            elif isinstance(obj, dimension.HorizontalDimension):
                seen = obj in self.__horiz_dimensions
            elif isinstance(obj, dimension.VerticalDimension):
                seen = obj in self.__vert_dimensions
            elif isinstance(obj, dimension.LinearDimension):
                seen = obj in self.__linear_dimensions
            elif isinstance(obj, dimension.RadialDimension):
                seen = obj in self.__radial_dimensions
            elif isinstance(obj, dimension.AngularDimension):
                seen = obj in self.__angular_dimensions
            else:
                raise TypeError, "Invalid type for in operation: " + `obj`
        return seen

    def clear(self):
        """Remove all the entities stored in this layer

clear()        
        """
        print "clearing layer: " + self.__name
        self.__point_tree.clear()
        self.__segment_tree.clear()
        self.__circle_tree.clear()
        self.__arc_tree.clear()
        self.__hcline_tree.clear()
        self.__vcline_tree.clear()
        self.__acline_tree.clear()                
        self.__cline_tree.clear()
        self.__ccircle_tree.clear()
        del self.__chamfers[:]
        del self.__fillets[:]
        del self.__leaders[:]
        del self.__linear_dimensions[:]
        del self.__horiz_dimensions[:]
        del self.__vert_dimensions[:]
        del self.__radial_dimensions[:]
        del self.__angular_dimensions[:]
        self.__visibility = True
        self.__scale = 1.0
        self.__objdict.clear()
        self.__externals.clear()

    def free(self):
        """Disconnect a layer from its parent.

free()
        """
        print "freeing layer: " + self.__name
        parent = self.__parent
        if parent is not None:
            parent.delChild(self)
        
    def getName(self):
        """Return the name of the Layer.

getName()
        """
        return self.__name

    def setName(self, name):
        """Set the name of the Layer.

setName(name)

The new must be a string, and cannot be None.
        """
        _n = name
        if _n is None:
            raise ValueError, "Layers must have a name."
        if not isinstance(_n, unicode):
            if isinstance(_n, str):
                _n = unicode(_n)
            else:
                raise TypeError, "Invalid name: " + `_n`
        self.__name = _n

    name = property(getName, setName, None, "Layer name.")

    def addLayer(self, l):
        """Add one Layer to another.

addLayer(l)

Add the contents of the Layer object "l" into the Layer. Only
the objects not found in the current Layer are added.
        """
        if not isinstance(l, Layer):
            raise TypeError, "Invalid Layer addition object: " + `l`
        objmap = {}
        self.update()
        l.update()
        for obj in l.__point_tree:
            sobj = self.__point_tree.binscan(obj)
            if sobj is None:
                self.addObject(obj)
            else:
                objmap[obj] = sobj
        for obj in l.__segment_tree:
            sobj = self.__segment_tree.binscan(obj)
            if sobj is None:
                p1, p2 = obj.getEndpoints()
                if p1 in objmap:
                    p1 = objmap[p1]
                if p2 in objmap:
                    p2 = objmap[p2]
                ns = segment.Segment(p1, p2, obj.getStyle(),
                                     obj.getLinetype(), obj.getColor(),
                                     obj.getThickness())
                self.addObject(ns)
            else:
                objmap[obj] = sobj
        for obj in l.__circle_tree:
            sobj = self.__circle_tree.binscan(obj)
            if sobj is None:
                cp = obj.getCenter()
                if cp in objmap:
                    cp = objmap[cp]
                nc = circle.Circle(cp, obj.getRadius(), obj.getStyle(),
                                   obj.getLinetype(), obj.getColor(),
                                   obj.getThickness())
                self.addObject(ns)
            else:
                objmap[obj] = sobj
        for obj in l.__arc_tree:
            sobj = self.__arc_tree.binscan(obj)
            if sobj is None:
                cp = obj.getCenter()
                if cp in objmap:
                    cp = objmap[cp]
                na = arc.Arc(cp, obj.getRadius(), obj.getStartAngle(),
                             obj.getEndAngle(), obj.getStyle(),
                             obj.getLinetype(), obj.getColor(),
                             obj.getThickness())
                self.addObject(na)
            else:
                objmap[obj] = sobj
        for obj in l.__hcline_tree:
            sobj = self.__hcline_tree.binscan(obj)
            if sobj is None:
                lp = obj.getLocation()
                if lp in objmap:
                    lp = objmap[lp]
                self.addObject(hcline.HCLine(lp))
        for obj in l.__vcline_tree:
            sobj = self.__vcline_tree.binscan(obj)
            if sobj is None:
                lp = obj.getLocation()
                if lp in objmap:
                    lp = objmap[lp]
                self.addObject(vcline.VCLine(lp))
        for obj in l.__acline_tree:
            sobj = self.__acline_tree.binscan(obj)
            if sobj is None:
                lp = obj.getLocation()
                if lp in objmap:
                    lp = objmap[lp]
                self.addObject(acline.ACLine(lp, obj.getAngle()))
        for obj in l.__cline_tree:
            sobj = self.__cline_tree.binscan(obj)
            if sobj is None:
                p1, p2 = obj.getKeypoints()
                if p1 in objmap:
                    p1 = objmap[p1]
                if p2 in objmap:
                    p2 = objmap[p2]
                self.addObject(cline.CLine(p1,p2))
        for obj in l.__ccircle_tree:
            sobj = self.__ccircle_tree.binscan(obj)
            if sobj is None:
                cp = obj.getCenter()
                if cp in objmap:
                    cp = objmap[cp]
                self.addObject(ccircle.CCircle(cp, obj.getRadius()))
        for obj in l.__chamfers:
            s1, s2 = obj.getSegments()
            if s1 in objmap:
                s1 = objmap[s1]
            if s2 in objmap:
                s2 = objmap[s2]
            nc = segjoint.Chamfer(s1, s2, obj.getLength(),
                                  obj.getStyle(), obj.getLinetype(),
                                  obj.getColor(), obj.getThickness())
            flag = True
            for lc in self.__chamfers:
                if nc == lc:
                    flag = False
                break
            if flag:
                self.addObject(nc)
        for obj in l.__fillets:
            s1, s2 = obj.getSegments()
            if s1 in objmap:
                s1 = objmap[s1]
            if s2 in objmap:
                s2 = objmap[s2]
            nf = segjoint.Fillet(s1, s2, obj.getRadius(),
                                 obj.getStyle(), obj.getLinetype(),
                                 obj.getColor(), obj.getThickness())
            flag = True
            for lf in self.__fillets:
                if nf == lf:
                    flag = False
                break
            if flag:
                self.addObject(nf)
        for _obj in l.getLayerEntities("leader"):
            if _obj not in self.__leaders:
                self.addObject(obj)
        for obj in l.__linear_dimensions:
            if obj not in self.__linear_dimensions:
                self.addObject(obj)
        for obj in l.__horiz_dimensions:
            if obj not in self.__horiz_dimensions:
                self.addObject(obj)
        for obj in l.__vert_dimensions:
            if obj not in self.__vert_dimensions:
                self.addObject(obj)
        for obj in l.__radial_dimensions:
            if obj not in self.__radial_dimensions:
                self.addObject(obj)
        for obj in l.__angular_dimensions:
            if obj not in self.__angular_dimensions:
                self.addObject(obj)

    def addObject(self, obj):
        """Add an object to this Layer.

addObject(obj)

The object should be a Point, Segment, Arc, Circle,
HCLine, VCLine, ACLine, CLine, CCircle, Chamfer,
Fillet, Leader, or Dimension. Anything else raises
a TypeError exception.
        """
        if obj in self.__objdict:
            raise ValueError, "Object already in layer: " + `obj`
        self.update()
        if isinstance(obj, point.Point):
            self._addPoint(obj)
        elif isinstance(obj, segment.Segment):
            self._addSegment(obj)
        elif isinstance(obj, arc.Arc):
            self._addArc(obj)
        elif isinstance(obj, circle.Circle):
            self._addCircle(obj)
        elif isinstance(obj, hcline.HCLine):
            self._addHCLine(obj)
        elif isinstance(obj, vcline.VCLine):
            self._addVCLine(obj)
        elif isinstance(obj, acline.ACLine):
            self._addACLine(obj)
        elif isinstance(obj, ccircle.CCircle):
            self._addCCircle(obj)
        elif isinstance(obj, cline.CLine):
            self._addCLine(obj)
        elif isinstance(obj, segjoint.Chamfer):
            self._addChamfer(obj)
        elif isinstance(obj, segjoint.Fillet):
            self._addFillet(obj)
        elif isinstance(obj, leader.Leader):
            self._addLeader(obj)
        elif isinstance(obj, dimension.AngularDimension):
            self._addAngularDimension(obj)
        elif isinstance(obj, dimension.RadialDimension):
            self._addRadialDimension(obj)
        elif isinstance(obj, dimension.HorizontalDimension):
            self._addHorizontalDimension(obj)
        elif isinstance(obj, dimension.VerticalDimension):
            self._addVerticalDimension(obj)
        elif isinstance(obj, dimension.LinearDimension):
            self._addLinearDimension(obj)
        else:
            raise TypeError, "Invalid object for storage: " + `obj`
            
    def _addPoint(self, p):
        """Add a Point object to the Layer.

_addPoint(p)

This method is private to the Layer object.
        """
        if p not in self.__point_tree:
            self.__point_tree.store(p)
            self.__objdict[p] = None

    def _addSegment(self, s):
        """Add a Segment object to the Layer.

_addSegment(s)

The Segment object "s" is added to the Layer if there is not
already a Segment in the layer between the same locations.
The Segment endpoints p1 and p2 _must_ be already stored in
this layer, otherwise a ValueError is raised.

This method is private to the Layer object.
        """
        if s not in self.__segment_tree:
            _p1, _p2 = s.getEndpoints()
            if _p1 not in self.__objdict:
                raise ValueError, "Segment p1 Point not found in Layer."
            if _p2 not in self.__objdict:
                raise ValueError, "Segment p2 Point not found in Layer."
            self.__segment_tree.store(s)
            self.__objdict[s] = None
            self._connectUser(_p1, s)
            self._connectUser(_p2, s)

            
    def _addCircle(self, c):
        """Add a Circle object to the Layer.

_addCircle(c)

The Circle object "c" is added to the Layer if there is not
already a Circle at that location with the same radius. The
Circle center must already be stored in the Layer, otherwise
a ValueError is raised.

This method is private to the layer object.
        """
        if c not in self.__circle_tree:
            _cp = c.getCenter()
            if _cp not in self.__objdict:
                raise ValueError, "Circle center Point not found in Layer." 
            self.__circle_tree.store(c)
            self.__objdict[c] = None
            self._connectUser(_cp, c)
            
    def _addArc(self, a):
        """Add an Arc object to the Layer.

_addArc(a)

The Arc object "a" is added to the Layer if there is not
already a Arc at that location with the same radius, start
angle, and eng_angle. The Arc center must already be stored
in the Layer, otherwise a ValueError is raised.

This method is private to the Layer object.
        """
        if a not in self.__arc_tree:
            _cp = a.getCenter()
            if _cp not in self.__objdict:
                raise ValueError, "Arc center Point not found in Layer."
            self.__arc_tree.store(a)
            self.__objdict[a] = None
            self._connectUser(_cp, a)
            for _ep in a.getEndpoints():
                _lp = self.hasPoint(_ep)
                if _lp is None:
                    _lp = point.Point(_ep)
                    self.addObject(_lp)
                self._connectUser(_lp, a)
                
    def _addHCLine(self, hcl):
        """Add an HCLine object to the Layer.

_addHCLine(hcl)

The HCLine object "hcl" is added to the Layer if there is not
already a HCLine at that location. The HCLine location point
must already be stored in the Layer, otherwise a ValueError is
raised.

This method is private the the Layer object.
        """
        if hcl not in self.__hcline_tree:
            _lp = hcl.getLocation()
            if _lp not in self.__objdict:
                raise ValueError, "HCLine location Point not found in Layer."
            self.__hcline_tree.store(hcl)
            self.__objdict[hcl] = None
            self._connectUser(_lp, hcl)
            
    def _addVCLine(self, vcl):
        """Add an VCLine object to the Layer.

_addVCLine(vcl)

The VCLine object "vcl" is added to the Layer if there is not
already a VCLine at that location. The VCLine location point
must already be stored in the Layer, otherwise a ValueError is
raised.

This method is private to the Layer object.
        """
        if vcl not in self.__vcline_tree:
            _lp = vcl.getLocation()
            if _lp not in self.__objdict:
                raise ValueError, "VCLine location Point not found in Layer."
            self.__vcline_tree.store(vcl)
            self.__objdict[vcl] = None
            self._connectUser(_lp, vcl)
            
    def _addACLine(self, acl):
        """Add an ACLine object to the Layer.

_addACLine(acl)

The ACLine object "acl" is added to the Layer if there is not
already a ACLine at that location. The ACLine location point
must already be stored in the Layer, otherwise a ValueError is
raised.

This method is private to the Layer object.
        """
        if acl not in self.__acline_tree:
            _lp = acl.getLocation()
            if _lp not in self.__objdict:
                raise ValueError, "ACLine location Point not found in Layer."
            self.__acline_tree.store(acl)
            self.__objdict[acl] = None
            self._connectUser(_lp, acl)

    def _addCCircle(self, cc):
        """Add an CCircle object to the Layer.

_addCCircle(cc)

The CCircle object "cc" is added to the Layer if there is not
already a CCircle at that location and with the same radius.
The CCircle center must already be stored in the Layer, otherwise
a ValueError is raised.

This method is private to the Layer object.
        """
        if cc not in self.__ccircle_tree:
            _cp = cc.getCenter()
            if _cp not in self.__objdict:
                raise ValueError, "CCircle center Point not found in Layer."
            self.__ccircle_tree.store(cc)
            self.__objdict[cc] = None
            self._connectUser(_cp, cc)

    def _addCLine(self, cl):
        """Add an CLine object to the Layer.

_addCLine(cl)

The CLine object "cl" is added to the Layer if there is not
already a CLine at that location. The CLine location point
must already be stored in the Layer, otherwise a ValueError is
raised.

This method is private to the Layer object.
        """
        if cl not in self.__cline_tree:
            _p1, _p2 = cl.getKeypoints()
            if _p1 not in self.__objdict:
                raise ValueError, "CLine p1 Point not found in Layer."
            if _p2 not in self.__objdict:
                raise ValueError, "CLine p2 Point not found in Layer."
            self.__cline_tree.store(cl)
            self.__objdict[cl] = None
            self._connectUser(_p1, cl)
            self._connectUser(_p2, cl)


    def _addChamfer(self, ch):
        """Add a Chamfer object to the Layer.

_addChamfer(ch)

The Chamfer object "ch" is added to the Layer if there is not
already a Chamfer found joining the two Segments. Both Segment
objects in the chamfer must also be stored in this Layer.

This method is private to the Layer object.
        """
        if ch not in self.__chamfers:
            _s1, _s2 = ch.getSegments()
            if _s1 not in self.__objdict:
                raise ValueError, "Chamfer s1 Segment not found in Layer."
            if _s2 not in self.__objdict:
                raise ValueError, "Chamfer s2 Segment not found in Layer."
            _new_chamfer = True
            for _ch in self.__chamfers:
                if _ch == ch:
                    _new_chamfer = False
                    break
            if _new_chamfer:
                self.__objdict[ch] = None
                self.__chamfers.append(ch)
                self._connectUser(_s1, ch)
                self._connectUser(_s2, ch)

    def _addFillet(self, f):
        """Add a Fillet object to the Layer.

_addFillet(f)

The Fillet object "f" is added to the Layer if there is not
already a Fillet found joining the two Segments. Both Segment
objects in the fillet must also be stored in this Layer.

This method is private to the Layer object.
        """
        if f not in self.__fillets:
            _s1, _s2 = f.getSegments()
            if _s1 not in self.__objdict:
                raise ValueError, "Fillet s1 Segment not found in Layer."
            if _s2 not in self.__objdict:
                raise ValueError, "Fillet s2 Segment not found in Layer."
            _new_fillet = True
            for _f in self.__fillets:
                if _f == f:
                    _new_fillet = False
                    break
            if _new_fillet:
                self.__objdict[f] = None
                self.__fillets.append(f)
                self._connectUser(_s1, f)
                self._connectUser(_s2, f)

    def _addLeader(self, l):
        """Add a Leader object to the Layer.

_addLeader(l)

The Leader object "l" is added to the Layer if there is not
already a Leader in the layer between the same locations.
The Leader endpoints p1, p2, and p3 _must_ be already stored in
this layer, otherwise a ValueError is raised.

This method is private to the Layer object.
        """
        if l not in self.__leaders:
            _p1, _p2, _p3 = l.getLeaderPoints()
            if _p1 not in self.__objdict:
                raise ValueError, "Leader p1 Point not found in Layer."
            if _p2 not in self.__objdict:
                raise ValueError, "Leader p2 Point not found in Layer."
            if _p3 not in self.__objdict:
                raise ValueError, "Leader p3 Point not found in Layer."
            self.__leaders.append(l)
            self.__objdict[l] = None
            self._connectUser(_p1, l)
            self._connectUser(_p2, l)
            self._connectUser(_p3, l)

    def _addAngularDimension(self, adim):
        """Add an AngularDimension object to the Layer.

_addAngularDimension(adim)

This method is private to the Layer object.
        """
        if adim not in self.__angular_dimensions:
            _layer1, _layer2, _layer3 = adim.getDimLayers()
            _p1, _p2, _p3 = adim.getDimPoints()
            _lp = _layer1.findObject(_p1)
            if _lp is not _p1:
                raise ValueError, "Dimension Point p1 not found in layer"
            _lp = _layer2.findObject(_p2)
            if _lp is not _p2:
                raise ValueError, "Dimension Point p2 not found in layer!"
            _lp = _layer3.findObject(_p3)
            if _lp is not _p3:
                raise ValueError, "Dimension Point p3 not found in layer!"
            self.__objdict[adim] = None            
            if _layer1 is self:
                self._connectUser(_p1, adim)
            else:
                _layer1.extRef(_p1, adim, self)
            if _layer2 is self:
                self._connectUser(_p2, adim)
            else:
                _layer2.extRef(_p2, adim, self)
            if _layer3 is self:
                self._connectUser(_p3, adim)
            else:
                _layer3.extRef(_p3, adim, self)
            self.__angular_dimensions.append(adim)

    def _addRadialDimension(self, rdim):
        """Add a RadialDimension object to the Layer.

_addRadialDimension(rdim)

This method is private to the Layer object.
        """
        if rdim not in self.__radial_dimensions:
            _layer = rdim.getDimLayer()
            _circle = rdim.getDimCircle()
            _lc = _layer.findObject(_circle)
            if _lc is not _circle:
                raise ValueError, "RadialDimension circular object not found in layer"
            self.__objdict[rdim] = None
            if _layer is self:
                self._connectUser(_circle, rdim)
            else:
                _layer.extRef(_circle, rdim, self)
            self.__radial_dimensions.append(rdim)

    def _addHorizontalDimension(self, hdim):
        """Add a HorizontalDimension object to the Layer.

_addHorizontalDimension(hdim)

This method is private to the Layer object.
        """
        if hdim not in self.__horiz_dimensions:
            _layer1, _layer2 = hdim.getDimLayers()
            _p1, _p2 = hdim.getDimPoints()
            _lp = _layer1.findObject(_p1)
            if _lp is not _p1:
                raise ValueError, "HorizontalDimension Point p1 not found in layer"
            _lp = _layer2.findObject(_p2)
            if _lp is not _p2:
                raise ValueError, "HorizontalDimension Point p2 not found in layer"
            self.__objdict[hdim] = None
            if _layer1 is self:
                self._connectUser(_p1, hdim)
            else:
                _layer1.extRef(_p1, hdim, self)
            if _layer2 is self:
                self._connectUser(_p2, hdim)
            else:
                _layer2.extRef(_p2, hdim, self)
            self.__horiz_dimensions.append(hdim)

    def _addVerticalDimension(self, vdim):
        """Add a VerticalDimension object to the Layer.

_addVerticalDimension(vdim)

This method is private to the Layer object.
        """
        if vdim not in self.__vert_dimensions:
            _layer1, _layer2 = vdim.getDimLayers()
            _p1, _p2 = vdim.getDimPoints()
            _lp = _layer1.findObject(_p1)
            if _lp is not _p1:
                raise ValueError, "VerticalDimension Point p1 not found in layer"
            _lp = _layer2.findObject(_p2)
            if _lp is not _p2:
                raise ValueError, "VerticalDimension Point p2 not found in layer"
            self.__objdict[vdim] = None
            if _layer1 is self:
                self._connectUser(_p1, vdim)
            else:
                _layer1.extRef(_p1, vdim, self)
            if _layer2 is self:
                self._connectUser(_p2, vdim)
            else:
                _layer2.extRef(_p2, vdim, self)
            self.__vert_dimensions.append(vdim)

    def _addLinearDimension(self, ldim):
        """Add a LinearDimension object to the Layer.

_addLinearDimension(ldim)

This method is private to the Layer object.
        """
        if ldim not in self.__linear_dimensions:
            _layer1, _layer2 = ldim.getDimLayers()
            _p1, _p2 = ldim.getDimPoints()
            _lp = _layer1.findObject(_p1)
            if _lp is not _p1:
                raise ValueError, "LinearDimension Point p1 not found in layer"
            _lp = _layer2.findObject(_p2)
            if _lp is not _p2:
                raise ValueError, "LinearDimension Point p2 not found in layer"
            self.__objdict[ldim] = None
            if _layer1 is self:
                self._connectUser(_p1, ldim)
            else:
                _layer1.extRef(_p1, ldim, self)
            if _layer2 is self:
                self._connectUser(_p2, ldim)
            else:
                _layer2.extRef(_p2, ldim, self)
            self.__linear_dimensions.append(ldim)

    def delObjects(self, objlist):
        """Remove a list of objects from this Layer.

delObjects(objlist)

The objlist argument must be a list containing objects
stored within this layer.
        """
        if not isinstance(objlist, list):
            raise TypeError, "Invalid object list: " + `objlist`
        for _obj in objlist:
            if _obj not in self.__objdict:
                raise ValueError, "Object not found in layer: " + `_obj`
        self.update()
        _removed = {}
        for _obj in objlist:
            if _obj not in _removed:
                _removed[_obj] = False
            if isinstance(_obj, point.Point):
                if _obj in self.__externals:
                    for _layer, _extobj in self.__externals[_obj]:
                        _layer.delObject(_extobj)
            elif isinstance(_obj, segment.Segment):
                _users = self.__objdict[_obj]
                if _users is not None:
                    for _user in _users:
                        self.delObject(_user)
                        _removed[_user] = True
                _p1, _p2 = _obj.getEndpoints()
                _removed[_p1] = True
                _removed[_p2] = True
            elif isinstance(_obj, arc.Arc):
                _users = self.__objdict[_obj]
                if _users is not None:
                    for _user in _users:
                        self.delObject(_user)
                        _removed[_user] = True
                _cp = _obj.getCenter()
                _removed[_cp] = True
            elif isinstance(_obj, (circle.Circle, ccircle.CCircle)):
                _users = self.__objdict[_obj]
                if _users is not None:
                    for _user in _users:
                        self.delObject(_user)
                        _removed[_user] = True
                _cp = _obj.getCenter()
                _removed[_cp] = True
            elif isinstance(_obj, (hcline.HCLine, vcline.VCLine, acline.ACLine)):
                _lp = _obj.getLocation()
                _removed[_lp] = True
            elif isinstance(_obj, cline.CLine):
                _p1, _p2 = _obj.getKeypoints()
                _removed[_p1] = True
                _removed[_p2] = True
            elif isinstance(_obj, (segjoint.Chamfer, segjoint.Fillet)):
                _users = self.__objdict[_obj]
                if _users is not None: # users are hatch paths
                    for _user in _users:
                        self.delObject(_user)
                        _removed[_user] = True
            elif isinstance(_obj, leader.Leader):
                _p1, _p2, _p3 = _obj.getLeaderPoints()
                _removed[_p1] = True
                _removed[_p2] = True
                _removed[_p3] = True
        for _obj in _removed:
            if not _removed[_obj]:
                self.delObject(_obj)

    def delObject(self, obj):
        """Remove an object from this Layer.

delObject(obj)

The object should be a Point, Segment, Arc, Circle,
HCLine, VCLine, ACLine, CLine, CCircle, Chamfer,
Fillet, Leader, or Dimension. Anything else raises
a TypeError exception.
        """
        if obj not in self.__objdict:
            raise ValueError, "Object not found in layer: " + `obj`
        self.update()
        if isinstance(obj, point.Point):
            self._delPoint(obj)
        elif isinstance(obj, segment.Segment):
            self._delSegment(obj)
        elif isinstance(obj, arc.Arc):
            self._delArc(obj)
        elif isinstance(obj, circle.Circle):
            self._delCircle(obj)
        elif isinstance(obj, hcline.HCLine):
            self._delHCLine(obj)
        elif isinstance(obj, vcline.VCLine):
            self._delVCLine(obj)
        elif isinstance(obj, acline.ACLine):
            self._delACLine(obj)
        elif isinstance(obj, cline.CLine):
            self._delCLine(obj)
        elif isinstance(obj, ccircle.CCircle):
            self._delCCircle(obj)
        elif isinstance(obj, segjoint.Chamfer):
            self._delChamfer(obj)
        elif isinstance(obj, segjoint.Fillet):
            self._delFillet(obj)
        elif isinstance(obj, leader.Leader):
            self._delLeader(obj)
        elif isinstance(obj, dimension.AngularDimension):
            self._delAngularDimension(obj)
        elif isinstance(obj, dimension.RadialDimension):
            self._delRadialDimension(obj)
        elif isinstance(obj, dimension.HorizontalDimension):
            self._delHorizontalDimension(obj)
        elif isinstance(obj, dimension.VerticalDimension):
            self._delVerticalDimension(obj)
        elif isinstance(obj, dimension.LinearDimension):
            self._delLinearDimension(obj)
        else:
            raise TypeError, "Invalid object for removal: " + `obj`

    def _delPoint(self, p):
        """Delete a Point from the Layer.

_delPoint(p)

This method is private to the Layer object.
        """
        _delete = True
        _users = self.__objdict[p]
        if _users is not None:
            for _user in _users:
                if not isinstance(_user, dimension.Dimension):
                    _delete = False
                    break
        if _delete:
            if p in self.__externals:
                for _layer, _extobj in self.__externals[p]:
                    _layer.delObject(_extobj)
            if _users is not None:
                _dims = _users[:] # copy _users - delObject() modifies it!
                for _dim in _dims:
                    self.delObject(_dim)
            del self.__objdict[p]
            self.__point_tree.remove(p)

    def  _connectUser(self, obj, user):
        """Store a reference between an object and its user in the objdict.

_connectUser(obj, user)

This method is private to the Layer.        
        """
        if obj in self.__objdict:
            if self.__objdict[obj] is None:
                self.__objdict[obj] = []
            _users = self.__objdict[obj]
            assert user in self.__objdict, "user not in __objdict"
            _seen = False
            for _user in _users:
                if _user is user:
                    _seen = True
                    break
            assert not _seen, "user already referenced by obj"
            _users.append(user)
            
    def _disconnectUser(self, obj, user):
        """Remove the linkage in the objdict of user of some object

_disconnectUser(obj, user)

This method is private to the Layer.
        """
        if obj in self.__objdict:
            _users = self.__objdict[obj]
            for _i in xrange(len(_users)):
                if user is _users[_i]:
                    del _users[_i]
                    break
            if not len(_users):
                self.__objdict[obj] = None

    def _delSegment(self, s):
        """Delete a Segment from the Layer.

_delSegment(s)

This method is private to the Layer object.
        """
        _users = self.__objdict[s]
        if _users is not None:
            for _obj in _users:
                self.delObject(_obj) # restore other segments?
        _p1, _p2 = s.getEndpoints()
        assert _p1 in self.__objdict, "Segment p1 Point not in objdict"
        self._disconnectUser(_p1, s)
        assert _p2 in self.__objdict, "Segment p2 Point not in objdict"
        self._disconnectUser(_p2, s)
        del self.__objdict[s]
        self.__segment_tree.remove(s)
        self.delObject(_p1) # remove possibly unused point _p1
        self.delObject(_p2) # remove possibly unused point _p2

    def _delCircle(self, c):
        """Delete a Circle from the Layer.

_delCircle(c)

This method is private to the Layer object.
        """
        assert not isinstance(c, arc.Arc), "Arc in _delCircle()"
        _users = self.__objdict[c]
        if _users is not None:
            for _user in _users:
                if isinstance(_user, dimension.RadialDimension):
                    _layer = _user.getDimLayer()
                    _layer.delObject(_user)
                else:
                    self.delObject(_user)
        _cp = c.getCenter()
        assert _cp in self.__objdict, "Circle center point not in objdict"
        self._disconnectUser(_cp, c)
        del self.__objdict[c]
        self.__circle_tree.remove(c)
        self.delObject(_cp) # remove possibly unused point _cp

    def _delArc(self, a):
        """Delete an Arc from the Layer.

_delArc(a)

This method is private to the Layer object.
        """
        _users = self.__objdict[a]
        if _users is not None:
            for _user in _users:
                if isinstance(_user, dimension.RadialDimension):
                    _layer = _user.getDimLayer()
                    _layer.delObject(_user)
                else:
                    self.delObject(_user)
        _cp = a.getCenter()
        assert _cp in self.__objdict, "Arc center point not in objdict"
        self._disconnectUser(_cp, a)
        for _ep in a.getEndpoints():
            _p = self.hasPoint(_ep)
            assert _p is not None, "Arc endpoint not found in layer"
            assert _p in self.__objdict, "Arc endpoint not in objdict"
            self._disconnectUser(_p, a)
            self.delObject(_p) # remove possibly unused point _p
        del self.__objdict[a]
        self.__arc_tree.remove(a)
        self.delObject(_cp) # remove possibly unused point _cp

    def _delHCLine(self, hcl):
        """Remove a HCLine object from the Layer.

_delHCLine(hcl)

This method is private to the Layer object.
        """
        _lp = hcl.getLocation()
        assert _lp in self.__objdict, "HCLine point not in objdict"
        self._disconnectUser(_lp, hcl)
        del self.__objdict[hcl]
        self.__hcline_tree.remove(hcl)
        self.delObject(_lp) # remove possibly unused point _lp

    def _delVCLine(self, vcl):
        """Remove a VCLine object from the Layer.

_delVCLine(vcl)

This method is private to the Layer object.
        """
        _lp = vcl.getLocation()
        assert _lp in self.__objdict, "VCLine point not in objdict"
        self._disconnectUser(_lp, vcl)
        del self.__objdict[vcl]
        self.__vcline_tree.remove(vcl)
        self.delObject(_lp) # remove possibly unused point _lp

    def _delACLine(self, acl):
        """Remove an ACLine object from the Layer.

_delACLine(acl)

This method is private to the Layer object.
        """
        _lp = acl.getLocation()
        assert _lp in self.__objdict, "ACLine point not in objdict"
        self._disconnectUser(_lp, acl)
        del self.__objdict[acl]
        self.__acline_tree.remove(acl)
        self.delObject(_lp) # remove possibly unused point _lp

    def _delCLine(self, cl):
        """Delete a CLine from the Layer.

_delCLine(s)

This method is private to the Layer object.
        """
        _p1, _p2 = cl.getKeypoints()                
        assert _p1 in self.__objdict, "CLine point p1 not in objdict"
        self._disconnectUser(_p1, cl)
        assert _p2 in self.__objdict, "CLine point p2 not in objdict"
        self._disconnectUser(_p2, cl)
        del self.__objdict[cl]
        self.__cline_tree.remove(cl)
        self.delObject(_p1) # remove possibly unused point _p1
        self.delObject(_p2) # remove possibly unused point _p2

    def _delCCircle(self, cc):
        """Delete a CCircle from the Layer.

_delCCircle(cc)

This method is private to the Layer object.
        """
        _cp = cc.getCenter()
        assert _cp in self.__objdict, "CCircle center point not in objdict"
        self._disconnectUser(_cp, cc)
        del self.__objdict[cc]
        self.__ccircle_tree.remove(cc)
        self.delObject(_cp) # remove possibly unused point _cp

    def _delChamfer(self, ch):
        """Remove a Chamfer from the Layer.

_delChamfer(ch)

This method is private to the Layer.
        """
        _chamfers = self.__chamfers
        _idx = None
        for _i in xrange(len(_chamfers)):
            if ch is _chamfers[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost chamfer from list"
        _s1, _s2 = ch.getSegments()
        assert _s1 in self.__objdict, "Chamfer s1 segment not in objdict"
        self._disconnectUser(_s1, ch)
        assert _s2 in self.__objdict, "Chamfer s2 segment not in objdict"
        self._disconnectUser(_s2, ch)            
        del self.__objdict[ch]
        del _chamfers[_idx] # restore the things the chamfer connects?

    def _delFillet(self, f):
        """Remove a Fillet from the Layer.

_delFillet(f)

This method is private to the Layer.
        """
        _fillets = self.__fillets
        _idx = None
        for _i in xrange(len(_fillets)):
            if f is _fillets[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost fillet from list"
        _s1, _s2 = f.getSegments()
        assert _s1 in self.__objdict, "Fillet s1 segment not in objdict"
        self._disconnectUser(_s1, f)
        assert _s2 in self.__objdict, "Fillet s2 segment not in objdict"
        self._disconnectUser(_s2, f)
        del self.__objdict[f]
        del _fillets[_idx] # restore the things the fillet connects?

    def _delLeader(self, l):
        """Delete a Leader from the Layer.

_delLeader(l)

This method is private to the Layer object.
        """
        _users = self.__objdict[l]
        if _users is not None:
            for _obj in _users:
                print "object bound to leader line: " + `_obj`
        _leaders = self.__leaders
        _idx = None
        for _i in xrange(len(_leaders)):
            if l is _leaders[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost leader from list"                
        _p1, _p2, _p3 = l.getLeaderPoints()
        assert _p1 in self.__objdict, "Leader p1 Point not in objdict"
        self._disconnectUser(_p1, l)
        assert _p2 in self.__objdict, "Leader p2 Point not in objdict"
        self._disconnectUser(_p2, l)
        assert _p3 in self.__objdict, "Leader p3 Point not in objdict"
        self._disconnectUser(_p3, l)
        del self.__objdict[l]
        del _leaders[_idx]
        self.delObject(_p1) # remove possibly unused point _p1
        self.delObject(_p2) # remove possibly unused point _p2
        self.delObject(_p3) # remove possibly unused point _p2
        
    def _delAngularDimension(self, adim):
        """Delete an AngularDimension from the Layer.

_delAngularDimension(adim)

This method is private to the Layer object.
        """
        _adims = self.__angular_dimensions
        _idx = None
        for _i in xrange(len(_adims)):
            if adim is _adims[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost adim in list"
        _layer1, _layer2, _layer3 = adim.getDimLayers()
        _p1, _p2, _p3 = adim.getDimPoints()
        if _layer1 is self:
            assert _p1 in self.__objdict, "ADim p1 not in objdict"
            self._disconnectUser(_p1, adim)
        else:
            _layer1.extDeref(_p1, adim)
        if _layer2 is self:
            assert _p2 in self.__objdict, "ADim p2 not in objdict"
            self._disconnectUser(_p2, adim)
        else:
            _layer2.extDeref(_p2, adim)
        if _layer3 is self:
            assert _p3 in self.__objdict, "ADim p3 not in objdict"
            self._disconnectUser(_p3, adim)
        else:
            _layer3.extDeref(_p3, adim)
        del self.__objdict[adim]
        del _adims[_idx]
        
    def _delRadialDimension(self, rdim):
        """Delete a RadialDimension from the Layer.

_delRadialDimension(rdim)

This method is private to the Layer object.
        """
        _rdims = self.__radial_dimensions
        _layer = rdim.getDimLayer()
        _circ = rdim.getDimCircle()
        if _layer is self:
            assert _circ in self.__objdict, "RDim circle not in objdict"
            self.__objdict[_circ] = None
        else:
            _layer.extDeref(_circ, rdim)
        for _i in xrange(len(_rdims)):
            if rdim is _rdims[_i]:
                del _rdims[_i]
                break
        del self.__objdict[rdim]

    def _delHorizontalDimension(self, hdim):
        """Delete an HorizontalDimension from the Layer.

_delHorizontalDimension(hdim)

This method is private to the Layer object.
        """
        _hdims = self.__horiz_dimensions
        _idx = None
        for _i in xrange(len(_hdims)):
            if hdim is _hdims[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost hdim in list"
        _layer1, _layer2 = hdim.getDimLayers()
        _p1, _p2 = hdim.getDimPoints()
        if _layer1 is self:
            assert _p1 in self.__objdict, "HDim p1 not in objdict"
            self._disconnectUser(_p1, hdim)
        else:
            _layer1.extDeref(_p1, hdim)
        if _layer2 is self:
            assert _p2 in self.__objdict, "HDim p2 not in objdict"
            self._disconnectUser(_p2, hdim)
        else:
            _layer2.extDeref(_p2, hdim)
        del self.__objdict[hdim]
        del _hdims[_idx]

    def _delVerticalDimension(self, vdim):
        """Delete an VerticalDimension from the Layer.

_delVerticalDimension(vdim)

This method is private to the Layer object.
        """
        _vdims = self.__vert_dimensions
        _idx = None
        for _i in xrange(len(_vdims)):
            if vdim is _vdims[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost vdim in list"
        _layer1, _layer2 = vdim.getDimLayers()
        _p1, _p2 = vdim.getDimPoints()
        if _layer1 is self:
            assert _p1 in self.__objdict, "VDim p1 not in objdict"
            self._disconnectUser(_p1, vdim)
        else:
            _layer1.extDeref(_p1, vdim)
        if _layer2 is self:
            assert _p2 in self.__objdict, "VDim p2 not in objdict"
            self._disconnectUser(_p2, vdim)
        else:
            _layer2.extDeref(_p2, vdim)
        del self.__objdict[vdim]
        del _vdims[_idx]

    def _delLinearDimension(self, ldim):
        """Delete an LinearDimension from the Layer.

_delLinearDimension(ldim)

This method is private to the Layer object.
        """
        _ldims = self.__linear_dimensions
        _idx = None
        for _i in xrange(len(_ldims)):
            if ldim is _ldims[_i]:
                _idx = _i
                break
        assert _idx is not None, "lost ldim in list"
        _layer1, _layer2 = ldim.getDimLayers()
        _p1, _p2 = ldim.getDimPoints()
        if _layer1 is self:
            assert _p1 in self.__objdict, "LDim p1 not in objdict"
            self._disconnectUser(_p1, ldim)
        else:
            _layer1.extDeref(_p1, ldim)
        if _layer2 is self:
            assert _p2 in self.__objdict, "LDim p2 not in objdict"
            self._disconnectUser(_p2, ldim)
        else:
            _layer2.extDeref(_p2, ldim)
        del self.__objdict[ldim]
        del _ldims[_idx]

    def usesObject(self, obj):
        """Return a list of objects tied to another object

usesObject(obj)

The object "obj" should be a Point, Segment, Circle, Arc,
Chamfer, or Fillet.
        """
        _users = []
        _objusers = None
        if isinstance(obj, (point.Point,
                            segment.Segment,
                            circle.Circle)):
            if obj in self.__objdict:
                _objusers = self.__objdict[obj]
        elif isinstance(obj, segjoint.Chamfer):
            for _c in self.__chamfers:
                if obj is _c:
                    _s1, _s2 = _c.getSegments()
                    assert _s1 in self.__objdict, "Chamfer s1 not in layer"
                    assert _s2 in self.__objdict, "Chamfer s2 not in layer"
                    _users.append(_s1)
                    _users.append(_s2)
        elif isinstance(obj, segjoint.Fillet):
            for _f in self.__fillets:
                if obj is _f:
                    _s1, _s2 = _f.getSegments()
                    assert _s1 in self.__objdict, "Fillet s1 not in layer"
                    assert _s2 in self.__objdict, "Fillet s2 not in layer"
                    _users.append(_s1)
                    _users.append(_s2)
        else:
            raise TypeError, "Invalid type for usesObject(): " + `obj`
        if _objusers is not None:
            for _user in _objusers:
                _users.append(_user)
        return _users

    def countUsers(self, obj):
        """Return the number of objects using a particular object.

countUsers(obj)

This method returns an integer value.
        """
        if obj not in self.__objdict:
            raise ValueError, "Object not stored in layer: " + `obj`
        _count = 0
        if isinstance(obj, (point.Point,
                            segment.Segment,
                            circle.Circle)):
            _users = self.__objdict[obj]
            if _users is not None:
                _count = len(_users)
        return _count

    def moveObjects(self, objlist, dx, dy):
        """Move the objects stored in the layer.

moveObject(objlist, dx, dy)

This method arguements are:

objlist: A list of objects stored in the layer
dx: The x-coordinate displacement
dy: The y-coordinate displacement
        """
        for _obj in objlist:
            if _obj not in self.__objdict:
                raise ValueError, "Object not found in layer: " + `_obj`
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        _pts = {}
        _objs = []
        for _obj in objlist:
            if isinstance(_obj, point.Point):
                _px, _py = _obj.getCoords()
                _pts[_obj] = self.hasPoint(((_px + _dx), (_py + _dy)))
            elif isinstance(_obj, segment.Segment):
                _p1, _p2 = _obj.getEndpoints()
                if _p1 not in _pts:
                    _p1x, _p1y = _p1.getCoords()
                    _pts[_p1] = self.hasPoint(((_p1x + _dx), (_p1y + _dy)))
                if _p2 not in _pts:
                    _p2x, _p2y = _p2.getCoords()            
                    _pts[_p2] = self.hasPoint(((_p2x + _dx), (_p2y + _dy)))
            elif isinstance(_obj, arc.Arc):
                _cp = _obj.getCenter()
                if _cp not in _pts:
                    _cx, _cy = _cp.getCoords()
                    _pts[_cp] = self.hasPoint(((_cx + _dx), (_cy + _dy)))
                for _ep in _obj.getEndpoints():
                    _lp = self.hasPoint(_ep)
                    assert _lp is not None, "Missing arc endpoint: " + `_ep`
                    if _lp not in _pts:
                        _lx, _ly = _lp.getCoords()
                        _pts[_lp] = self.hasPoint(((_lx + _dx), (_ly + _dy)))
            elif isinstance(_obj, (circle.Circle,
                                   ccircle.CCircle)):
                _cp = _obj.getCenter()
                if _cp not in _pts:
                    _cx, _cy = _cp.getCoords()
                    _pts[_cp] = self.hasPoint(((_cx + _dx), (_cy + _dy)))
            elif isinstance(_obj, (hcline.HCLine,
                                   vcline.VCLine,
                                   acline.ACLine)):
                _lp = _obj.getLocation()
                if _lp not in _pts:
                    _lx, _ly = _lp.getCoords()
                    _pts[_lp] = self.hasPoint(((_lx + _dx), (_ly + _dy)))
            elif isinstance(_obj, cline.CLine):
                _p1, _p2 = _obj.getKeypoints()
                if _p1 not in _pts:
                    _p1x, _p1y = _p1.getCoords()
                    _pts[_p1] = self.hasPoint(((_p1x + _dx), (_p1y + _dy)))
                if _p2 not in _pts:
                    _p2x, _p2y = _p2.getCoords()            
                    _pts[_p2] = self.hasPoint(((_p2x + _dx), (_p2y + _dy)))
            elif isinstance(_obj, (segjoint.Chamfer,
                                   segjoint.Fillet)):
                _s1, _s2 = _obj.getSegments()
                _p1, _p2 = _s1.getEndpoints()
                if _p1 not in _pts:
                    _p1x, _p1y = _p1.getCoords()
                    _pts[_p1] = self.hasPoint(((_p1x + _dx), (_p1y + _dy)))
                if _p2 not in _pts:
                    _p2x, _p2y = _p2.getCoords()            
                    _pts[_p2] = self.hasPoint(((_p2x + _dx), (_p2y + _dy)))
                _p1, _p2 = _s2.getEndpoints()
                if _p1 not in _pts:
                    _p1x, _p1y = _p1.getCoords()
                    _pts[_p1] = self.hasPoint(((_p1x + _dx), (_p1y + _dy)))
                if _p2 not in _pts:
                    _p2x, _p2y = _p2.getCoords()            
                    _pts[_p2] = self.hasPoint(((_p2x + _dx), (_p2y + _dy)))
            elif isinstance(_obj, leader.Leader):
                _objs.append(_obj)
                _p1, _p2, _p3 = _obj.getLeaderPoints()
                if _p1 not in _pts:
                    _p1x, _p1y = _p1.getCoords()
                    _pts[_p1] = self.hasPoint(((_p1x + _dx), (_p1y + _dy)))
                if _p2 not in _pts:
                    _p2x, _p2y = _p2.getCoords()
                    _pts[_p2] = self.hasPoint(((_p2x + _dx), (_p2y + _dy)))
                if _p3 not in _pts:
                    _p3x, _p3y = _p3.getCoords()
                    _pts[_p3] = self.hasPoint(((_p3x + _dx), (_p3y + _dy)))
            elif isinstance(_obj, dimension.Dimension):
                _dtx, _dty = _obj.getLocation()
                _obj.setLocation((_dtx + _dx), (_dty + _dy))
            else:
                raise TypeError, "Unexpected object:" + `_obj`
        for _pt in _pts:
            _lp = _pts[_pt]        
            if _lp is None:
                _pt.move(_dx, _dy)
            else:
                for _user in self.usesObject(_pt):
                    if isinstance(_user, segment.Segment):
                        _p1, _p2 = _user.getEndpoints()
                        if _p1 is _pt:
                            _user.setP1(_lp)
                        elif _p2 is _pt:
                            _user.setP2(_lp)
                        else:
                            raise ValueError, "Segment/Point mismatch"
                    elif isinstance(_user, arc.Arc):
                        _cp = _user.getCenter()
                        if _cp is _pt:
                            _user.setCenter(_lp)
                    elif isinstance(_user, (circle.Circle,
                                            ccircle.CCircle)):
                        _user.setCenter(_lp)
                    elif isinstance(_user, (hcline.HCLine,
                                            vcline.VCLine,
                                            acline.ACLine)):
                        _user.setLocation(_lp)
                    elif isinstance(_user, cline.CLine):
                        _p1, _p2 = _user.getKeypoints()
                        if _p1 is _pt:
                            _user.setP1(_lp)
                        elif _p2 is _pt:
                            _user.setP2(_lp)
                        else:
                            raise ValueError, "CLine/Point mismatch"
                    elif isinstance(_user, leader.Leader):
                        _p1, _p2, _p3 = _user.getLeaderPoints()
                        if _p1 is _pt:
                            _user.setP1(_pt)
                        elif _p2 is _pt:
                            _user.setP2(_pt)
                        elif _p3 is _pt:
                            _user.setP3(_pt)
                        else:
                            raise ValueError, "Leader/Point mismatch"
                    elif isinstance(_user, dimension.LinearDimension):
                        _l1, _l2 = _user.getDimLayers()
                        _p1, _p2 = _user.getDimLayers()
                        if _l1 is self and _p1 is _pt:
                            _user.setP1(_l1, _lp)
                        if _l2 is self and _p2 is _pt:
                            _user.setP2(_l2, _lp)
                    elif isinstance(_user, dimension.AngularDimension):
                        _vl, _l1, _l2 = _user.getDimLayers()
                        _vp, _p1, _p2 = _user.getDimLayers()
                        if _vl is self and _vp is _pt:
                            _user.setVertexPoint(_l1, _lp)
                        if _l1 is self and _p1 is _pt:
                            _user.setP1(_l1, _lp)
                        if _l2 is self and _p2 is _pt:
                            _user.setP2(_l2, _lp)
                    else:
                        raise TypeError, "Unknown user: " + `_user`
                    self._disconnectUser(_pt, _user)
            self.delObject(_pt)
        for _obj in _objs:
            if isinstance(_obj, leader.Leader):
                _obj.calcArrowPoints()
            else:
                raise TypeError, "Unexpected object: " + `_obj`

    def moveObject(self, obj, dx, dy):
        """Move an object stored in the layer.

moveObject(obj, dx, dy)

The object "obj" must be stored in this layer
        """
        import warnings
        warnings.warn("moveObject() is deprecated - use moveObjects()",
                      DeprecationWarning)
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        if obj not in self.__objdict:
            raise ValueError, "Object not found in layer: " + `obj`
        if isinstance(obj, point.Point):
            _x, _y = obj.getCoords()
            _px = _x + _dx
            _py = _y + _dy
            _users = self.__objdict[obj]
            _lp = self.hasPoint((_px, _py))            
            if _users is not None and len(_users):
                if _lp is None:
                    obj.move(_dx, _dy)
                else:
                    pass # WRONG, but this method is going away!!!
            else: 
                if _lp is None:
                    obj.move(_dx, _dy)
                else:
                    self.delObject(obj)
        elif isinstance(obj, arc.Arc):
            _cp = obj.getCenter()
            _x, _y = _cp.getCoords()
            _px = _x + _dx
            _py = _y + _dy
            _lcp = self.hasPoint((_px, _py))
            if _lcp is None:
                _lcp = point.Point(_px, _py)
                self.addObject(_lcp)
            _ep1, _ep2 = obj.getEndpoints()
            #
            # test first endpoint
            #
            _epx, _epy = _ep1 
            _lex = _epx + _dx
            _ley = _epy + _dy
            _lp1 = self.hasPoint((_lex, _ley))
            if _lp1 is None:
                _lp1 = point.Point(_lex, _ley)
                self.addObject(_lp1)
            #
            # test second endpoint
            #
            _epx, _epy = _ep2
            _lex = _epx + _dx
            _ley = _epy + _dy
            _lp2 = self.hasPoint((_lex, _ley))
            if _lp2 is None:
                _lp2 = point.Point(_lex, _ley)
                self.addObject(_lp2)
            obj.setCenter(_lcp)
            self._connectUser(_lcp, obj)
            self._connectUser(_lp1, obj)
            self._connectUser(_lp2, obj)
            self.delObject(_cp)
            _lp = self.hasPoint(_ep1)
            if _lp is not None:
                self.delObject(_lp)
            _lp = self.hasPoint(_ep2)
            if _lp is not None:
                self.delObject(_lp)                
        elif isinstance(obj, circle._Circle):
            _cp = obj.getCenter()
            _x, _y = _cp.getCoords()
            _px = _x + _dx
            _py = _y + _dy
            _lcp = self.hasPoint((_px, _py))
            if _lcp is None:
                _users = self.__objdict[_cp]
                if (_users is not None and
                    len(_users) == 1 and
                    _users[0] is obj):
                    obj.move(_dx, _dy)
                else:
                    _lcp = point.Point(_px, _py)
                    self.addObject(_lcp)
                    obj.setCenter(_lcp)
                    self._connectUser(_lcp, obj)
            else:
                obj.setCenter(_lcp)
                self._connectUser(_lcp, obj)
            self.delObject(_cp)
        elif isinstance(obj, (hcline.HCLine, vcline.VCLine, acline.ACLine)):
            _sp = obj.getLocation()
            _x, _y = _sp.getCoords()
            _px = _x + _dx
            _py = _y + _dy
            _lsp = self.hasPoint((_px, _py))
            if _lsp is None:
                _users = self.__objdict[_sp]
                if (_users is not None and
                    len(_users) == 1 and
                    _users[0] is obj):
                    obj.move(_dx, _dy)
                else:
                    _lsp = point.Point(_px, _py)
                    self.addObject(_lsp)
                    obj.setLocation(_lsp)
                    self._connectUser(_lsp, obj)
            else:
                obj.setLocation(_lsp)
                self._connectUser(_lsp, obj)
            self.delObject(_sp)
        elif isinstance(obj, (segment.Segment, cline.CLine)):
            if isinstance(obj, segment.Segment):
                _p1, _p2 = obj.getEndpoints()
            else:
                _p1, _p2 = obj.getKeypoints()
            #
            # check p1
            #
            _x, _y = _p1.getCoords()
            _px = _x + _dx
            _py = _y + _dy
            _lp = self.hasPoint((_px, _py))
            if _lp is None:
                _users = self.__objdict[_p1]
                if (_users is not None and
                    len(_users) == 1 and
                    _users[0] is obj):
                    obj.getP1().move(_dx,_dy)
                else:
                    _lp = point.Point(_px, _py)
                    self.addObject(_lp)
                    obj.setP1(_lp)
                    self._connectUser(_lp, obj)
            else:
                obj.setP1(_lp)
                self._connectUser(_lp, obj)
            #
            # check p2
            #
            _x, _y = _p2.getCoords()
            _px = _x + _dx
            _py = _y + _dy
            _lp = self.hasPoint((_px, _py))
            if _lp is None:
                _users = self.__objdict[_p2]
                if (_users is not None and
                    len(_users) == 1 and
                    _users[0] is obj):
                    obj.getP2().move(_dx, _dy)
                else:
                    _lp = point.Point(_px, _py)
                    self.addObject(_lp)
                    obj.setP2(_lp)
                    self._connectUser(_lp, obj)
            else:
                obj.setP2(_lp)
                self._connectUser(_lp, obj)
            self.delObject(_p1)
            self.delObject(_p2)
        elif isinstance(obj, (segjoint.Chamfer, segjoint.Fillet)):
            _s1, _s2 = obj.getSegments()
            # self.moveObject(_s1)
            # self.moveObject(_s2)
        elif isinstance(obj, dimension.Dimension):
            _x, _y = obj.getLocation()
            _nx = _x + _dx
            _ny = _y + _dy
            obj.setLocation(_nx, _ny)

    def findObject(self, obj):
        """Return an object in the layer that is equivalent to a test object.

findObject(obj)

This method returns None if a suitable object is not found.
        """
        _retobj = None
        if obj in self.__objdict:
            _retobj = obj
        else:
            self.update()
            _objtree = None
            _objlist = None
            if isinstance(obj, point.Point):
                _objtree = self.__point_tree
            elif isinstance(obj, segment.Segment):
                _objtree = self.__segment_tree
            elif isinstance(obj, arc.Arc):
                _objtree = self.__arc_tree
            elif isinstance(obj, circle.Circle):
                _objtree = self.__circle_tree
            elif isinstance(obj, hcline.HCLine):
                _objtree = self.__hcline_tree
            elif isinstance(obj, vcline.VCLine):
                _objtree = self.__vcline_tree
            elif isinstance(obj, acline.ACLine):
                _objtree = self.__acline_tree
            elif isinstance(obj, cline.CLine):
                _objtree = self.__cline_tree
            elif isinstance(obj, ccircle.CCircle):
                _objtree = self.__ccircle_tree
            elif isinstance(obj, segjoint.Fillet):
                _objlist = self.__fillets
            elif isinstance(obj, segjoint.Chamfer):
                _objlist = self.__chamfers
            elif isinstance(obj, leader.Leader):
                _objlist = self.__leaders
            elif isinstance(obj, dimension.HorizontalDimension):
                _objlist = self.__horiz_dimensions
            elif isinstance(obj, dimension.VerticalDimension):
                _objlist = self.__vert_dimensions
            elif isinstance(obj, dimension.LinearDimension):
                _objlist =  self.__linear_dimensions
            elif isinstance(obj, dimension.RadialDimension):
                _objlist = self.__radial_dimensions
            elif isinstance(obj, dimension.AngularDimension):
                _objlist = self.__angular_dimensions
            else:
                raise TypeError, "Invalid type for in operation: " + `obj`
            if _objtree is not None:
                _retobj = _objtree.binscan(obj)
            elif _objlist is not None:
                _seen = True
                try:
                    _idx = _objlist.index(obj)
                except ValueError:
                    _seen = False
                if _seen:
                    _retobj = _objlist[_idx]
        return _retobj

    def hasPoint(self, p, tol=tolerance.TOL):
        """Find a Point in the layer.

hasPoint(p [,tol])

There is a single required argument:

p: Either a Point object or a tuple of two-floats

There is one optional argument:

tol: A float equal or greater than 0

The function returns an existing Point object found at
the same location, or None.
        """
        _p = p
        if not isinstance(_p, point.Point):
            _p = point.Point(p)
        _lp = None
        if _p in self.__objdict:
            _lp = _p
        else:
            _t = tolerance.toltest(tol)
            self.update()
            _lp = self.__point_tree.binscan(_p, _t)
        return _lp

    def hasPointAt(self, x, y):
        """Find a Point in the layer at a specific location.

hasPointAt(x, y)

The two required arguments are:

x: A float giving the "x" coordinate
y: A float giving the "y" coordinate

This method returns a the Point if one is found at
the location, or None otherwise.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        self.update()
        return self.__point_tree.atCoords(_x, _y)

    def hasCircleAt(self, r, x, y):
        """Find a Circle of some radius in the layer at a specific location.

hasCircleAt(r, x, y)

The three required arguments are:

r: A float giving the Circle radius
x: A float giving the "x" coordinate
y: A float giving the "y" coordinate

This method returns a the Circle if one is found at
the location, or None otherwise.
        """
        _r = r
        if not isinstance(_r, float):
            _r = float(_r)
        if _r < 1e-10:
            raise ValueError, "Invalid radius: " + `_r`
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        self.update()
        return self.__circle_tree.atCoords(_r, _x, _y)

    def extRef(self, obj, ext, l):
        """Store a reference of an object in another layer in this layer.

extRef(obj, ext, l)

There are three required arguments to this function:

obj: An object in this layer - currently this must be a Point or a Circle
ext: An object in another layer
l: The layer holding the external object

The object in the other layer must call this function so
that a reference to itself can be associated to the object
stored in this layer.
        """
        if not isinstance(obj, (point.Point, circle.Circle)):
            raise TypeError, "Invalid object: " + `obj`
        if obj not in self.__objdict:
            raise ValueError, "Object not found in layer: " + `obj`
        if ext is None:
            raise ValueError, "External object cannot be None."
        if not isinstance(l, Layer):
            raise TypeError, "Invalid layer argument: " + `l`
        if l is self:
            raise ValueError, "extRef() called for self as layer"
        if obj not in self.__externals:
            self.__externals[obj] = []
        self.__externals[obj].append((l, ext))

    def extDeref(self, obj, ext):
        """Release a reference to an object in another layer from a object.

extDeref(obj, ext)

There are two required arguments to this function:

obj: An object in this layer - Currently this must be a Point or a Circle
ext: An object in another layer

The object in the other layer must call this function so
that a reference to itself can be removed from the external
reference dictionary kept in this layer.
        """
        if not isinstance(obj, (point.Point, circle.Circle)):
            raise TypeError, "Invalid object: " + `obj`
        if obj not in self.__objdict:
            raise ValueError, "Object not found in this layer: " + `obj`
        _users = self.__externals[obj]
        for _i in xrange(len(_users)):
            _layer, _extobj = _users[_i]
            if ext is _extobj:
                del _users[_i]
                break
        if not len(self.__externals[obj]):
            del self.__externals[obj]

    def getExtRefs(self, obj):
        """Return the list of external objects tied to an object in this layer.

getExtRefs(obj)

The argument obj must be an object in this layer. The returned list
contains tuples of the form (layer, object).
        """
        _refs = []
        if obj in self.__externals:
            for _tup in self.__externals[obj]:
                _refs.append(_tup)
        return _refs
    
    def mapPoint(self, p, tol=tolerance.TOL, count=2):
        """Find a Point in the layer

mapPoint(p [,tol, count])

There is a single required argument:

p: Either a Point object or a tuple of two-floats

There are two optional arguments:

tol: A float equal or greater than 0 for distance tolerance comparisons.
count: An integer value indicating the largest number of objects to
       return. By default this value is 2.

Setting "count" to None or a negative value will result in the maximum
number of objects being unlimited.

This method tests the objects in the Layer to see if the
Point can can be mapped on to any of them. The returned list
consists of tuples in the form:

(obj, pt)

Where "obj" is the object the point was mapped to and "pt"
is the projected point on the object.
        """
        _scan = True
        _hits = []
        _p = p
        if not isinstance(_p, point.Point):
            _p = point.Point(p)
        _t = tolerance.toltest(tol)
        _count = count
        if _count is None:
            _count = sys.maxint
        else:
            if not isinstance(_count, int):
                _count = int(count)
            if _count < 0: # What if _count == 0 ???
                _count = sys.maxint
        self.update()
        _lp = self.hasPoint(_p, _t)
        if _lp is not None:
            assert _lp in self.__objdict, "Point not in pointdict"
            _users = self.__objdict[_lp]
            if _users is not None:
                for _user in _users:
                    _hits.append((_user, _lp))
                    if len(_hits) == _count:
                        _scan = False
                        break
        if _scan:
            for _tup in self.__segment_tree.mapPoint(_p, _t, _count):
                _hits.append(_tup)
                if len(_hits) == _count:
                    _scan = False
                    break
        if _scan:
            for _tup in self.__circle_tree.mapPoint(_p, _t, _count):
                _hits.append(_tup)
                if len(_hits) == _count:
                    _scan = False
                    break
        if _scan:
            for _tup in self.__arc_tree.mapPoint(_p, _t, _count):
                _hits.append(_tup)
                if len(_hits) == _count:
                    _scan = False
                    break
        if _scan: # max of one object with hcline_tree
            for _tup in self.__hcline_tree.mapPoint(_p, _t):
                _hits.append(_tup)
                if len(_hits) == _count:
                    _scan = False
                    break
        if _scan: # max of one object with vcline_tree
            for _tup in self.__vcline_tree.mapPoint(_p, _t):
                _hits.append(_tup)
                if len(_hits) == _count:
                    _scan = False
                    break
        if _scan:
            for _tup in self.__acline_tree.mapPoint(_p, _t, _count):
                _hits.append(_tup)
                if len(_hits) == _count:
                    _scan = False
                    break
        if _scan:
            for _tup in self.__ccircle_tree.mapPoint(_p, _t, _count):
                _hits.append(_tup)
                if len(_hits) == _count:
                    _scan = False
                    break
        if _scan:
            for _tup in self.__cline_tree.mapPoint(_p, _t, _count):
                _hits.append(_tup)
                if len(_hits) == _count:
                    _scan = False
                    break
        if _scan:
            for _leader in self.__leaders:
                _pt = _leader.mapPoint(_p, _t)
                if _pt is not None:
                    _hits.append((_leader, _pt))
                    if len(_hits) == _count:
                        break
        _px, _py = _p.getCoords()                    
        if _scan:
            for _ldim in self.__linear_dimensions:
                if _ldim.onDimension(_px, _py, _t):
                    _hits.append((_ldim, None))
                    if len(_hits) == _count:
                        break
        if _scan:
            for _hdim in self.__horiz_dimensions:
                if _hdim.onDimension(_px, _py, _t):
                    _hits.append((_hdim, None))
                    if len(_hits) == _count:
                        break
        if _scan:
            for _vdim in self.__vert_dimensions:
                if _vdim.onDimension(_px, _py, _t):
                    _hits.append((_vdim, None))
                    if len(_hits) == _count:
                        break
        if _scan:
            for _rdim in self.__radial_dimensions:
                if _rdim.onDimension(_px, _py, _t):
                    _hits.append((_rdim, None))
                    if len(_hits) == _count:
                        break
        if _scan:
            for _adim in self.__angular_dimensions:
                if _adim.onDimension(_px, _py, _t):
                    _hits.append((_adim, None))
                    if len(_hits) == _count:
                        break
        return _hits

    def getLayerEntities(self, entity):
        """Get all of a particular type of entity in the Layer.

getLayerEntities(entity)

The argument "entity" should be one of the following:
point, segment, circle, arc, hcline, vcline, acline,
cline, ccircle, chamfer, fillet, linear_dimension,
horizontal_dimenions, vertical_dimension, radial_dimension,
or angular_dimension
        """
        if entity == "point":
            objs = self.__point_tree[:]
        elif entity == "segment":
            objs = self.__segment_tree[:]
        elif entity == "circle":
            objs = self.__circle_tree[:]
        elif entity == "arc":
            objs = self.__arc_tree[:]
        elif entity == "hcline":
            objs = self.__hcline_tree[:]
        elif entity == "vcline":
            objs = self.__vcline_tree[:]
        elif entity == "acline":
            objs = self.__acline_tree[:]
        elif entity == "cline":
            objs = self.__cline_tree[:]
        elif entity == "ccircle":
            objs = self.__ccircle_tree[:]
        elif entity == "chamfer":
            objs = self.__chamfers[:]
        elif entity == "fillet":
            objs = self.__fillets[:]
        elif entity == "leader":
            objs = self.__leaders[:]
        elif entity == "linear_dimension":
            objs = self.__linear_dimensions[:]
        elif entity == "horizontal_dimension":
            objs = self.__horiz_dimensions[:]
        elif entity == "vertical_dimension":
            objs = self.__vert_dimensions[:]
        elif entity == "radial_dimension":
            objs = self.__radial_dimensions[:]
        elif entity == "angular_dimension":
            objs = self.__angular_dimensions[:]
        else:
            raise ValueError, "Invalid layer entity: " + `entity`
        return objs
    
    def addChild(self, child):
        """Add a child Layer to a Layer.

addChild(child)

There is no limit to the number of children for a Layer.
        """
        if not isinstance(child, Layer):
            raise TypeError, "Invalid child Layer: " + `child`
        child.__parent = self
        self.__children.append(child)

    def delChild(self, child):
        """Remove a child Layer from the Layer.

delChild(child)

The child layer will not be removed if it has children
of its own.
        """
        if not isinstance(child, Layer):
            raise TypeError, "Invalid child Layer: " + `child`
        seen = False
        index = 0
        for kid in self.__children:
            if child is kid:
                seen = True
                break
            index = index + 1
        if not seen:
            raise ValueError, "Invalid child Layer: " + `child`
        if not child.hasChildren():
            del self.__children[index]

    def hasChildren(self):
        """Test if there are children of this Layer.

hasChildren()

The function returns 1 if the Layer has children, 0 otherwise.
        """
        val = False
        if len(self.__children) > 0:
            val = True
        return val

    def getChildren(self):
        """Get the children of a Layer.

getChildren()

This function returns a list of children Layers for a Layer.
        """
        return self.__children[:]
    
    def getParent(self):
        """Get the Layer's parent.

getParent()        
        """
        return self.__parent

    def setParent(self, parent):
        """Set the Layer's parent.

setParent(parent)

Set the parent layer for a Layer object. The parent cannot
be set to None.
        """
        if parent is None:
            raise ValueError, "Layer parent cannot be set to None."
        if not isinstance(parent, Layer):
            raise TypeError, "Invalid parent layer: " + `parent`
        self.__parent = parent

    parent = property(getParent, setParent, None, "Parent of a layer.")

    def getVisibility(self):
        """Return the visibility value of the Layer.

getVisibility()        
        """
        return self.__visibility

    def setVisibility(self, val):
        """Set the visibility of the Layer.

setVisibility(val)

The argument "val" should be either "True" or "False".
        """
        if val is not True and val is not False:
            raise ValueError, "Invalid visibility value: " + `val`
        self.__visibility = val

    visibility = property(getVisibility, setVisibility, None, "Visibility.")
    
    def getScale(self):
        """Return the scale factor of the Layer.

getScale()        
        """
        return self.__scale

    def setScale(self, scale):
        """Set the scale factor for the Layer.

setScale(scale)

The scale factor must be a positive float value greater than 0.
        """
        _s = scale
        if not isinstance(_s, float):
            _s = float(scale)
        if _s < 1e-10:
            raise ValueError, "Invalid scale factor: " + `_s`
        self.__scale = _s

    scale = property(getScale, setScale, None, "Layer scale factor.")
    
    def getBoundary(self):
        """Return the maximum and minimum values of the object in the Layer.

getBoundary()

The function returns a tuple holding four float values:

(xmin, ymin, xmax, _ymax)

A default value of (-1.0, -1.0, 1.0, 1.0) is returned for a Layer
containing no objects.
        """
        _defaults = True
        _xmin = -1.0
        _ymin = -1.0
        _xmax = 1.0
        _ymax = 1.0
        for _obj in self.__point_tree:
            _x, _y = _obj.getCoords()            
            if _defaults:
                _xmin = _x
                _ymin = _y
                _xmax = _x
                _ymax = _y
                _defaults = False
            else:
                if _x < _xmin:
                    _xmin = _x
                if _x > _xmax:
                    _xmax = _x
                if _y < _ymin:
                    _ymin = _y
                if _y > _ymax:
                    _ymax = _y
        _ctrees = self.__circle_tree + self.__arc_tree + self.__ccircle_tree
        for _obj in _ctrees:
            _x, _y = _obj.getCenter().getCoords()
            _rad = _obj.getRadius()
            if _defaults:
                _xmin = _x - _rad
                _ymin = _y - _rad
                _xmax = _x + _rad
                _ymax = _y + _rad
                _defaults = False
            else:
                _val = _x - _rad
                if _val < _xmin:
                    _xmin = _val
                _val = _y - _rad
                if _val < _ymin:
                    _ymin = _val
                _val = _x + _rad
                if _val > _xmax:
                    _xmax = _val
                _val = _y + _rad
                if _val > _ymax:
                    _ymax = _val
        _dims = (self.__linear_dimensions +
                 self.__horiz_dimensions +
                 self.__vert_dimensions +
                 self.__radial_dimensions +
                 self.__angular_dimensions)
        for _dim in _dims:
            _dx, _dy = _dim.getLocation()
            _dxpts = _dim.getDimXPoints()
            _dypts = _dim.getDimYPoints()
            _dxmin = min(_dx, min(_dxpts))
            _dxmax = max(_dx, max(_dxpts))
            _dymin = min(_dy, min(_dypts))
            _dymax = max(_dy, max(_dypts))
            if _defaults:
                _xmin = _dxmin
                _ymin = _dymin
                _xmax = _dxmax
                _ymax = _dymax
                _defaults = False
            else:
                if _dxmin < _xmin:
                    _xmin = _dxmin
                if _dymin < _ymin:
                    _ymin = _dymin
                if _dxmax > _xmax:
                    _xmax = _dxmax
                if _dymax > _ymax:
                    _ymax = _dymax
        return (_xmin, _ymin, _xmax, _ymax)

    def objsInRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return a all the objects in the Layer visible within the bounds.

objsInRegion(xmin, ymin, xmax, ymax[, fully])

The function has four required arguments:

xmin: The minimum x-value of the region
ymin: The minimum y-value of the region
xmax: The maximum x-value of the region
ymax: The maximum y-value of the region

There is a single optional argument:

fully: A True/False value indicating if the object must be
       entirely within the region [fully=True], or can
       merely pass through [fully=False]. The default value
       is False.

The function returns a list of objects.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if xmax < xmin:
            raise ValueError, "Value error: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Value error: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        self.update()
        _objs = []
        for _obj in self.__point_tree:
            _x, _y = _obj.getCoords()
            if _x > _xmax:
                break
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__segment_tree:
            _p1, _p2 = _obj.getEndpoints()
            _p1x, _p1y = _p1.getCoords()
            _p2x, _p2y = _p2.getCoords()
            _pxmin = min(_p1x, _p2x) - 1e-10
            if _pxmin > _xmax:
                break
            _pymin = min(_p1y, _p2y)
            _pxmax = max(_p1x, _p2x)
            _pymax = max(_p1y, _p2y)
            if _xmin > _pxmax or _ymin > _pymax or _ymax < _pymin:
                continue
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax, fully):
                _objs.append(_obj)
        _ctrees = (self.__circle_tree +
                   self.__arc_tree +
                   self.__ccircle_tree)
        for _obj in _ctrees:
            _x, _y = _obj.getCenter().getCoords()
            _rad = _obj.getRadius()
            _cxmin = _x - _rad
            if _cxmin > _xmax:
                break
            _cymin = _y - _rad
            _cxmax = _x + _rad
            _cymax = _y + _rad
            if _xmin > _cxmax or _ymin > _cymax or _ymax < _cymin:
                continue
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax, fully):
                _objs.append(_obj)
        for _obj in self.__hcline_tree:
            _x, _y = _obj.getLocation().getCoords()
            if _y > _ymax:
                break
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__vcline_tree:
            _x, _y = _obj.getLocation().getCoords()
            if _x > _xmax:
                break
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__acline_tree:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__cline_tree:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__chamfers:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__leaders: # fixme
            _p1, _p2, _p3 = _obj.getLeaderPoints()
            _p1x, _p1y = _p1.getCoords()
            _p2x, _p2y = _p2.getCoords()
            _p3x, _p3y = _p3.getCoords()
            if _xmin > max(_p1x, _p2x, _p3x):
                break
            if ((_xmax < min(_p1x, _p2x, _p3x)) or
                (_ymax < min(_p1y, _p2y, _p3y)) or
                (_ymin > max(_p1y, _p2y, _p3y))):
                continue
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__fillets:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__linear_dimensions:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__horiz_dimensions:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)
        for _obj in self.__vert_dimensions:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)            
        for _obj in self.__radial_dimensions:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)            
        for _obj in self.__angular_dimensions:
            if _obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _objs.append(_obj)            
        return _objs

    def update(self):
        """Check that the objects in this layer are stored correctly.

update()

This function checks that the objects held in this layer are kept
in the proper order. Also, any duplicated objects that may have
be created due to modifying entities in the layer are removed.
        """
        if (self.__segment_tree.isModified() or
            self.__circle_tree.isModified() or
            self.__arc_tree.isModified() or
            self.__hcline_tree.isModified() or
            self.__vcline_tree.isModified() or
            self.__acline_tree.isModified() or
            self.__cline_tree.isModified() or
            self.__ccircle_tree.isModified()):
            self._sweepPointUsers()
        self._sweepPointTree()
        self._sweepSegmentTree()
        self._sweepCircleTree()
        self._sweepArcTree()
        self._sweepSPCLineTree(self.__hcline_tree)
        self._sweepSPCLineTree(self.__vcline_tree)
        self._sweepSPCLineTree(self.__acline_tree)
        self._sweepCLineTree()
        self._sweepCCircleTree()

    def _sweepPointUsers(self):
        """Check that all objects associated with a point still do.

_sweepPointUsers()

This method is private to the Layer object.
        """
        for _pt in self.__point_tree:
            assert _pt in self.__objdict, "Lost point from objdict"
            _users = self.__objdict[_pt]            
            _i = 0
            _idx = []
            if _users is not None:
                for _user in _users:
                    if isinstance(_user, segment.Segment):
                        _p1, _p2 = _user.getEndpoints()
                        if _pt is not _p1 and _pt is not _p2:
                            _idx.append(_i)
                    elif isinstance(_user, arc.Arc):
                        _cp = _user.getCenter()
                        _ep1, _ep2 = _user.getEndpoints()
                        if _pt is not _cp and _pt != _ep1 and _pt != _ep2:
                            _idx.append(_i)                        
                    elif isinstance(_user, circle._Circle):
                        if _pt is not _user.getCenter():
                            _idx.append(_i)
                    elif isinstance(_user, (hcline.HCLine,
                                            vcline.VCLine,
                                            acline.ACLine)):
                        _lp = _user.getLocation()
                        if _pt is not _lp:
                            _idx.append(_i)
                    elif isinstance(_user, cline.CLine):
                        _p1, _p2 = _user.getKeypoints()
                        if _pt is not _p1 and _pt is not _p2:
                            _idx.append(_i)
                    elif isinstance(_user, dimension.Dimension):
                        if isinstance(_user, dimension.LinearDimension):
                            _p1, _p2 = _user.getDimPoints()
                            if _pt is not _p1 and _pt is not _p2:
                                _idx.append(_i)
                        elif isinstance(_user, dimension.AngularDimension):
                            _p1, _p2, _p3 = _user.getDimPoints()
                            if (_pt is not _p1 and
                                _pt is not _p2 and
                                _pt is not _p3):
                                _idx.append(_i)
                    _i = _i + 1
            if len(_idx):
                _idx.reverse()
                for _index in _idx:
                    del _users[_index]

    def _sweepPointTree(self):
        """Validate the layer Point storage.

_sweepPointTree()

This method is private to the Layer object.
        """
        _modpts = self.__point_tree.getModified()
        if len(_modpts):
            _deadpts = self.__point_tree.clean()
            for _pt in _deadpts:
                _lp = self.__point_tree.binscan(_pt)
                assert _lp is not None, "lost point in point_tree: " + str(_pt)
                _users = self.__objdict[_pt]
                if _users is not None:
                    for _user in _users:
                        if isinstance(_user, segment.Segment):
                            _p1, _p2 = _user.getEndpoints()
                            if _pt is _p1:
                                _user.setP1(_lp)
                            elif _pt is _p2:
                                _user.setP2(_lp)
                            else:
                                pass
                        elif isinstance(_user, circle._Circle):
                            _user.setCenter(_lp)
                        elif isinstance(_user, (hcline.HCLine,
                                                vcline.VCLine,
                                                acline.ACLine)):
                            _user.setLocation(_lp)
                        elif isinstance(_user, cline.CLine):
                            _p1, _p2 = _user.getKeypoints()
                            if _pt is _p1:
                                _user.setP1(_lp)
                            elif _pt is _p2:
                                _user.setP2(_lp)
                            else:
                                pass
                        elif isinstance(_user, dimension.Dimension):
                            if isinstance(_user, dimension.LinearDimension):
                                _p1, _p2 = _user.getDimPoints()
                                if _pt is _p1:
                                    _user.setP1(_lp)
                                elif _pt is _p2:
                                    _user.setP2(_lp)
                                else:
                                    pass
                            elif isinstance(_user, dimension.AngularDimension):
                                _cp, _p1, _p2 = _user.getDimPoints()
                                if _pt is _cp:
                                    _user.setCenterPoint(_lp)
                                elif _pt is _p1:
                                    _user.setP1(_lp)
                                elif _pt is _p2:
                                    _user.setP2(_lp)
                                else:
                                    pass
                            else:
                                raise TypeError, "Missing type for keypoint modification:" + `_user`

                        self._connectUser(_lp, _user)
                if self.__externals.has_key(_pt):
                    for _layer, _extobj in self.__externals[_pt]:
                        if isinstance(_extobj, dimension.LinearDimension):
                            _p1, _p2 = _extobj.getDimPoints()
                            if _pt is _p1:
                                _extobj.setP1(_lp)
                            elif _pt is _p2:
                                _extobj.setP2(_lp)
                            else:
                                pass
                        elif isinstance(_extobj, dimension.RadialDimension):
                            _extobj.setCenterPoint(_lp)
                        elif isinstance(_extobj, dimension.AngularDimension):
                            _cp, _p1, _p2 = _extobj.getDimPoints()
                            if _pt is _cp:
                                _extobj.setCenterPoint(_lp)
                            elif _pt is _p1:
                                _extobj.setP1(_lp)
                            elif _pt is _p2:
                                _extobj.setP2(_lp)
                            else:
                                pass
                del self.__objdict[_pt]                            
            for _pt in _modpts:
                if _pt not in _deadpts:
                    for _obj in self.__objdict[_pt]:
                        _obj.modified()
                    _pt.reset()

    def _sweepSegmentTree(self):
        """Validate the layer segment storage.

_sweepSegmentTree()

This method is private to the Layer.
        """
        _modsegs = self.__segment_tree.getModified()
        if len(_modsegs):
            _deadsegs = self.__segment_tree.clean()
            for _seg in _modsegs:
                if _seg in _deadsegs:
                    continue
                if _seg.length() < 1e-10:
                    _deadsegs.append(_seg)
                else:
                    _seg.reset()
                    for _pt in _seg.getEndpoints():
                        assert _pt in self.__objdict, "Segment point not in objdict:" + `_pt`
                        _add_flag = True
                        _users = self.__objdict[_pt]
                        if _users is not None:
                            for _user in _users:
                                if _user is _seg:
                                    _add_flag = False
                                    break
                            if _add_flag:
                                self._connectUser(_pt, _seg)
            for _seg in _deadsegs:
                _keepseg = self.__segment_tree.binscan(_seg)
                assert _keepseg is not None, "Lost segment: " + str(_seg)
                _p1, _p2 = _seg.getEndpoints()
                self._disconnectUser(_p1, _seg)
                self._disconnectUser(_p2, _seg)
                _users = self.__objdict[_seg]
                if _users is not None: # fixme - hatching ...
                    for _user in _users:
                        self.delObject(_user)
                del self.__objdict[_seg]
                self.delObject(_p1)
                self.delObject(_p2)

    def _sweepCircleTree(self):
        """Validate the layer Circle storage.

_sweepCircleTree(spctree)

This method is private to the Layer.
        """
        _modcircs = self.__circle_tree.getModified()
        if len(_modcircs):
            _deadcircs = self.__circle_tree.clean()
            for _circ in _modcircs:
                if _circ in _deadcircs:
                    continue
                if _circ.getRadius() < 1e-10: # this probably can't happen
                    _deadcircs.append(_circ)
                else:
                    _circ.reset()
                    _cp = _circ.getCenter()
                    assert _cp in self.__objdict, "Circle center not in objdict:" + `_cp`
                    _add_flag = True
                    _users = self.__objdict[_cp]
                    if _users is not None:
                        for _user in _users:
                            if _user is _circ:
                                _add_flag = False
                                break
                        if _add_flag:
                            self._connectUser(_cp, _circ)
                    if self.__objdict[_circ] is not None:
                        self.__objdict[_circ].modified()
                    if _circ in self.__externals:
                        for _layer, _extobj in self.__externals[_circ]:
                            _extobj.modified()
            for _circ in _deadcircs:
                _keepcirc = self.__circle_tree.binscan(_circ)
                assert _keepcirc is not None, "Lost Circle: " + `_circ`
                if _circ in self.__externals:
                    for _layer, _extobj in self.__externals[_circ]:
                        _extobj.setDimCircle(_keepcirc)
                _users = self.__objdict[_circ]
                if _users is not None:
                    for _user in _users:
                        if isinstance(_user, dimension.RadialDimension):
                            _user.setDimCircle(_keepcirc)
                        else:
                            pass # fixme - hatching ...
                _cp = _circ.getCenter()
                self._disconnectUser(_cp, _circ)
                del self.__objdict[_circ]
                self.delObject(_cp)

    def _sweepArcTree(self):
        """Validate the layer Arc storage.

_sweepArcTree(spctree)

This method is private to the Layer.
        """
        _modarcs = self.__arc_tree.getModified()
        if len(_modarcs):
            _deadarcs = self.__arc_tree.clean()
            for _arc in _modarcs:
                if _arc in _deadarcs:
                    continue
                if _arc.getRadius() < 1e-10:
                    _deadarcs.append(_arc)
                else:
                    _arc.reset()
                    _cp = _arc.getCenter()
                    assert _cp in self.__objdict, "Arc center not in objdict:" + `_cp`
                    _add_flag = True
                    _users = self.__objdict[_cp]
                    if _users is not None:
                        for _user in _users:
                            if _user is _arc:
                                _add_flag = False
                                break
                        if _add_flag:
                            self._connectUser(_cp, _arc)
                    for _ep in _arc.getEndpoints():
                        _lp = self.hasPoint(_ep)
                        assert _lp in self.__objdict, "Arc endpoint not in objdict:" + `_lp`
                        _add_flag = True
                        _users = self.__objdict[_lp]
                        if _users is not None:
                            for _user in _users:
                                if _user is _arc:
                                    _add_flag = False
                                    break
                            if _add_flag:
                                self._connectUser(_lp, _arc)
                    if self.__objdict[_arc] is not None:
                        self.__objdict[_arc].modified()
                    if _arc in self.__externals:
                        for _layer, _extobj in self.__externals[_arc]:
                            _extobj.modified()
            for _arc in _deadarcs:
                _keeparc = self.__arc_tree.binscan(_arc)
                assert _keeparc is not None, "Lost Arc: " + `_arc`
                if _arc in self.__externals:
                    for _layer, _extobj in self.__externals[_arc]:
                        _extobj.setDimCircle(_keeparc)
                _users = self.__objdict[_arc]
                if _users is not None:
                    for _user in _users:
                        if isinstance(_user, dimension.RadialDimension):
                            _user.setDimCircle(_keeparc)
                        elif isinstance(_user, dimension.AngularDimension):
                            pass # fixme
                        else:
                            pass # fixme - hatching ...
                _cp = _arc.getCenter()
                self._disconnectUser(_cp, _arc)
                for _ep in _arc.getEndpoints():
                    _p = self.hasPoint(_ep)
                    self._disconnectUser(_p, _arc)
                    self.delObject(_p)
                del self.__objdict[_arc]
                self.delObject(_cp)

    def _sweepSPCLineTree(self, spctree):
        """Validate the storage of single point construction line objects.

_sweepSPCLineTree(spctree)

This method is private to the Layer.
        """
        _modspclines = spctree.getModified()
        if len(_modspclines):
            _deadspclines = spctree.clean()
            for _spcline in _modspclines:
                if _spcline not in _deadspclines:
                    _spcline.reset()
                    _pt =  _spcline.getLocation()
                    assert _pt in self.__objdict, "spcline point not in objdict:" + str(_spcline)
                    _add_flag = True
                    _users = self.__objdict[_pt]
                    assert _users is not None, "SPCLine not bound to location point"
                    for _user in _users:
                        if _user is _spcline:
                            _add_flag = False
                            break
                    if _add_flag:
                        self._connectUser(_pt, _spcline)
            for _spcline in _deadspclines:
                _pt = _spcline.getLocation()
                self._disconnectUser(_pt, _spcline)
                del self.__objdict[_spcline]
                self.delObject(_pt)

    def _sweepCLineTree(self):
        """Validate the layer CLine storage.

_sweepCLineTree()

This method is private to the Layer.
        """
        _modclines = self.__cline_tree.getModified()
        if len(_modclines):
            _deadclines = self.__segment_tree.clean()
            for _cline in _modclines:
                if _cline not in _deadclines:
                    _cline.reset()
                    for _pt in _cline.getKeypoints():
                        assert _pt in self.__objdict, "CLine keypoint not in objdict:" + `_pt`
                        _add_flag = True
                        _users = self.__objdict[_pt]
                        assert _users is not None, "CLine not bound to keypoint"
                        for _user in _users:
                            if _user is _cline:
                                _add_flag = False
                                break
                        if _add_flag:
                            self._connectUser(_pt, _cline)
            for _cline in _deadclines:
                _p1, _p2 = _cline.getKeypoints()
                self._disconnectUser(_p1, _cline)
                self._disconnectUser(_p2, _cline)
                del self.__objdict[_cline]
                self.delObject(_p1)
                self.delObject(_p2)

    def _sweepCCircleTree(self):
        """Validate the layer CCircle storage.

_sweepCCircleTree()

This method is private to the Layer object.
        """
        _modccs = self.__ccircle_tree.getModified()
        if len(_modccs):
            _deadccs = self.__ccircle_tree.clean()
            for _cc in _modccs:
                if _cc in _deadccs:
                    continue
                if _cc.getRadius() < 1e-10:
                    _deadccs.append(_cc)
                else:
                    _cc.reset()
                    _cp = _cc.getCenter()
                    assert _cp in self.__objdict, "CCircle center not in objdict:" + `_cp`
                    _add_flag = True
                    _users = self.__objdict[_cp]
                    assert _users is not None, "CCircle not bound to center"
                    for _user in _users:
                        if _user is _cc:
                            _add_flag = False
                            break
                    if _add_flag:
                        self._connectUser(_cp, _cc)
            for _cc in _deadccs:
                _cp = _cc.getCenter()
                self._disconnectUser(_cp, _cc)
                del self.__objdict[_cc]
                self.delObject(_cp)
