#
# Copyright (c) 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# classes for leader lines
#

import math
import array

import graphicobject
import tolerance
import style
import linetype
import color
import point
import tree
import maptree
import util

class _Leader(object):
    """A class representing a leader line.

A leader line is usually used as a visual connector between
some text and an entity in a drawing. Leader lines cannot be
used to define an edge of some shape.

A _Leader object has the following attributes:
p1: A _Point object representing the first end point.
p2: A _Point object representing the leader mid point.
p3: A _Point object representing the final end point.
arrowsize: The size of the arrow at the end of the leader line

A _Leader object has the following methods:

getLeaderPoints(): Return the points defining the _Leader
{get/set}P1(): Get/Set the _Leader first endpoint.
{get/set}P2(): Get/Set the _Leader midpoint
{get/set}P3(): Get/Set the _Leader final endpoint
{get/set}ArrowSize(): Get/Set the _Leader line arrow size.
calcArrowPoints(): Calculate where the _Leader arrow points are.
getArrowPoints(): Return where the _Leader arrow points are.
move(): Move the _Leader.
inRegion(): Test if the _Leader is visible in some area.
clone(): Make an identical copy of a _Leader.
onLeader(): See if a point lies within some distance of a _Leader.
    """

    # member functions
    
    def __init__(self, p1, p2, p3, size=1.0):
        """Initialize a _Leader object.

_Leader(p1, p2, p3[, size])

The following arguments are required:

p1: _Leader first endpoint - may be a _Point or a two-item tuple of floats
p2: _Leader mid point - may be a _Point or a two-item tuple of floats
p3: _Leader final endpoint - may be a _Point or a two-item tuple of floats

Argument "size" is optional. It gives the size of the arrow at
the end of the leader line, and defaults to 1.0.
        """
        _p1 = p1
        if not isinstance(_p1, point._Point):
            _p1 = point._Point(p1)
        _p2 = p2
        if not isinstance(_p2, point._Point):
            _p2 = point._Point(p2)
        if _p1 == _p2:
            raise ValueError, "_Leader objects cannot have equivalent points."
        _p3 = p3
        if not isinstance(_p3, point.Point):
            _p3 = point._Point(_p3)
        if _p1 == _p3:
            raise ValueError, "_Leader objects cannot have equivalent points."
        if _p2 == _p3:
            raise ValueError, "_Leader objects cannot have equivalent points."
        _size = size
        if not isinstance(_size, float):
            _size = float(size)
        if _size < 0.0:
            raise ValueError, "Invalid arrow size: %g" % _size
        self.__p1 = _p1
        self.__p2 = _p2
        self.__p3 = _p3
        self.__size = _size
        self.__arrow_pts = array.array('d', [0.0, 0.0, 0.0, 0.0])
        self.calcArrowPoints()
        
    def __str__(self):
        return "Leader: %s to %s to %s" % (self.__p1, self.__p2, self.__p3)

    def __eq__(self, obj):
        """Compare two leader lines for equality.
        """
        if not isinstance(obj, _Leader):
            raise TypeError, "Invalid leader comparison: " + `obj`
        _sp1 = self.__p1
        _sp2 = self.__p2
        _sp3 = self.__p3
        _p1, _p2, _p3 = obj.getLeaderPoints()
        _res = False
        if (_sp1 == _p1 and _sp2 == _p2 and _sp3 == _p3):
            _res = True
        return _res

    def __ne__(self, obj):
        """Compare two leader lines for inequality.
        """
        if not isinstance(obj, _Leader):
            raise TypeError, "Invalid leader comparison: " + `obj`
        _sp1 = self.__p1
        _sp2 = self.__p2
        _sp3 = self.__p3
        _p1, _p2, _p3 = obj.getLeaderPoints()
        _res = True
        if (_sp1 == _p1 and _sp2 == _p2 and _sp3 == _p3):
            _res = False
        return _res
    
    def __old_cmp__(self, obj):
        """Compare two _Leader lines.

FIXME:
_Leader comparison works as follows:
1) Get points of both leaders L1 and L2
2) Get minima and maxima values for the "x" and "y" coordinates

The return value is determined as follows:

minimum "x" of L1 < minimum "x" of L2 => return -1
maximum "x" of L1 > maximum "x" of L2 => return 1
minimum "y" of L1 < minimum "y" of L2 => return -1
maximum "y" of L1 > maximum "y" of L2 => return 1

Endpoint comparison now is needed to determine if
the Legments are equal.

if L1.p1.x == L2.p1.x and L1.p1.y == L2.p2.y or
   L1.p1.x == L2.p2.x and s1.p1.y == L2.p2.y ==> return 0
   
If the the above is not true, then the "x" coordinates of
the first endpoint of each segment are compared.

minimum "x" of L1E1 < minimum "x" of L2E1 => return -1
maximum "x" of L1E1 > maximum "x" of L2E1 => return 1

Legment/Point comparison works as follows:
1) Get endpoints of L and coordinates of P
2) Calculate minima and maxima of segment endpoints

The function returns the following

1: P.x < L.xmin
0: L.xmin <= P.x <= L.xmax and L.ymin <= P.y <= L.ymax
-1: Any other condition
        """
        if not isinstance(obj, (point._Point, _Leader)):
            raise TypeError, "Invalid object for _Leader comparison: " + `obj`
        _sx1, _sy1 = self.__p1.getCoords()
        _sx2, _sy2 = self.__p2.getCoords()
        _sx3, _sy3 = self.__p3.getCoords()
        _sxmin = min(_sx1, _sx2, _sx3)
        if isinstance(obj, _Leader):
            _p1, _p2, _p3 = obj.getLeaderPoints()
            _x1, _y1 = _p1.getCoords()
            _x2, _y2 = _p2.getCoords()
            _x3, _y3 = _p3.getCoords()
            _xmin = min(_x1, _x2, _x3)
            # print "sxmin = %g; xmin = %g" % (_sxmin, _xmin)
            if abs(_sxmin - _xmin) < 1e-10:
                _symin = min(_sy1, _sy2, _sy3)
                _ymin = min(_y1, _y2, _y3)
                # print "symin = %g; ymin = %g" % (_symin, _ymin)
                if abs(_symin - _ymin) < 1e-10:
                    _sxmax = max(_sx1, _sx2, _sx3)
                    _xmax = max(_x1, _x2, _x3)
                    # print "sxmax = %g; xmax = %g" % (_sxmax, _xmax)
                    if abs(_sxmax - _xmax) < 1e-10:
                        _symax = max(_sy1, _sy2, _sy3)
                        _ymax = max(_y1, _y2, _sy3)
                        # print "symax = %g; ymax = %g" % (_symax, _ymax)
                        if abs(_symax - _ymax) < 1e-10:
                            _x1diff = abs(_sx1 - _x1)
                            _y1diff = abs(_sy1 - _y1)
                            _x2diff = abs(_sx1 - _x2)
                            _y2diff = abs(_sy1 - _y2)
                            _x3diff = abs(_sx3 - _x3)
                            _y3diff = abs(_sy3 - _y3)
                            if ((_x1diff < 1e-10 and _y1diff < 1e-10) or
                                (_x2diff < 1e-10 and _y2diff < 1e-10)):
                                # print "common endpoints ..."
                                # print "self: " + str(self)
                                # print "obj: " + str(obj)
                                _val = 0
                            elif _sx1 > _x1:
                                _val = 1
                            else:
                                _val = -1
                        elif _symax > _ymax:
                            _val = 1
                        else:
                            _val = -1
                    elif _sxmax > _xmax:
                        _val = 1
                    else:
                        _val = -1
                elif _symin > _ymin:
                    _val = 1
                else:
                    _val = -1
            elif _sxmin > _xmin:
                _val = 1
            else:
                _val = -1
        else:
            _x, _y = obj.getCoords()
            if _x < _sxmin - 1e-10:
                _val = 1
            else:
                _val = -1
                _symin = min(_sy1, _sy2, _sy3) - 1e-10
                _sxmax = max(_sx1, _sx2, _sx3) + 1e-10
                _symax = max(_sy1, _sy2, _sy3) + 1e-10
                if _x < _sxmax and _symin < _y < _symax:
                    _val = 0
        return _val
        
    def getLeaderPoints(self):
        """Get the points defining the _Leader.

getLeaderPoints()

This function returns a tuple containing the three _Point objects
that are the defining points of the _Leader
        """
        return (self.__p1, self.__p2, self.__p3)

    def getP1(self):
        """Return the first endpoint _Point of the _Leader

getP1()        
        """
        return self.__p1

    def setP1(self, p):
        """Set the first endpoint _Point of the _Leader

setP1(p)

The argument "p" should be a _Point.
        """
        if not isinstance(p, point._Point):
            raise TypeError, "Invalid _Point for p1 endpoint: " + `p`
        self.__p1 = p
    
    p1 = property(getP1, setP1, None, "First endpoint of the Leader.")

    def getP2(self):
        """Return the midpoint _Point of the _Leader.

getP2()        
        """
        return self.__p2

    def setP2(self, p):
        """Set the midpoint _Point of the _Leader.

setP2(p)

The argument "p" should be a _Point.
        """
        if not isinstance(p, point._Point):
            raise TypeError, "Invalid Point for p2 endpoint: " + `p`
        self.__p2 = p

    p2 = property(getP2, setP2, None, "Leader midpoint.")

    def getP3(self):
        """Return the final _Point of the _Leader.

getP3()        
        """
        return self.__p3

    def setP3(self, p):
        """Set the final endpoint _Point of the _Leader.

setP3(p)

The argument "p" should be a _Point.
        """
        if not isinstance(p, point._Point):
            raise TypeError, "Invalid Point for p3 endpoint: " + `p`
        self.__p3 = p

    p3 = property(getP3, setP3, None, "Third endpoint of the Leader.")

    def getArrowSize(self):
        """Return the size of the leader line arrow.

getArrowSize()
        """
        return self.__size

    def setArrowSize(self, size):
        """Set the size of the leader line arrow.

setSize(size)

Argument "size" should be a float greater than or
equal to 0.0.
        """
        _size = size
        if not isinstance(_size, float):
            _size = float(size)
        if _size < 0.0:
            raise ValueError, "Invalid arrow size: %g" % _size
        self.__size = _size
        self.calcArrowPoints()

    arrowsize = property(getArrowSize, setArrowSize,
                    None, "_Leader line arrow size.")

    def calcArrowPoints(self):
        """Calculate where the _Leader arrow points are.

calcArrowPoints()
        """
        _x1, _y1 = self.__p2.getCoords()
        _x2, _y2 = self.__p3.getCoords()
        if abs(_x2 - _x1) < 1e-10: # vertical
            if _y2 > _y1:
                _sine = 1.0
                _cosine = 0.0
            else:
                _sine = -1.0
                _cosine = 0.0
        elif abs(_y2 - _y1) < 1e-10: # horizontal
            if _x2 > _x1:
                _sine = 0.0
                _cosine = 1.0
            else:
                _sine = 0.0
                _cosine = -1.0
        else:
            _angle = math.atan2((_y2 - _y1), (_x2 - _x1))
            _sine = math.sin(_angle)
            _cosine = math.cos(_angle)
        _size = self.__size
        _height = _size/5.0
        # p1 -> (x,y) = (-size, _height)
        self.__arrow_pts[0] = (_cosine * (-_size) - _sine * _height) + _x2
        self.__arrow_pts[1] = (_sine * (-_size) + _cosine * _height) + _y2
        # p2 -> (x,y) = (-size, -_height)
        self.__arrow_pts[2] = (_cosine * (-_size) - _sine *(-_height)) + _x2
        self.__arrow_pts[3] = (_sine * (-_size) + _cosine *(-_height)) + _y2

    def getArrowPoints(self):
        """Return the endpoints of the _Leader arrow.

getArrowPoints()

This method returns an array holding four float values.
        """
        return self.__arrow_pts
    
    def move(self, dx, dy):
        """Move a _Leader.

move(dx, dy)

The first argument gives the "x" coordinate displacement,
and the second gives the "y" coordinate displacement. Both
values should be floats.
        """
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        self.__p1.move(_dx, _dy)
        self.__p2.move(_dx, _dy)
        self.__p3.move(_dx, _dy)
        self.calcArrowPoints()

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest _Point on the _Leader to the point p.

mapPoint(p [, tol])

The function has one required argument:

p: A _Point object or a tuple of two floats

There is a single optional argument:

tol: A float value equal or greater than 0.

This function is used to map a possibly near-by _Point object to an
actual _Point on the _Leader. If the distance between the actual
_Point and the _Point used as an argument is less than the Tolerance,
the actual _Point is returned. Otherwise, this function returns None.
        """
        _p = p
        if not isinstance(_p, point._Point):
            _p = point._Point(p)
        _t = tolerance.toltest(tol)
        _xp, _yp = _p.getCoords()
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _x3, _y3 = self.__p3.getCoords()
        _xmin = min(_x1, _x2, _x3) - _t - 1e-10
        _ymin = min(_y1, _y2, _y3) - _t - 1e-10
        _ap = None
        if _xp > _xmin and _yp > _ymin:
            _xmax = max(_x1, _x2, _x3) + _t + 1e-10
            _ymax = max(_y1, _y2, _y3) + _t + 1e-10
            if _xp < _xmax and _yp < _ymax:
                _sqlen = pow((_x2 - _x1), 2) + pow((_y2 - _y1), 2)
                _r = ((_xp - _x1)*(_x2 - _x1) + (_yp - _y1)*(_y2 - _y1))/_sqlen
                if _r < 0.0 : _r = 0.0
                if _r > 1.0 : _r = 1.0
                _px = _x1 + _r * (_x2 - _x1)
                _py = _y1 + _r * (_y2 - _y1)
                _xsep = abs(_px - _xp)
                _ysep = abs(_py - _yp)
                if ((_xsep < _t or abs(_xsep - _t) < 1e-10) and
                    (_ysep < _t or abs(_ysep - _t) < 1e-10)):
                    _ap = point._Point(_px, _py)
                if _ap is None:
                    _sqlen = pow((_x3 - _x2), 2) + pow((_y3 - _y2), 2)
                    _r = ((_xp - _x2)*(_x3 - _x2) + (_yp - _y2)*(_y3 - _y2))/_sqlen
                    if _r < 0.0 : _r = 0.0
                    if _r > 1.0 : _r = 1.0
                    _px = _x2 + _r * (_x3 - _x2)
                    _py = _y2 + _r * (_y3 - _y2)
                    _xsep = abs(_px - _xp)
                    _ysep = abs(_py - _yp)
                    if ((_xsep < _t or abs(_xsep - _t) < 1e-10) and
                        (_ysep < _t or abs(_ysep - _t) < 1e-10)):
                        _ap = point._Point(_px, _py)
        return _ap

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a _Leader exists within a region.

isRegion(xmin, ymin, xmax, ymax[, fully])

The four arguments define the boundary of an area, and the
method returns True if the _Leader lies within that area. If
the optional argument "fully" is used and is True, then all
the _Leader points must lie within the boundary. Otherwise,
the method returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        # print "xmin: %g" % _xmin
        # print "ymin: %g" % _ymin
        # print "xmax: %g" % _xmax
        # print "ymax: %g" % _ymax
        _x1, _y1 = self.__p1.getCoords()
        _x2, _y2 = self.__p2.getCoords()
        _x3, _y3 = self.__p3.getCoords()
        # print "x1: %g" % _x1
        # print "y1: %g" % _y1
        # print "x2: %g" % _x2
        # print "y2: %g" % _y2
        # print "x3: %g" % _x3
        # print "y3: %g" % _y3
        _pxmin = min(_x1, _x2, _x3)
        _pymin = min(_y1, _y2, _y3)
        _pxmax = max(_x1, _x2, _x3)
        _pymax = max(_y1, _y2, _y3)
        # print "pxmin: %g" % _pxmin
        # print "pymin: %g" % _pymin
        # print "pxmax: %g" % _pxmax
        # print "pymax: %g" % _pymax
        _val = False
        if (_xmin < _pxmin and _xmax > _pxmax and
            _ymin < _pymin and _ymax > _pymax):
            _val = True
        elif (not fully and
              not (_pxmax < _xmin or _pymax < _ymin or
                   _pxmin > _xmax or _pymin > _ymax)):
            if _xmin < _x1 < _xmax and _ymin < _y1 < _ymax:
                _val = True
            elif _xmin < _x2 < _xmax and _ymin < _y2 < _ymax:
                _val = True
            elif _xmin < _x3 < _xmax and _ymin < _y3 < _ymax:
                _val = True
            else:
                _val = False # fixme
        return _val

    def clone(self):
        """Create an identical copy of a _Leader.

clone()
        """
        _cp1 = self.__p1.clone()
        _cp2 = self.__p2.clone()
        _cp3 = self.__p3.clone()
        _size = self.__size
        return _Leader(_cp1, _cp2, _cp3, _size)

#
# Leader line default style bits
#

def_linetype = linetype.Linetype('Solid_Default', None)
def_color = color.Color(0xcc9933)
def_thickness = 1.0

def_leader_style = style.Style("Leader_Default", def_linetype,
                               def_color, def_thickness)

class Leader(_Leader, tree.TreeItem, graphicobject.GraphicObject):
    """The storable and displayed representation of _Leader objects.

The Leader class is derived from _Leader, treeItem, and
GraphicObject classes, so it shares all the attributes and
methods of those classes.
    """
    def __init__(self, p1, p2, p3, size=1.0,
                 st=def_leader_style, lt=None, col=None, t=None):
        _p1 = p1
        if not isinstance(_p1, point.Point):
            _p1 = point.Point(p1)
        _p2 = p2
        if not isinstance(_p2, point.Point):
            _p2 = point.Point(p2)
        if _p1 == _p2:
            raise ValueError, "Leader objects cannot have equivalent points."
        _p3 = p3
        if not isinstance(_p3, point.Point):
            _p3 = point.Point(p3)
        if _p1 == _p3:
            raise ValueError, "Leader objects cannot have equivalent points."
        if _p2 == _p3:
            raise ValueError, "Leader objects cannot have equivalent points."
        _size = size
        if not isinstance(_size, float):
            _size = float(size)
        if _size < 0.0:
            raise ValueError, "Invalid Leader arrow size: %g" % _size
        _Leader.__init__(self, _p1, _p2, _p3, _size)
        tree.TreeItem.__init__(self)
        graphicobject.GraphicObject.__init__(self, st, lt, col, t)


    def isModified(self):
        """Indicate if the Leader has been changed.

isModified()

This method extends the TreeItem::isModified() method.
        """
        _p1, _p2, _p3 = self.getLeaderPoints()
        if _p1.isModified() or _p2.isModified() or _p3.isModified():
            self.modified()
        return tree.TreeItem.isModified(self)
    
    def setP1(self, p):
        """Set the first endpoint Point of the Leader

setP1(p)

This method extends _Leader::setP1() and sets the modified
flag indicating that the Leader has been altered.
        """
        _Leader.setP1(self, p)
        self.modified()
    
    p1 = property(_Leader.getP1, setP1, None,
                  "First endpoint of the Leader.")

    def setP2(self, p):
        """Set the midpoint Point of the Leader.

setP2(p)

This method extends _Leader::setP2() and sets the modified
flag indicating that the Leader has been altered.
        """
        _Leader.setP2(self, p)
        self.modified()
    
    p2 = property(_Leader.getP2, setP2, None,
                  "Midpoint of the Leader.")

    def setP3(self, p):
        """Set the final endpoint Point of the Leader.

setP3(p)

This method extends _Leader::setP3() and sets the modified
flag indicating that the Leader has been altered.
        """
        _Leader.setP3(self, p)
        self.modified()
    
    p3 = property(_Leader.getP3, setP3, None,
                  "Final endpoint of the Leader.")                  

    def move(self, dx, dy):
        """Move a Leader.

move(dx, dy)

This method extends _Leader::move() and sets the modified
flag indicating that the Leader has been altered.
        """
        _Leader.move(self, dx, dy)
        self.modified()

    def clone(self):
        """Create an identical copy of a Leader.

clone()

This method overrides the _Leader::clone() method.
        """
        _p1, _p2, _p3 = self.getLeaderPoints()
        _cp1 = _p1.clone()
        _cp2 = _p2.clone()
        _cp3 = _p3.clone()
        _size = self.getArrowSize()
        _st = self.getStyle()
        _lt = self.getLinetype()
        _col = self.getColor()
        _th = self.getThickness()
        return Leader(_cp1, _cp2, _cp3, _size, _st, _lt, _col, _th)

    def mapPoint(self, p, tol=tolerance.TOL):
        """Return the nearest Point on the Leader to the point p.

mapPoint(p [, tol])

This function extends _Leader::mapPoint().
        """
        _pt = _Leader.mapPoint(self, p, tol)
        if _pt is not None:
            _x, _y = _pt.getCoords()
            _pt = point.Point(_x, _y)
        return _pt

class LeaderTree(maptree.PointMapTree):
    """A Tree class specific for Leader objects.

This class is based on the PointMapTree class, and holds
only Leader objects.
    """
    def __init__(self):
        """Initialize a LeaderTree.

LeaderTree()
        """
        maptree.PointMapTree.__init__(self, Leader)

    def binscan(self, obj, tol=tolerance.TOL):
        """Find a Leader object held in a LeaderTree via binary search.

binscan(obj [, tol])

This method has one required parameter:

obj: Either a Leader object or a tuple of two-points

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     Leaders are close enough to be considered the same.

If a Leader object is found in the LeaderTree within the
tolerance, the index of that object in the LeaderTree is
returned. Otherwise, the function returns None.
        """
        _o = obj
        if isinstance(_o, Leader):
            _p1, _p2, _p3 = _o.getLeaderPoints()
        elif isinstance(_o, tuple):
            _l = len(_o)
            if _l != 3:
                raise ValueError, "Tuple must contain exactly three objects."
            _p1 = _o[0]
            if not isinstance(_p1, point.Point):
                _p1 = point.Point(_o[0])
            _p2 = _o[1]
            if not isinstance(_p2, point.Point):
                _p2 = point.Point(_o[1])
            _p3 = _o[2]
            if not isinstance(_p3, point.Point):
                _p3 = point.Point(_o[3])
        else:
            raise TypeError, "Invalid type for LeaderTree search: " + `obj`
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "LeaderTree in modified state."
        _ro = None
        if len(self):
            _xmin = min(_p1.x, _p2.x, _p3.x)
            _ymin = min(_p1.y, _p2.y, _p3.y)
            _xmax = max(_p1.x, _p2.x, _p3.x)
            _ymax = max(_p1.y, _p2.y, _p3.y)
            # print "xmin: %g" % _xmin
            # print "ymin: %g" % _ymin
            # print "xmax: %g" % _xmax
            # print "ymax: %g" % _ymax
            _scanlist = []
            _lo = 0
            _hi = len(self)
            # print "lo: 0; hi: %d" % _hi
            while _lo < _hi:
                _mid = (_hi+_lo)//2
                if _mid in _scanlist:
                    break
                _scanlist.append(_mid)
                # print "testing %d" % _mid
                _sp1, _sp2, _sp3 = self[_mid].getLeaderPoints()
                # print _sp1
                # print _sp2
                # print _sp3
                _sxmin = min(_sp1.x, _sp2.x, _sp3.x)
                # print "sxmin: %g" % _sxmin
                if (_xmin - _sxmin) > -1e-10:
                    _lo = _mid + 1
                else:
                    _hi = _mid
                    # print "lo: %d; hi: %d" % (_lo, _hi)
            for _i in xrange(_mid, -1, -1):
                # print "index: %d" % _i
                _leader = self[_i]
                _sp1, _sp2, _sp3 = _leader.getLeaderPoints()
                _sxmin = min(_sp1.x, _sp2.x, _sp3.x)
                if _sxmin < _xmin - _t - 1e-10:
                    break
                if abs(_sxmin - _xmin) < _t + 1e-10:
                    _symin = min(_sp1.y, _sp2.y, _sp3.y)
                    if abs(_symin - _ymin) < _t + 1e-10:
                        _sxmax = max(_sp1.x, _sp2.x, _sp3.x)
                        if abs(_sxmax - _xmax) < _t + 1e-10:
                            _symax = max(_sp1.y, _sp2.y, _sp3.y)
                            if abs(_symax - _ymax) < _t + 1e-10:
                                _ro = _leader
                                break
        return _ro

    def scan(self, obj, tol=tolerance.TOL):
        """Find a Leader object held in a LeaderTree via linear search.

scan(obj [, tol])

This method has one required parameter:

obj: Either a Leader object or a tuple of two-points

There is a single optional argument:

tol: A float equal or greater than 0 used to decide if two
     Leaders are close enough to be considered equal.

If a Leader object is found in the LeaderTree within the
tolerance, the index of that object in the LeaderTree is
returned. Otherwise, the function returns None.
        """
        _o = obj
        if isinstance(_o, Leader):
            _p1, _p2 = _o.getEndpoints()
        elif isinstance(_o, tuple):
            _l = len(_o)
            if _l != 2:
                raise ValueError, "Tuple must contain exactly two objects."
            _p1 = _o[0]
            if not isinstance(_p1, point.Point):
                _p1 = point.Point(_o[0])
            _p2 = _o[1]
            if not isinstance(_p2, point.Point):
                _p2 = point.Point(_o[1])
        else:
            raise TypeError, "Invalid type for LeaderTree search: " + `obj`
        _t = tolerance.toltest(tol)
        if self.isModified():
            raise ValueError, "LeaderTree in modified state."
        _xmin = min(_p1.x, _p2.x)
        _ymin = min(_p1.y, _p2.y)
        _xmax = max(_p1.x, _p2.x)
        _ymax = max(_p1.y, _p2.y)
        _ro = None
        for _sobj in self:
            _sp1, _sp2 = _sobj.getEndpoints()
            _sxmin = min(_sp1.x, _sp2.x)
            if _sxmin > _xmin + _t + 1e-10:
                break
            if abs(_sxmin - _xmin) < _t + 1e-10:
                _symin = min(_sp1.y, _sp2.y)
                if abs(_symin - _ymin) < _t + 1e-10:
                    _sxmax = max(_sp1.x, _sp2.x)
                    if abs(_sxmax - _xmax) < _t + 1e-10:
                        _symax = max(_sp1.y, _sp2.y)
                        if abs(_symax - _ymax) < _t + 1e-10:
                            _ro = _sobj
                            break
        return _ro
