#
# Copyright (c) 2002, Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# handle setting and retrieving of the various options
# and preferences in an image
#
# none of these functions should be called directly - they
# are all meant as private functions for an Image object

import units
import dimension
import text
import style
import color
import linetype

#
# generic type tests
#

def _test_boolean_value(optdict, key, value):
    if (value is True or
        value is False or
        value == 0 or
        value == 1): # these should go away ...
        optdict[key] = value
    else:
        raise ValueError, "Invalid value: " + `value`
    
def _test_int_value(optdict, key, value, min=None, max=None):
    _v = value
    if not isinstance(_v, int):
        _v = int(value)
    if min is not None:
        assert isinstance(min, int), "Invalid minimum value: " + `min`
        if _v < min:
            raise ValueError, "Invalid value: " + `_v`
    if max is not None:
        assert isinstance(max, int), "Invalid maximum value: " + `max`
        if _v > max:
            raise ValueError, "Invalid value: " + `_v`
    optdict[key] = _v

def _test_float_value(optdict, key, value, min=None, max=None):
    _v = value
    if not isinstance(_v, float):
        _v = float(value)
    if min is not None:
        assert isinstance(min, float), "Invalid minimum value: " + `min`
        if _v < min:
            raise ValueError, "Invalid value: " + `_v`
    if max is not None:
        assert isinstance(max, float), "Invalid maximum value: " + `max`
        if _v > max:
            raise ValueError, "Invalid value: " + `_v`
    optdict[key] = _v

def _test_unicode_value(optdict, key, value):
    _u = value
    if not isinstance(_u, unicode):
        _u = unicode(value)
    optdict[key] = _u
    
def _test_color_value(optdict, key, value):
    if not isinstance(value, color.Color):
        raise TypeError, "Invalid color: " + `value`
    optdict[key] = value

def _test_font_family(optdict, key, value):
    if (value is None or
        (not isinstance(value, str)) or
        value == ''):
        raise ValueError, "Invalid font family: " + `value`
    optdict[key] = value
    
def _test_font_weight(optdict, key, value):
    if (value != text.NORMAL and
        value != text.LIGHT and
        value != text.BOLD and
        value != text.HEAVY):
        raise ValueError, "Invalid font weight: " + `value`
    optdict[key] = value

def _test_font_style(optdict, key, value):
    if (value != text.NORMAL and
        value != text.OBLIQUE and
        value != text.ITALIC):
        raise ValueError, "Invalid font style: " + `value`
    optdict[key] = value

def _test_units(optdict, key, value):
    if (value != units.MILLIMETERS and
        value != units.MICROMETERS and
        value != units.METERS and
        value != units.KILOMETERS and
        value != units.INCHES and
        value != units.FEET and
        value != units.YARDS and
        value != units.MILES):
        raise ValueError, "Invalid unit: " + `value`
    optdict[key] = value
    
#
# dimension related tests
#

def _set_dim_primary_font_family(optdict, opt, family):
    _test_font_family(optdict, opt, family)

def _set_dim_primary_font_size(optdict, opt, size):
    _test_int_value(optdict, opt, size, min=1)

def _set_dim_primary_font_style(optdict, opt, st):
    _test_font_style(optdict, opt, st)

def _set_dim_primary_font_weight(optdict, opt, weight):
    _test_font_weight(optdict, opt, weight)

def _set_dim_primary_font_color(optdict, opt, col):
    _test_color_value(optdict, opt, col)

def _set_dim_primary_prefix(optdict, opt, pfx):
    _test_unicode_value(optdict, opt, pfx)

def _set_dim_primary_suffix(optdict, opt, pfx):
    _test_unicode_value(optdict, opt, pfx)

def _set_dim_primary_units(optdict, opt, u):
    _test_units(optdict, opt, u)
    
def _set_dim_primary_precision(optdict, opt, prec):
    _test_int_value(optdict, opt, prec, min=0, max=15)

def _set_dim_primary_print_zero(optdict, opt, flag):
    _test_boolean_value(optdict, opt, flag)

def _set_dim_primary_trail_decimal(optdict, opt, flag):
    _test_boolean_value(optdict, opt, flag)

def _set_dim_secondary_font_family(optdict, opt, family):
    _test_font_family(optdict, opt, family)

def _set_dim_secondary_font_size(optdict, opt, size):
    _test_int_value(optdict, opt, size, min=1)

def _set_dim_secondary_font_style(optdict, opt, st):
    _test_font_style(optdict, opt, st)

def _set_dim_secondary_font_weight(optdict, opt, weight):
    _test_font_weight(optdict, opt, weight)

def _set_dim_secondary_font_color(optdict, opt, col):
    _test_color_value(optdict, opt, col)

def _set_dim_secondary_prefix(optdict, opt, pfx):
    _test_unicode_value(optdict, opt, pfx)

def _set_dim_secondary_suffix(optdict, opt, pfx):
    _test_unicode_value(optdict, opt, pfx)

def _set_dim_secondary_units(optdict, opt, u):
    _test_units(optdict, opt, u)
    
def _set_dim_secondary_precision(optdict, opt, prec):
    _test_int_value(optdict, opt, prec, min=0, max=15)

def _set_dim_secondary_print_zero(optdict, opt, flag):
    _test_boolean_value(optdict, opt, flag)

def _set_dim_secondary_trail_decimal(optdict, opt, flag):
    _test_boolean_value(optdict, opt, flag)
    
def _set_dim_offset(optdict, opt, offset):
    _test_float_value(optdict, opt, offset, min=0.0)

def _set_dim_extension(optdict, opt, ext):
    _test_float_value(optdict, opt, ext, min=0.0)

def _set_dim_color(optdict, opt, col):
    _test_color_value(optdict, opt, col)

def _set_dim_dual_mode(optdict, opt, mode):
    _test_boolean_value(optdict, opt, mode)
    
def _set_dim_position(optdict, opt, pos):
    if (pos != dimension.Dimension.SPLIT and
        pos != dimension.Dimension.ABOVE and
        pos != dimension.Dimension.BELOW):
        raise ValueError, "Invalid dimension position: " + `pos`
    optdict[opt] = pos

def _set_dim_endpoint(optdict, opt, ept):
    if (ept != dimension.Dimension.NO_ENDPOINT and
        ept != dimension.Dimension.ARROW and
        ept != dimension.Dimension.FILLED_ARROW and
        ept != dimension.Dimension.SLASH and
        ept != dimension.Dimension.CIRCLE):
        raise ValueError, "Invalid endpoint value: " + `ept`
    optdict[opt] = ept

def _set_dim_endpoint_size(optdict, opt, size):
    _test_float_value(optdict, opt, size, min=0.0)

def _set_radial_dim_primary_prefix(optdict, opt, pfx):
    _test_unicode_value(optdict, opt, pfx)

def _set_radial_dim_primary_suffix(optdict, opt, sfx):
    _test_unicode_value(optdict, opt, sfx)

def _set_radial_dim_secondary_prefix(optdict, opt, sfx):
    _test_unicode_value(optdict, opt, sfx)

def _set_radial_dim_secondary_suffix(optdict, opt, pfx):
    _test_unicode_value(optdict, opt, pfx)

def _set_radial_dim_dia_mode(optdict, opt, flag):
    _test_boolean_value(optdict, opt, flag)
    
def _set_angular_dim_primary_prefix(optdict, opt, pfx):
    _test_unicode_value(optdict, opt, pfx)

def _set_angular_dim_primary_suffix(optdict, opt, sfx):
    _test_unicode_value(optdict, opt, sfx)

def _set_angular_dim_secondary_prefix(optdict, opt, pfx):
    _test_unicode_value(optdict, opt, pfx)

def _set_angular_dim_secondary_suffix(optdict, opt, sfx):
    _test_unicode_value(optdict, opt, sfx)

def _set_angular_dim_small_angle_mode(optdict, opt, flag):
    _test_boolean_value(optdict, opt, flag)
                                    
#
# text related tests
#

def _set_font_family(optdict, opt, family):
    _test_font_family(optdict, opt, family)

def _set_font_style(optdict, opt, st):
    _test_font_style(optdict, opt, st)

def _set_font_weight(optdict, opt, weight):
    _test_font_weight(optdict, opt, weight)

def _set_font_size(optdict, opt, size):
    _test_int_value(optdict, opt, size, min=1)

def _set_font_color(optdict, opt, col):
    _test_color_value(optdict, opt, col)
    
def _set_chamfer_length(optdict, opt, length):
    _test_float_value(optdict, opt, length, min=0.0)

def _set_fillet_radius(optdict, opt, radius):
    _test_float_value(optdict, opt, radius)

def _set_line_style(optdict, opt, st):
    if not isinstance(st, style.Style):
        raise TypeError, "Invalid line style: " + `st`
    optdict[opt] = st
    optdict['LINE_COLOR'] = st.getColor()
    optdict['LINE_TYPE'] = st.getLinetype()
    optdict['LINE_THICKNESS'] = st.getThickness()

def _set_line_color(optdict, opt, col):
    _test_color_value(optdict, opt, col)

def _set_line_type(optdict, opt, lt):
    if not isinstance(lt, linetype.Linetype):
        raise TypeError, "Invalid line type: " + `lt`
    optdict[opt] = lt

def _set_line_thickness(optdict, opt, thickness):
    _test_float_value(optdict, opt, thickness, min=0.0)

def _set_units(optdict, opt, u):
    _test_units(optdict, opt, u)

def _set_highlight_points(optdict, opt, flag):
    _test_boolean_value(optdict, opt, flag)

def _set_inactive_layer_color(optdict, opt, col):
    _test_color_value(optdict, opt, col)

def _set_background_color(optdict, opt, col):
    _test_color_value(optdict, opt, col)

def _set_autosplit(optdict, opt, col):
    _test_boolean_value(optdict, opt, col)
    
test_dict = {
    'DIM_PRIMARY_FONT_FAMILY' : _set_dim_primary_font_family,
    'DIM_PRIMARY_FONT_SIZE' : _set_dim_primary_font_size,
    'DIM_PRIMARY_FONT_WEIGHT' : _set_dim_primary_font_weight,
    'DIM_PRIMARY_FONT_STYLE' : _set_dim_primary_font_style,
    'DIM_PRIMARY_FONT_COLOR' : _set_dim_primary_font_color,
    'DIM_PRIMARY_PREFIX' : _set_dim_primary_prefix,
    'DIM_PRIMARY_SUFFIX' : _set_dim_primary_suffix,
    'DIM_PRIMARY_PRECISION' : _set_dim_primary_precision,
    'DIM_PRIMARY_UNITS' : _set_dim_primary_units,
    'DIM_PRIMARY_LEADING_ZERO' : _set_dim_primary_print_zero,
    'DIM_PRIMARY_TRAILING_DECIMAL': _set_dim_primary_trail_decimal,
    'DIM_SECONDARY_FONT_FAMILY' : _set_dim_secondary_font_family,
    'DIM_SECONDARY_FONT_SIZE' : _set_dim_secondary_font_size,
    'DIM_SECONDARY_FONT_WEIGHT' : _set_dim_secondary_font_weight,
    'DIM_SECONDARY_FONT_STYLE' : _set_dim_secondary_font_style,
    'DIM_SECONDARY_FONT_COLOR' : _set_dim_secondary_font_color,
    'DIM_SECONDARY_PREFIX' : _set_dim_secondary_prefix,
    'DIM_SECONDARY_SUFFIX' : _set_dim_secondary_suffix,
    'DIM_SECONDARY_PRECISION' : _set_dim_secondary_precision,
    'DIM_SECONDARY_UNITS' : _set_dim_secondary_units,
    'DIM_SECONDARY_LEADING_ZERO' : _set_dim_secondary_print_zero,
    'DIM_SECONDARY_TRAILING_DECIMAL': _set_dim_secondary_trail_decimal,
    'DIM_OFFSET' : _set_dim_offset,
    'DIM_EXTENSION': _set_dim_extension,
    'DIM_COLOR' : _set_dim_color,
    'DIM_POSITION': _set_dim_position,
    'DIM_ENDPOINT': _set_dim_endpoint,
    'DIM_ENDPOINT_SIZE' : _set_dim_endpoint_size,
    'DIM_DUAL_MODE' : _set_dim_dual_mode,
    'RADIAL_DIM_PRIMARY_PREFIX' : _set_radial_dim_primary_prefix,
    'RADIAL_DIM_PRIMARY_SUFFIX' : _set_radial_dim_primary_suffix,
    'RADIAL_DIM_SECONDARY_PREFIX' : _set_radial_dim_secondary_prefix,
    'RADIAL_DIM_SECONDARY_SUFFIX' : _set_radial_dim_secondary_suffix,
    'RADIAL_DIM_DIA_MODE' : _set_radial_dim_dia_mode,
    'ANGULAR_DIM_PRIMARY_PREFIX' : _set_angular_dim_primary_prefix,
    'ANGULAR_DIM_PRIMARY_SUFFIX' : _set_angular_dim_primary_suffix,
    'ANGULAR_DIM_SECONDARY_PREFIX' : _set_angular_dim_secondary_prefix,
    'ANGULAR_DIM_SECONDARY_SUFFIX' : _set_angular_dim_secondary_suffix,
    'FONT_FAMILY' : _set_font_family,
    'FONT_STYLE' : _set_font_style,
    'FONT_WEIGHT' : _set_font_weight,
    'FONT_SIZE' : _set_font_size,
    'FONT_COLOR' : _set_font_color,
    'CHAMFER_LENGTH' : _set_chamfer_length,
    'FILLET_RADIUS' : _set_fillet_radius,
    'UNITS' : _set_units,
    'LINE_STYLE' : _set_line_style,
    'LINE_COLOR' : _set_line_color,
    'LINE_TYPE' : _set_line_type,
    'LINE_THICKNESS': _set_line_thickness,
    'HIGHLIGHT_POINTS' : _set_highlight_points,
    'INACTIVE_LAYER_COLOR' : _set_inactive_layer_color,
    'BACKGROUND_COLOR' : _set_background_color,
    'AUTOSPLIT' : _set_autosplit,
    }

def test_option(optdict, opt, val):
    if test_dict.has_key(opt):
        test_dict[opt](optdict, opt, val)
    else:
        optdict[opt] = val
