#
# Copyright (c) 2002, Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# basic text functionality
#

import color

#
# common constant for style and weight
#

NORMAL = 0

#
# styles
#

OBLIQUE = 1
ITALIC = 2

#
# weights
#

LIGHT = 1
BOLD = 2
HEAVY = 3

#
# default text color
#

text_color = color.Color(255,255,255)

class TextStyle(object):
    """A class for describing text string properties.

A TextStyle object has the following attributes:

family: The font family
style: The font style
size: The font size
weight: The font weight
color: The font color

A TextStyle object has the following methods:

{get/set}Family(): Get/Set the font family.
{get/set}Size(): Get/Set the font size.
{get/set}Style(): Get/Set the font style.
{get/set}Weight(): Get/Set the font weight.
{get/set}Color(): Get/Set the font color.
    """
    def __init__(self, family="Sans", size=12, style=NORMAL, weight=NORMAL,
                 col=text_color):
        """Initialize a TextStyle object.

ts = TextStyle(family, size, style, weight, col)

Each argument has a default value:

family: Sans
size: 12
style: NORMAL
weight: NORMAL
col: A default color (white)
        """
        _f = family
        if _f is None: # need better error checking ...
            raise ValueError, "The font family cannot be None."
        _sz = size
        if not isinstance(_sz, int):
            _sz = int(size)
        if _sz < 1:
            raise ValueError, "Invalid text size: " + `_sz`
        _st = style
        if (_st != NORMAL and
            _st != OBLIQUE and
            _st != ITALIC):
            raise ValueError, "Invalid text style: " + `_st`
        _w = weight
        if (_w != NORMAL and
            _w != LIGHT and
            _w != BOLD and
            _w != HEAVY):
            raise ValueError, "Invalid text weight: " + `_w`
        if not isinstance(col, color.Color):
            raise TypeError, "Invalid color: " + `col`
        self.__family = _f
        self.__size = _sz
        self.__style = _st
        self.__weight = _w
        self.__color = col

    def getFamily(self):
        """Return the font family.

getFamily()        
        """
        return self.__family

    def setFamily(self, family): # what should default be?
        """Set the font family.

setFamily(family)
        """
        if not isinstance(family, str):
            raise
        self. __family = family

    family = property(getFamily, setFamily, None, "Text font family")

    def getSize(self):
        """Return the font size.

getSize()        
        """
        return self.__size

    def setSize(self, size=12):
        """Set the font size.

setSize([size])

The argument "size" should be an integer value
greater than 0. If this method is called without
arguments, the font size defaults to 12.
        """
        _s = size
        if not isinstance(_s, int):
            _s = int(size)
        if _s < 1:
            raise ValueError, "Invalid text size: " + `_s`
        self.__size = _s

    size = property(getSize, setSize, None, "Text font size")

    def getStyle(self):
        """Return the font style.

getStyle()        
        """
        return self.__style

    def setStyle(self, style=NORMAL):
        """Set the font style.

setStyle([style])

The argument "style" should be one of the following:

NORMAL
OBLIQUE
ITALIC

If the method is invoked without an argument, the font
style is set to NORMAL.
        """
        if (style != NORMAL and
            style != OBLIQUE and
            style != ITALIC):
            raise ValueError, "Invalid text style: " + `style`
        self.__style = style

    style = property(getStyle, setStyle, None, "Text font style")

    def getWeight(self):
        """Return the font weight.

getWeight()        
        """
        return self.__weight

    def setWeight(self, weight=NORMAL):
        """Set the font weight.

setWeight([weight])

The argument "weight" should be one of the following:

NORMAL
LIGHT
BOLD
HEAVY

If the method is called without an argument, the weight
is set to NORMAL.
        """
        if (weight != NORMAL and
            weight != LIGHT and
            weight != BOLD and
            weight != HEAVY):
            raise ValueError, "Invalid text weight: " + `weight`
        self.__weight = weight

    weight = property(getWeight, setWeight, None, "Text font weight")

    def getColor(self):
        """Return the font color.

getColor()        
        """
        return self.__color

    def setColor(self, col=text_color):
        """Set the font color.

setColor([color])

The argument "color" should be a color argument. If the method
is called without an argument, the font color is set to the
default text color.
        """
        if not isinstance(col, color.Color):
            raise TypeError, "Invalid color: " + `col`
        self.__color = col

    color = property(getColor, setColor, None, "Text color")


#
# the font_prop_map and parse_font() should be moved as
# they are GTK specific ...
#

font_prop_map = {
    'Oblique' : 'style',
    'Italic' : 'style',
    'Ultra-Light' : 'weight',
    'Light' : 'weight',
    'Medium' : 'weight',
    'Semi-Bold' : 'weight',
    'Bold' : 'weight',
    'Ultra-Bold' : 'weight',
    'Heavy' : 'weight',
    'Ultra-Condensed' : 'stretch',
    'Extra-Condensed' : 'stretch',
    'Condensed' : 'stretch',
    'Semi-Condensed' : 'stretch',
    'Semi-Expanded' : 'stretch',
    'Expanded' : 'stretch',
    'Extra-Expanded' : 'stretch',
    'Ultra-Expanded' : 'stretch',
    }

def parse_font(fontstr):
    size = 12
    weight = NORMAL
    style = NORMAL
    stretch = NORMAL
    family = 'Sans'
    if fontstr != '':
        fontlist = fontstr.split()
        fontlist.reverse()
        if fontlist[0].isdigit():
            size = fontlist.pop(0)
            _sz = int(size)
        while (fontlist[0] in font_prop_map):
            prop = fontlist.pop(0)
            item = font_prop_map[prop]
            # print "prop: " + prop
            # print "item: " + item
            if item == 'style':
                if prop == 'Oblique':
                    style = 1
                elif prop == 'Italic':
                    style = 2
                else:
                    style = NORMAL # default
            elif item == 'weight':
                if (prop == 'Ultra-Light' or
                    prop == 'Light' or
                    prop == 'Medium'):
                    weight = 1
                elif (prop == 'Semi-Bold' or
                      prop == 'Bold'):
                    weight = 2
                elif (prop == 'Ultra-Bold' or
                      prop == 'Heavy'):
                    weight = 3
                else:
                    weight = NORMAL
            elif item == 'stretch':
                stretch = prop # fixme - add stretching bits
            else:
                raise ValueError, "Unknown font property: " + item
        fontlist.reverse()
        if len(fontlist):
            family = ' '.join(fontlist)
    return (family, style, weight, stretch, size)
    
    
